/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Component, ComponentEventMap } from "../Component.js";
import { ObservableListenerOpts } from "../Observable.js";
import { Button } from "../Button.js";
import { MaterialIcon } from "../MaterialIcon";
import { Menu } from "../menu";
/**
 * @inheritDoc
 */
export interface FieldEventMap<Type> extends ComponentEventMap<Type> {
    /**
     * Fires when the field changes. It fires on blur.
     *
     * @param field
     */
    change: (field: Type, newValue: any, oldValue: any) => void;
    /**
     * Fires when setValue() is called
     *
     * @param field
     * @param newValue
     * @param oldValue
     */
    setvalue: (field: Type, newValue: any, oldValue: any) => void;
    /**
     * Fires when field is reset
     *
     * @param field
     * @param newValue
     * @param oldValue
     */
    reset: (field: Type, newValue: any, oldValue: any) => void;
    /**
     * Fires when validated
     *
     * Use {@see setInvalid()} to mark field invalid
     *
     * @param field
     */
    validate: (field: Type) => void;
    /**
     * Fires when the field is invalid
     *
     * Use {@see setInvalid()} to mark field invalid
     *
     * @param field
     */
    invalid: (field: Type) => void;
}
export interface Field extends Component {
    on<K extends keyof FieldEventMap<this>>(eventName: K, listener: Partial<FieldEventMap<this>>[K], options?: ObservableListenerOpts): void;
    fire<K extends keyof FieldEventMap<this>>(eventName: K, ...args: Parameters<FieldEventMap<any>[K]>): boolean;
}
export declare abstract class Field extends Component {
    private _buttons?;
    private toolbar?;
    private _wrap?;
    private _labelEl?;
    private _icon;
    private iconEl?;
    constructor(tagName?: keyof HTMLElementTagNameMap);
    readonly isFormField = true;
    protected baseCls: string;
    private _name?;
    private _required;
    private _readOnly;
    private _label;
    protected _value: any;
    protected resetValue: any;
    invalidMsg: string;
    private _hint;
    private hintEl?;
    /**
     * Used for "change" event
     * @protected
     */
    protected oldValue?: string;
    /**
     * Validate the field on blur
     */
    validateOnBlur: boolean;
    /**
     * Fires a change event if the field's value is different since it got focus
     * @protected
     */
    protected fireChangeOnBlur: boolean;
    protected onFocusOut(e: FocusEvent): void;
    /**
     * Return true if the field was modified
     */
    isDirty(): boolean;
    protected onFocusIn(e: FocusEvent): void;
    protected captureValueForChange(): void;
    protected internalRender(): HTMLElement;
    isFocusable(): boolean;
    get el(): HTMLElement;
    /**
     * A wrapper DIV element that contains input and toolbar for input buttons like an expand button for a drop down
     */
    get wrap(): HTMLDivElement;
    protected renderControl(): void;
    private renderButtons;
    /**
     * When buttons with menus are added it is handy to delay the validation on blur.
     * Because when the user will click something in the menu it will blur the field and you probably don't
     * want it to validate at that point. It's important that the menu will return focus to the field
     * and sets the value afterward.
     *
     * @param menu
     * @protected
     */
    protected setupMenu(menu: Menu): void;
    protected createControl(): HTMLElement | undefined;
    /**
     * Render buttons inside the text field
     *
     * @example
     * ```
     * buttons: [
     * 				 		btn({icon: "clear", handler:(btn) => (btn.parent!.parent! as Field).value = ""})
     * 					]
     * ```
     * @param buttons
     */
    set buttons(buttons: Button[] | undefined);
    get buttons(): Button[] | undefined;
    protected createHint(): HTMLDivElement | void;
    protected createLabel(): HTMLElement | void;
    private getLabelText;
    /**
     * Form element name which will be the key in values
     * If omitted the field won't be included in the form values.
     */
    get name(): string;
    /**
     * The field's name
     */
    set name(name: string);
    get required(): boolean;
    /**
     * Required or not
     */
    set required(required: boolean);
    get label(): string;
    /**
     * The field's label
     */
    set label(label: string);
    get icon(): MaterialIcon | "" | undefined;
    /**
     * The field's label
     */
    set icon(icon: MaterialIcon | "" | undefined);
    get hint(): string;
    /**
     * The field's hint text
     */
    set hint(hint: string);
    get readOnly(): boolean;
    /**
     * Make the field read only
     */
    set readOnly(readOnly: boolean);
    /**
     * Set the field value
     *
     * @param v
     * @param useForReset When true, this value is recorded for use when reset() is used.
     */
    set value(v: any);
    protected internalSetValue(v: any, old: any): void;
    /**
     * Helper to fire "change" event. Child classes must implement this.
     *
     * @param suppressSetValue If another helper component like a date picker just called "set value" then you can
     *  set this to true to prevent firing twice.
     */
    protected fireChange(): void;
    get value(): any;
    /**
     * Reset the field value to the initial value
     * The initial value is set by setValue() when "useForReset" is set to true.
     * @see setValue()
     */
    reset(): void;
    /**
     * Set the field as invalid and set a message
     *
     * @param msg
     */
    setInvalid(msg: string): void;
    /**
     * Check if the field is empty
     */
    isEmpty(): boolean;
    protected validate(): void;
    protected setValidityState(input: HTMLInputElement | HTMLTextAreaElement | HTMLSelectElement): void;
    protected applyInvalidMsg(): void;
    /**
     * Clears the invalid state
     */
    clearInvalid(): void;
    /**
     * Checks if the field is valid
     */
    isValid(): boolean;
    private createIcon;
}
//# sourceMappingURL=Field.d.ts.map