/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Component, createComponent } from "./Component.js";
import { DraggableComponent } from "./DraggableComponent.js";
/**
 * Splitter
 *
 * Used to resize panels
 */
export class Splitter extends DraggableComponent {
    /**
     *
     * @param resizeComponentPredicate 	The component to resize in height or width.
     *    Can be a component ID, itemId property, Component instance or custom function
     *
     */
    constructor(resizeComponentPredicate) {
        super("hr");
        this.resizeComponentPredicate = resizeComponentPredicate;
        /**
         * The minimum size it will set. Note that you can also put a min-width or min-height on the element with css.
         */
        this.minSize = 50;
        this.on("beforerender", () => {
            // find component to resize if it's an id string
            if (!(resizeComponentPredicate instanceof Component)) {
                this._resizeComponent = this.parent.findChild(this.resizeComponentPredicate);
                if (!this._resizeComponent) {
                    console.warn(this.resizeComponentPredicate);
                    throw "Could not find component to resize!";
                }
            }
            else {
                this._resizeComponent = resizeComponentPredicate;
            }
            if (this._state) {
                this.restoreState(this._state);
            }
        });
    }
    restoreState(state) {
        // we don't know the component to resize yet so store the state until
        // this component is added to a parent.
        this._state = state;
        if (this._resizeComponent) {
            if (this._state) {
                if (this._state.width)
                    this._resizeComponent.width = this._state.width;
                if (this._state.height)
                    this._resizeComponent.height = this._state.height;
            }
        }
    }
    buildState() {
        return this.resizeWidth ? { width: this._resizeComponent.el.offsetWidth } : { height: this._resizeComponent.el.offsetHeight };
    }
    internalRender() {
        const el = super.internalRender();
        this.on("dragstart", (comp, dragData, e) => {
            //resize width if this is a vertical splitter
            if (this.resizeWidth === undefined) {
                this.resizeWidth = this.el.offsetHeight > this.el.offsetWidth;
            }
            // if invert is undefined then auto detect based on the component order
            if (this.invert === undefined) {
                const splitterIndex = this.parent.findItemIndex(this);
                const resizeCmpIndex = this.parent.findItemIndex(this._resizeComponent);
                this.invert = splitterIndex < resizeCmpIndex;
            }
            if (this.resizeWidth) {
                dragData.data.startWidth = this._resizeComponent.el.offsetWidth;
            }
            else {
                dragData.data.startHeight = this._resizeComponent.el.offsetHeight;
            }
        });
        this.on("drag", (dc, dragData, ev) => {
            if (this.resizeWidth) {
                let offset = dragData.x - dragData.startX;
                if (this.invert) {
                    offset *= -1;
                }
                let width = Math.max(this.minSize, dragData.data.startWidth + offset);
                if (this.maxSize) {
                    width = Math.min(this.maxSize, width);
                }
                this._resizeComponent.width = width;
            }
            else {
                let offset = dragData.y - dragData.startY;
                if (this.invert) {
                    offset *= -1;
                }
                let height = Math.max(this.minSize, dragData.data.startHeight + offset);
                if (this.maxSize) {
                    height = Math.min(this.maxSize, height);
                }
                this._resizeComponent.height = height;
            }
        });
        this.on("drop", () => {
            this.saveState();
        });
        return el;
    }
}
/**
 * Shorthand function to create {@see Splitter}
 */
export const splitter = (config) => createComponent(new Splitter(config.resizeComponentPredicate), config);
//# sourceMappingURL=Splitter.js.map