/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Component } from "./Component.js";
import { Toolbar } from "./Toolbar.js";
import { DraggableComponent, DraggableComponentEventMap } from "./DraggableComponent.js";
import { Config, ObservableListenerOpts } from "./Observable.js";
/**
 * @inheritDoc
 */
export interface WindowEventMap<Type> extends DraggableComponentEventMap<Type> {
    /**
     * Fires when the window is closed
     *
     * @param window
     */
    close: (window: Type) => void;
    /**
     * Fires when the window is maximized
     *
     * @param window
     */
    maximize: (window: Type) => void;
    /**
     * Fires when the window is restored after being maximized
     *
     * @param window
     */
    unmaximize: (window: Type) => void;
}
export interface Window {
    on<K extends keyof WindowEventMap<this>>(eventName: K, listener: Partial<WindowEventMap<this>>[K], options?: ObservableListenerOpts): void;
    fire<K extends keyof WindowEventMap<this>>(eventName: K, ...args: Parameters<WindowEventMap<any>[K]>): boolean;
}
/**
 * Window component
 *
 * @example
 *
 * ```
 * const win = Window.create({
 * 	title: "Hello World",
 * 	items: [Component.create({tagName: "h1", cls: "pad", html: "Just saying hi!"})]
 * });
 *
 * win.open();
 * ```
 */
export declare class Window extends DraggableComponent {
    constructor();
    protected baseCls: string;
    /**
     * Maximize the window
     */
    maximized: boolean;
    /**
     * Enable tool to maximize window
     */
    maximizable: boolean;
    /**
     * Enable tool to close window
     */
    closable: boolean;
    /**
     * Enable tool to collapse window
     */
    collapsible: boolean;
    /**
     * Make the window modal so the user can only interact with this window.
     */
    modal: boolean;
    private titleCmp;
    private header;
    private modalOverlay;
    private resizeObserver?;
    /**
     * Return focus to element focussed before opening it when closing the window
     * @private
     */
    private focussedBeforeOpen?;
    private initMaximizeTool;
    protected applyTitle(): void;
    protected getDragHandle(): HTMLElement;
    getHeader(): Toolbar;
    set collapsed(collapsed: boolean);
    get collapsed(): boolean;
    protected internalRender(): HTMLElement;
    private observerResize;
    protected buildState(): Record<string, any>;
    protected restoreState(s: Record<string, any>): void;
    /**
     * Open the window by rendering it into the DOM body element
     * Use show()
     * @deprecated
     */
    open(): boolean;
    show(): boolean;
    private shrinkToFit;
    private disableBodyScroll;
    private enableBodyScroll;
    /**
     * @inheritDoc
     */
    remove(): boolean;
    /**
     * Close the window by removing it
     */
    close(): void;
    /**
     * Center the window in the screen
     */
    center(): this;
    /**
     * Returns true if the window is maximized
     */
    isMaximized(): boolean;
    /**
     * Grow window to the maximum of the viewport
     */
    maximize(): this;
    /**
     * Make the window smaller than the viewport
     */
    unmaximize(): this;
    static error(msg: string): Promise<void>;
    /**
     * Show modal alert window
     *
     */
    static alert(text: any, title?: string): Promise<void>;
    /**
     * Prompt the user for a text input value.
     *
     * Returns a promise with the input or undefined if the user cancelled
     *
     */
    static prompt(text: string, inputLabel: string, defaultValue?: string, title?: string): Promise<string | undefined>;
    /**
     * Ask the user for confirmation
     */
    static confirm(text: string, title?: string): Promise<boolean>;
}
/**
 * Shorthand function to create {@see Window}
 *
 * @param config
 * @param items
 */
export declare const win: (config?: Config<Window, WindowEventMap<Window>>, ...items: Component[]) => Window;
//# sourceMappingURL=Window.d.ts.map