/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { ContainerField, ContainerFieldValue } from "./ContainerField.js";
import { Config, ObservableListenerOpts } from "../Observable.js";
import { Component } from "../Component.js";
import { FieldEventMap } from "./Field.js";
export interface FormEventMap<Type, ValueType extends ContainerFieldValue = ContainerFieldValue> extends FieldEventMap<Type> {
    /**
     * Fires when the form is submitted. The event is fired after calling the handler.
     *
     * You can check if the form is valid after submitting with {@see form.isValid()}
     *
     * @param form
     */
    submit: (form: Type, handlerResponse: any) => any;
    /**
     * Not fired by the framework. But comes in handy when you extend this form and add a cancel button
     *
     * @param form
     */
    cancel: (form: Type) => any;
}
export interface Form<ValueType extends ContainerFieldValue = ContainerFieldValue> extends ContainerField<ValueType> {
    on<K extends keyof FormEventMap<this, ValueType>>(eventName: K, listener: Partial<FormEventMap<this, ValueType>>[K], options?: ObservableListenerOpts): void;
    fire<K extends keyof FormEventMap<this, ValueType>>(eventName: K, ...args: Parameters<FormEventMap<any, ValueType>[K]>): boolean;
    get el(): HTMLFormElement;
    get value(): ValueType;
    set value(v: Partial<ValueType>);
}
/**
 * Form component
 *
 * @example
 * ```
 * const form = form({
 * 	cls: "scroll fit",
 * 	handler: () => {
 *
 * 		console.log(form.getValues());
 *
 * 		const sub = <ContainerField>form.findField("sub");
 * 		const test1 = <TextField>sub.findField("test1");
 * 		test1.setInvalid("Hey something went wrong!");
 * 	},
 * 	items: [
 * 		Fieldset.create({
 * 			items: [
 *
 * 				TextField.create({
 * 					label: "Required field",
 * 					name: "test",
 * 					required: true
 * 				}),
 * 				DateField.create({
 * 					label: "Date",
 * 					name: "date"
 *
 * 				}),
 * 				HtmlField.create({
 * 					label: "Html"
 * 				}),
 *
 * 		    // This will create a "sub" object in the form values with the child components
 * 				ContainerField.create({
 * 					name: "sub",
 * 					items: [
 * 						TextField.create({
 * 							label: "A freaking long stupid label",
 * 							name: "test1",
 * 						}),
 * 						TextField.create({
 * 							label: "Test 2",
 * 							name: "test2",
 * 						}),
 * 					]
 * 				}),
 * 				CheckboxField.create({
 * 					label: "A checkbox label comes after",
 * 					name: "checkbox"
 * 				})
 * 			]
 * 		})
 * 	]
 * });
 *
 * @example Password validation
 *
 * ```
 * 	TextField.create({
 * 		type: "password",
 * 		label: "Password",
 * 		name: "password",
 * 		listeners: {
 * 			validate: (field) => {
 * 				const form = field.findAncestorByInstanceType(Form)!;
 * 				if(field.getValue() != form.findField("confirm")!.getValue()) {
 * 					field.setInvalid("The passwords don't match");
 * 				}
 * 			}
 * 		},
 * 	}),
 *
 * 	TextField.create({
 * 		itemId: "confirm",//item ID used instead of name so this field won't be submitted
 * 		type: "password",
 * 		label: "Confirm password"
 * 	}),
 *
 * ```
 *
 */
export declare class Form<ValueType extends ContainerFieldValue = ContainerFieldValue> extends ContainerField<ValueType> {
    protected baseCls: string;
    hideLabel: boolean;
    constructor();
    /**
     * Executed when form is submitted.
     *
     * If a promise is returned the "submit" event will fire after it has been resolved.
     *
     * @param form
     */
    handler: ((this: this, form: Form<ValueType>) => any | Promise<any>) | undefined;
    protected internalRender(): HTMLElement;
    /**
     * Get all form values
     */
    getValues(): {
        [key: string]: any;
    };
    /**
     * Set form field values
     *
     * @param v
     */
    setValues(v: Partial<ValueType>): void;
    /**
     * @inheritDoc
     */
    reset(): void;
    /**
     * Validates the form and submits it using the handler function passed with the config.
     */
    submit(): Promise<void>;
    setInvalid(msg: string): void;
}
/**
 * Shorthand function to create {@see Form}
 *
 * @param config
 * @param items
 */
export declare const form: <ValueType extends ContainerFieldValue = ContainerFieldValue>(config?: Config<Form<ValueType>, FormEventMap<Form<ValueType>, ContainerFieldValue>> | undefined, ...items: Component[]) => Form<ValueType>;
//# sourceMappingURL=Form.d.ts.map