/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Observable } from "../Observable.js";
import { ArrayUtil } from "../../util/ArrayUtil.js";
import { createComponent } from "../Component";
/**
 * Row selection model
 */
export class RowSelect extends Observable {
    constructor(list) {
        super();
        this.list = list;
        this._selected = [];
        /**
         * Last selected index used for multi selection with shift
         * @private
         */
        this.lastIndex = -1;
        this.multiSelect = true;
        this.hasKeyUpListener = false;
        this.list.on('beforerender', (me) => {
            const tableEl = me.el;
            tableEl.cls('+rowselect');
            tableEl.addEventListener("keydown", (e) => {
                this.onKeyDown(e);
            });
            me.on('rowmousedown', (table, index, row, e) => {
                this.onRowMouseDown(table, index, e);
            });
            tableEl.addEventListener("focus", (e) => {
                if (!this.selected.length && this.list.store.get(0)) {
                    this.selected = [0];
                }
            });
        });
    }
    clear() {
        this.selected = [];
    }
    selectAll() {
        const selected = [];
        for (let i = 0, c = this.list.store.count(); i < c; i++) {
            selected.push(i);
        }
        this.selected = selected;
    }
    /**
     * Get selected indexes
     *
     * Note that this is a copy and can't be edited directly. You have to set the selected property with a changed array.
     */
    get selected() {
        return [...this._selected];
    }
    set selected(newSelection) {
        this.setSelected(newSelection);
    }
    onRowMouseDown(_list, index, e) {
        let selection = this.selected;
        if (e.shiftKey && this.multiSelect) {
            const start = Math.min(index, this.lastIndex);
            const end = Math.max(index, this.lastIndex);
            for (let i = start; i <= end; i++) {
                if (selection.indexOf(i) === -1)
                    selection.push(i);
            }
        }
        else if ((e.ctrlKey || e.metaKey) && this.multiSelect) {
            const currentIndex = selection.indexOf(index);
            if (currentIndex > -1) {
                selection.splice(currentIndex, 1);
            }
            else {
                selection.push(index);
            }
        }
        else {
            selection = [index];
        }
        this.selected = selection;
    }
    add(index) {
        if (this._selected.indexOf(index) > -1) {
            return;
        }
        this._selected.push(index);
        this.fire('rowselect', this, index);
        this.fire('selectionchange', this);
    }
    /**
     * Remove an index from the selection
     *
     * @param index
     * @param silent Don't fire events
     */
    remove(index, silent = false) {
        const selectedIndex = this._selected.indexOf(index);
        if (selectedIndex == -1) {
            return;
        }
        this._selected.splice(selectedIndex, 1);
        if (!silent) {
            this.fire('rowdeselect', this, index);
            this.fire('selectionchange', this);
        }
    }
    /**
     * Set selected indexes
     *
     * @param newSelection
     * @param silent Suspends 'selectionchange' event
     */
    setSelected(newSelection, silent = false) {
        const old = this._selected;
        this._selected = newSelection;
        const deselect = ArrayUtil.diff(old, newSelection);
        const select = ArrayUtil.diff(newSelection, old);
        deselect.forEach(i => {
            this.fire('rowdeselect', this, i);
        });
        select.forEach((i, index) => {
            this.fire('rowselect', this, i);
        });
        if (newSelection.length) {
            this.lastIndex = newSelection[0];
        }
        const change = (select.length > 0 || deselect.length > 0);
        if (!silent && change) {
            this.fire('selectionchange', this);
        }
        return change;
    }
    onKeyDown(e) {
        if (e.key == "Escape") {
            this.clear();
            return;
        }
        if (e.key == "Shift") {
            this.shiftStartIndex = this.lastIndex;
        }
        if (e.key != "ArrowDown" && e.key != "ArrowUp") {
            return;
        }
        e.preventDefault();
        let index = 0, change = false;
        if (e.key == "ArrowDown") {
            if (this.lastIndex == this.list.store.count() - 1) {
                return;
            }
            index = this.lastIndex + 1;
        }
        else if (e.key == "ArrowUp") {
            if (this.lastIndex == 0) {
                return;
            }
            index = this.lastIndex - 1;
        }
        if (e.shiftKey && this.multiSelect) {
            const selected = this.selected;
            if ((e.key == "ArrowDown" && index > this.shiftStartIndex) || (e.key == "ArrowUp" && index < this.shiftStartIndex)) {
                if (selected.indexOf(index) == -1) {
                    selected.push(index);
                }
            }
            else {
                const removeIndex = selected.indexOf(this.lastIndex);
                if (removeIndex > -1) {
                    selected.splice(removeIndex, 1);
                }
            }
            change = this.setSelected(selected, true);
        }
        else {
            change = this.setSelected([index], true);
            this.list.focusRow(index);
        }
        if (change && !this.hasKeyUpListener) {
            this.hasKeyUpListener = true;
            this.list.el.addEventListener('keyup', () => {
                this.fire('selectionchange', this);
                this.hasKeyUpListener = false;
            }, { once: true });
        }
        this.lastIndex = index;
    }
}
/**
 * Shorthand function to create {@see RowSelect}
 *
 * @param config
 */
export const rowselect = (config) => createComponent(new RowSelect(config.list), config);
//# sourceMappingURL=RowSelect.js.map