/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Field } from "./Field.js";
import { createComponent } from "../Component.js";
import { E } from "../../util/Element.js";
/**
 * Checkbox field
 *
 * @see Form
 */
export class CheckboxField extends Field {
    /**
     *
     * @param type Render the checkbox as a checkbox, switch or toggle button
     */
    constructor(type = 'box') {
        super();
        this.type = type;
        this.baseCls = 'goui-form-field check';
        // we handle it with the native change event here
        this.fireChangeOnBlur = false;
        this.type = type;
        this.el.cls(type, true);
    }
    applyTitle() {
        if (this.title) {
            this.input.title = this.title;
        }
    }
    focus(o) {
        if (!this.input) {
            super.focus(o);
        }
        this.input?.focus(o);
    }
    createLabel() {
        const old = this.label;
        this.label = "";
        const span = super.createLabel();
        this.label = old;
        return span;
    }
    createControl() {
        const el = E('input').on("change", () => {
            this.fireChange();
        });
        el.type = "checkbox";
        el.required = this.required;
        el.name = this.name;
        el.readOnly = this.readOnly;
        el.checked = !!this.value;
        this.input = el;
        if (this.invalidMsg) {
            this.applyInvalidMsg();
        }
        const control = E('div', this.input);
        if (this.label) {
            const lbl = E('span', this.label).cls('box-label');
            if (this._cls) {
                lbl.cls("+" + this._cls);
            }
            control.append(lbl);
        }
        return control;
    }
    set readOnly(readOnly) {
        super.readOnly = readOnly;
        if (this.input) {
            this.input.readOnly = this.readOnly;
        }
    }
    get readOnly() {
        return super.readOnly;
    }
    setInvalid(msg) {
        super.setInvalid(msg);
        if (this.rendered) {
            this.applyInvalidMsg();
        }
    }
    applyInvalidMsg() {
        super.applyInvalidMsg();
        this.input.setCustomValidity(this.invalidMsg);
        //check if el is visible (https://stackoverflow.com/questions/19669786/check-if-element-is-visible-in-dom)
        if (this.input.offsetParent) {
            this.input.reportValidity();
        }
        if (this.invalidMsg != "") {
            //clear the field on change
            this.input.addEventListener('input', () => {
                this.clearInvalid();
            }, { once: true });
        }
    }
    clearInvalid() {
        super.clearInvalid();
        this.applyInvalidMsg();
    }
    getInput() {
        return this.input;
    }
    set value(v) {
        if (this.input) {
            this.input.checked = v;
        }
        super.value = v;
    }
    get value() {
        if (!this.input) {
            return super.value;
        }
        else {
            return this.input.checked;
        }
    }
    set name(name) {
        super.name = name;
        if (this.input) {
            this.input.name = this.name;
        }
    }
    get name() {
        return super.name;
    }
    validate() {
        super.validate();
        //this implements the native browser validation
        if (!this.input.validity.valid) {
            this.setInvalid(this.input.validationMessage);
        }
    }
}
/**
 * Shorthand function to create {@see CheckboxField}
 *
 * @param config
 */
export const checkbox = (config) => createComponent(new CheckboxField(config?.type || 'box'), config);
//# sourceMappingURL=CheckboxField.js.map