import { Field } from "./Field";
import { comp, Component, createComponent } from "../Component";
import { btn } from "../Button";
import { t } from "../../Translate";
/**
 * Chips component
 */
export class ChipsField extends Field {
    constructor() {
        super(...arguments);
        this.baseCls = 'goui-form-field chips';
        /**
         * Function that transforms the user text input to a chip.
         *
         * If it returns an empty value no chip will be rendered
         *
         * @param text
         */
        this.textInputToValue = async (text) => {
            return text;
        };
        /**
         * Renders a value to the chip component
         * @param value
         */
        this.chipRenderer = async (chip, value) => {
            chip.items.get(0).text = value;
        };
        this.selectedIndex = -1;
    }
    get editor() {
        return this._editor;
    }
    createControl() {
        this.items.on("datachanged", collection => {
            this.el.classList.toggle("has-chips", this.items.count() > 1);
        });
        this.chipsContainer = document.createElement("div");
        this.chipsContainer.classList.add("control-wrap");
        this.wrap.append(this.chipsContainer);
        this._editor = comp({
            id: Component.uniqueID(),
            attr: {
                contentEditable: "true"
            },
            cls: "editor"
        });
        this.el.setAttribute("for", this._editor.id);
        this._editor.el.addEventListener("keydown", (ev) => {
            this.onEditorKeyDown(ev);
        });
        this.el.addEventListener("click", (ev) => {
            if (this.selectedIndex == -1) {
                this._editor.focus();
            }
        });
        this._editor.el.addEventListener("focus", (ev) => {
            this.clearSelection();
        });
        this._editor.render(this.chipsContainer);
        this.items.add(this._editor);
        return undefined;
    }
    onEditorKeyDown(ev) {
        this.clearInvalid();
        switch (ev.key) {
            case "Enter":
                ev.preventDefault();
                const chip = this.createChip();
                this.textInputToValue(this._editor.text).then((value) => {
                    if (!value) {
                        return;
                    }
                    return this.chipRenderer(chip, value).then(() => {
                        this.items.insert(-1, chip);
                        this.value.push(value);
                    });
                });
                this._editor.text = "";
                break;
            case "ArrowRight":
                if (this.selectedIndex > -1) {
                    this.select(this.selectedIndex + 1);
                }
                break;
            case "ArrowLeft":
                if (this.selectedIndex == -1) {
                    if (this._editor.html == "") {
                        ev.preventDefault();
                        this.select(this.items.count() - 2);
                    }
                }
                else {
                    this.select(this.selectedIndex - 1);
                }
                break;
            case "Delete":
            case "Backspace":
                ev.stopPropagation();
                if (this._editor.html == "") {
                    if (this.selectedIndex == -1) {
                        ev.preventDefault();
                        this.select(this.items.count() - 2);
                    }
                    else {
                        this.items.get(this.selectedIndex).remove();
                        this.value.splice(this.selectedIndex, 1);
                        this.select(this.selectedIndex);
                    }
                    // this.focus();
                }
                break;
        }
    }
    createChip() {
        const chip = comp({
            cls: "chip hbox"
        }, comp({}), btn({
            title: t("Remove"),
            icon: "cancel",
            handler: (btn) => {
                this.captureValueForChange();
                const index = this.items.indexOf(chip);
                chip.remove();
                this.value.splice(index, 1);
                this.fireChange();
            }
        }));
        chip.el.addEventListener("click", () => {
            this.select(this.items.indexOf(chip));
            // this.focus();
        });
        return chip;
    }
    get itemContainerEl() {
        return this.chipsContainer;
    }
    clearSelection() {
        if (this.selectedIndex > -1) {
            const item = this.items.get(this.selectedIndex);
            if (item) {
                item.el.classList.remove("selected");
            }
            this.selectedIndex = -1;
        }
    }
    select(index) {
        if (index < 0) {
            return;
        }
        this.clearSelection();
        if (index > (this.items.count() - 2)) {
            this._editor.focus();
            return;
        }
        this.selectedIndex = index;
        this.items.get(this.selectedIndex).el.classList.add("selected");
    }
    internalSetValue(v, old) {
        if (this.rendered) {
            //remove all chips except the editor (last item).
            for (let i = 0; i < this.items.count() - 1; i++) {
                this.items.removeAt(i);
            }
            while (this.items.count() > 1) {
                this.items.removeAt(0);
            }
            this.renderValue();
        }
    }
    internalRender() {
        const el = super.internalRender();
        this.renderValue();
        return el;
    }
    renderValue() {
        if (!this.value) {
            return;
        }
        this.value.forEach((v) => {
            const chip = this.createChip();
            this.chipRenderer(chip, v).then(() => {
                this.items.insert(-1, chip);
            });
        });
    }
    focus(o) {
        this.editor.focus(o);
    }
}
/**
 * Shorthand function to create {@see ChipsField}
 *
 * @param config
 */
export const chips = (config) => createComponent(new ChipsField(), config);
//# sourceMappingURL=ChipsField.js.map