/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { ContainerField } from "./ContainerField.js";
import { Notifier } from "../../Notifier.js";
import { createComponent } from "../Component.js";
import { t } from "../../Translate.js";
/**
 * Form component
 *
 * @example
 * ```
 * const form = form({
 * 	cls: "scroll fit",
 * 	handler: () => {
 *
 * 		console.log(form.getValues());
 *
 * 		const sub = <ContainerField>form.findField("sub");
 * 		const test1 = <TextField>sub.findField("test1");
 * 		test1.setInvalid("Hey something went wrong!");
 * 	},
 * 	items: [
 * 		Fieldset.create({
 * 			items: [
 *
 * 				TextField.create({
 * 					label: "Required field",
 * 					name: "test",
 * 					required: true
 * 				}),
 * 				DateField.create({
 * 					label: "Date",
 * 					name: "date"
 *
 * 				}),
 * 				HtmlField.create({
 * 					label: "Html"
 * 				}),
 *
 * 		    // This will create a "sub" object in the form values with the child components
 * 				ContainerField.create({
 * 					name: "sub",
 * 					items: [
 * 						TextField.create({
 * 							label: "A freaking long stupid label",
 * 							name: "test1",
 * 						}),
 * 						TextField.create({
 * 							label: "Test 2",
 * 							name: "test2",
 * 						}),
 * 					]
 * 				}),
 * 				CheckboxField.create({
 * 					label: "A checkbox label comes after",
 * 					name: "checkbox"
 * 				})
 * 			]
 * 		})
 * 	]
 * });
 *
 * @example Password validation
 *
 * ```
 * 	TextField.create({
 * 		type: "password",
 * 		label: "Password",
 * 		name: "password",
 * 		listeners: {
 * 			validate: (field) => {
 * 				const form = field.findAncestorByInstanceType(Form)!;
 * 				if(field.getValue() != form.findField("confirm")!.getValue()) {
 * 					field.setInvalid("The passwords don't match");
 * 				}
 * 			}
 * 		},
 * 	}),
 *
 * 	TextField.create({
 * 		itemId: "confirm",//item ID used instead of name so this field won't be submitted
 * 		type: "password",
 * 		label: "Confirm password"
 * 	}),
 *
 * ```
 *
 */
export class Form extends ContainerField {
    constructor() {
        super("form");
        this.baseCls = "goui-form";
        this.hideLabel = true;
    }
    internalRender() {
        const el = super.internalRender();
        // disable browser validation
        this.el.noValidate = true;
        el.addEventListener("submit", (event) => {
            event.preventDefault();
            this.submit();
        });
        el.addEventListener("reset", (e) => {
            e.preventDefault();
            this.reset();
        });
        //Submit forms on CTRL + ENTER. (For html area or text area's)
        el.addEventListener("keydown", (e) => {
            if ((e.ctrlKey || e.metaKey) && e.key == "Enter") {
                e.preventDefault();
                if (document.activeElement && "blur" in document.activeElement) {
                    document.activeElement.blur();
                }
                this.submit();
            }
        });
        return el;
    }
    /**
     * Get all form values
     */
    getValues() {
        return this.value;
    }
    /**
     * Set form field values
     *
     * @param v
     */
    setValues(v) {
        this.value = v;
    }
    /**
     * @inheritDoc
     */
    reset() {
        this.findFields().forEach((field) => {
            field.reset();
        });
    }
    /**
     * Validates the form and submits it using the handler function passed with the config.
     */
    async submit() {
        const el = this.el;
        this.clearInvalid();
        if (this.isValid()) {
            el.cls(['+valid', '-invalid']);
            let handlerResponse = undefined;
            if (this.handler) {
                try {
                    handlerResponse = await this.handler(this);
                }
                catch (e) {
                    el.cls(['-valid', '+invalid']);
                    const msg = typeof (e) == "string" ? e : e.message;
                    Notifier.error(msg);
                    return;
                }
            }
            this.fire("submit", this, handlerResponse);
            //this.reset();
        }
        else {
            el.cls(['-valid', '+invalid']);
            const invalid = this.findFirstInvalid();
            if (invalid) {
                invalid.focus();
            }
            this.setInvalid(t('You have errors in your form. The invalid fields are marked.'));
        }
    }
    setInvalid(msg) {
        super.setInvalid(msg);
        Notifier.error(msg);
    }
}
/**
 * Shorthand function to create {@see Form}
 *
 * @param config
 * @param items
 */
export const form = (config, ...items) => createComponent(new Form, config, items);
//# sourceMappingURL=Form.js.map