/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Store } from "../data/Store.js";
import { ObjectUtil } from "../util/index.js";
import { createComponent } from "../component/index.js";
/**
 * DataSourceStore class
 *
 * Uses an {@see AbstractDataSource} to present data in the view.
 * @category Data
 */
export class DataSourceStore extends Store {
    constructor(dataSource) {
        super();
        this.dataSource = dataSource;
        this.queryParams = {};
        this.hasMore = false;
        // public properties?: string[] = [];
        /**
         * True when loaded at least once.
         */
        this.loaded = false;
        /**
         * Reload when the datasource changes
         */
        this.monitorChanges = true;
        /**
         * Builds record from entity
         * @param entity
         */
        this.buildRecord = async (entity) => entity;
        this.dataSource.on('change', (dataSource, changes) => {
            this.onChange(dataSource, changes);
        });
    }
    /**
     * Reloads the store when the datasource changes
     *
     * @protected
     */
    async onChange(DataSource, changes) {
        if (this.loaded && this.monitorChanges) {
            void this.reload();
        }
    }
    async internalLoad(append = false) {
        const queryParams = structuredClone(this.queryParams);
        queryParams.sort = this.sort;
        if (queryParams.limit) {
            //to see if the remote has more data we query one ID more.
            queryParams.limit++;
        }
        const queryResponse = await this.dataSource.query(queryParams);
        if (this.queryParams.limit) {
            // check if the server has more data.
            this.hasMore = queryResponse.ids.length > this.queryParams.limit;
            if (this.hasMore) {
                queryResponse.ids.pop();
            }
        }
        const getResponse = await this.dataSource.get(queryResponse.ids);
        const entities = await this.fetchRelations(getResponse.list), records = await Promise.all(entities.map(this.buildRecord));
        this.loadData(records, append);
        this.loaded = true;
        return records;
    }
    async fetchRelations(records) {
        if (!this.relations) {
            return records;
        }
        let relationName;
        const promises = [];
        for (relationName in this.relations) {
            const rel = this.relations[relationName];
            let id;
            for (let i = 0, l = records.length; i < l; i++) {
                id = ObjectUtil.path(records[i], rel.path);
                if (id) {
                    promises.push(rel.dataSource.single(id).then((e) => {
                        if (e) {
                            records[i][relationName] = e;
                        }
                    }));
                }
            }
        }
        return Promise.all(promises).then(() => records);
    }
    reload() {
        this.queryParams.position = 0;
        return super.reload();
    }
    loadNext(append = false) {
        if (!this.queryParams.limit) {
            throw new Error("Limit must be set for pagination");
        }
        this.queryParams.position = this.queryParams.position || 0;
        this.queryParams.position += this.queryParams.limit;
        return this.load(append);
    }
    loadPrevious() {
        if (!this.queryParams.limit) {
            throw new Error("Limit and position must be set!");
        }
        this.queryParams.position = this.queryParams.position || 0;
        this.queryParams.position -= this.queryParams.limit;
        return this.load(false);
    }
    hasNext() {
        return this.hasMore;
    }
    hasPrevious() {
        return this.queryParams.position > 0;
    }
    /**
     * Load more data when this element is scrolled down
     *
     * @param el
     */
    addScrollLoader(el) {
        const onScroll = () => {
            const pixelsLeft = el.scrollHeight - el.scrollTop - el.offsetHeight;
            if (pixelsLeft < 100) {
                if (!this.loading && this.hasNext()) {
                    void this.loadNext(true);
                }
            }
        };
        el.addEventListener("scroll", onScroll, { passive: true });
        // this will fill the empty space on firt load.
        this.on("load", (store, records, append) => {
            // use set timeout otherwise this.loading is still true
            setTimeout(() => {
                onScroll();
            });
        });
        return onScroll;
    }
}
/**
 * Shorthand function to create a {@see DataSourceStore}
 *
 * @param config
 */
export const datasourcestore = (config) => createComponent(new DataSourceStore(config.dataSource), config);
//# sourceMappingURL=DataSourceStore.js.map