var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};

// goui/script/util/FunctionUtil.ts
var BufferedFunction = class {
  /**
   * Constructor
   * @param delay Delay to execute function
   * @param fn Function to buffer
   */
  constructor(delay, fn) {
    this.delay = delay;
    this.fn = fn;
    this.delay = delay;
    this.fn = fn;
  }
  /**
   * Bugger the function with the delay set on this class
   *
   * @param args
   */
  buffer(args = []) {
    this.cancel();
    this.id = window.setTimeout(() => {
      this.cancel();
      this.fn.apply(null, args);
    }, this.delay);
  }
  /**
   * Cancel the function call
   */
  cancel() {
    if (this.id) {
      clearTimeout(this.id);
      this.id = void 0;
    }
  }
  /**
   * Check if it's still pending for execution
   */
  pending() {
    return this.id !== void 0;
  }
};
var FunctionUtil = class {
  /**
   * Buffer the function with this number of milliseconds.
   * When the function is called multiple times within the delay period only the last call will execute.
   *
   * @param delay
   * @param fn
   */
  static buffer(delay, fn) {
    const bf = new BufferedFunction(delay, fn);
    return (...args) => {
      bf.buffer(args);
    };
  }
  /**
   *
   * Delay function execution
   *
   * @param delay
   * @param fn
   */
  static delay(delay, fn) {
    return (...args) => {
      const bf = new BufferedFunction(delay, fn);
      bf.buffer(args);
    };
  }
  /**
   * Execute on the next repaint
   *
   * The onRepaint method tells the browser that you wish to perform an animation and requests
   * that the browser calls a specified function to update an animation before the next repaint.
   *
   * @link https://developer.mozilla.org/en-US/docs/Web/API/window/requestAnimationFrame
   * @param fn
   */
  static onRepaint(fn) {
    let frame = -1;
    return (...args) => {
      if (frame > -1) {
        cancelAnimationFrame(frame);
      }
      frame = window.requestAnimationFrame(() => {
        fn.apply(null, args);
      });
    };
  }
  /**
   * Will combine the given functions into one.
   *
   * The newFn function will be called with the return value of origFn function + the parameters of origFn
   * and function will return the value of newFn
   *
   * @param origFn
   * @param newFn
   *
   * @example
   * ```
   * function test(a:number, b:string) : string {
   * 	return b;
   * }
   *
   * FunctionUtil.createSequence(test, function(r, a, b) {
   * 	return r + "b";
   * })
   * ```
   *
   */
  static createSequence(origFn, newFn) {
    return function(...args) {
      const r = origFn.apply(this, args);
      return newFn.call(this, r, ...args);
    };
  }
  /**
   * Create a combined function of an orignal and new function. The new function will be called
   * before the original,
   *
   * @param origFn
   * @param newFn
   *
   * @example
   * ```
   * export function playgroundTableOverride() {
   * 	PlaygroundTable.prototype.render = FunctionUtil.createInterceptor(
   * 		PlaygroundTable.prototype.render,
   * 		function(this:PlaygroundTable) {
   * 			this.el.classList.add("cls-added-by-override");
   * 		}
   * 	)
   * }
   * ```
   */
  static createInterceptor(origFn, newFn) {
    return function(...args) {
      newFn.apply(this, args);
      return origFn.apply(this, args);
    };
  }
};

// goui/script/component/Observable.ts
var Observable = class {
  /**
   * Add a listener
   *
   * @param eventName
   * @param listener
   * @param options
   */
  on(eventName, listener, options) {
    const unbindkey = listener;
    if (options) {
      if (options.buffer !== void 0) {
        listener = FunctionUtil.buffer(options.buffer, listener);
      }
      if (options.once) {
        listener = this.once(eventName, listener);
      }
      if (options.delay) {
        listener = FunctionUtil.delay(options.delay, listener);
      }
    }
    this.lisnrs = this.lisnrs || {};
    this.lisnrs[eventName] = this.lisnrs[eventName] || [];
    if (options == null ? void 0 : options.unshift) {
      this.lisnrs[eventName].unshift({ listener, options, unbindkey });
    } else {
      this.lisnrs[eventName].push({ listener, options, unbindkey });
    }
  }
  once(eventName, listener) {
    const newfn = (...args) => {
      listener.apply(null, args);
      setTimeout(() => {
        this.un(eventName, listener);
      });
    };
    return newfn;
  }
  /**
   * Remove listener
   *
   * @param eventName
   * @param listener
   */
  un(eventName, listener) {
    if (!this.lisnrs || !this.lisnrs[eventName]) {
      return false;
    }
    for (let i = 0, l = this.lisnrs[eventName].length; i < l; i++) {
      const l2 = this.lisnrs[eventName][i];
      if (l2.unbindkey === listener) {
        this.lisnrs[eventName].splice(i, 1);
        return true;
      }
    }
    return false;
  }
  /**
   * Fire an event
   *
   * When a listener returns false this function will return false too.
   *
   * @param eventName
   * @param args
   */
  fire(eventName, ...args) {
    if (!this.lisnrs || !this.lisnrs[eventName]) {
      return true;
    }
    for (let l of this.lisnrs[eventName]) {
      if (l.listener.apply(null, args) === false) {
        return false;
      }
    }
    return true;
  }
  relayEvent(comp2, type) {
    comp2.on(type, (...args) => {
      this.fire(type, ...args);
    });
  }
};

// goui/script/State.ts
var State = class _State {
  static get() {
    return new _State();
  }
  hasItem(id) {
    return localStorage.getItem("state-" + id) !== null;
  }
  getItem(id) {
    const json = localStorage.getItem("state-" + id);
    if (!json) {
      return {};
    }
    return JSON.parse(json);
  }
  setItem(id, state) {
    localStorage.setItem("state-" + id, JSON.stringify(state));
  }
};

// goui/script/util/ArrayUtil.ts
var ArrayUtil = class {
  /**
   * Get array of all values of array1 that are NOT present in array2
   *
   * Opposite of intersect()
   *
   * @param array1
   * @param array2
   */
  static diff(array1, array2) {
    return array1.filter((i) => {
      return array2.indexOf(i) === -1;
    });
  }
  /**
   * Get array of all values of array1 that are ALSO present in array2
   *
   * @param array1
   * @param array2
   */
  static intersect(array1, array2) {
    return array1.filter((i) => {
      return array2.indexOf(i) !== -1;
    });
  }
  /**
   * Sort an array by multiple fields
   *
   * @param array
   * @param comparators
   */
  static multiSort(array, comparators) {
    if (!comparators.length) {
      return array;
    }
    const keySort = (a, b, isAscending) => {
      const direction = isAscending ? 1 : -1;
      if (a === b) {
        return 0;
      }
      return a > b ? direction : -1 * direction;
    };
    return array.sort((a, b) => {
      let sorted = 0;
      let index = 0;
      while (sorted === 0 && index < comparators.length) {
        const key = comparators[index].property;
        if (key) {
          sorted = keySort(a[key], b[key], comparators[index].isAscending || comparators[index].isAscending === void 0);
          index++;
        }
      }
      return sorted;
    });
  }
};

// goui/script/util/Browser.ts
var Browser = class {
  constructor() {
    this.cache = {};
    this.ua = navigator.userAgent.toLowerCase();
  }
  check(r, cacheKey) {
    if (!this.cache[cacheKey]) {
      this.cache[cacheKey] = r.test(this.ua);
    }
    return this.cache[cacheKey];
  }
  isWebkit() {
    return this.check(/webkit/, "isWebkit");
  }
  isChrome() {
    return this.check(/\bchrome\b/, "isChrome");
  }
  isSafari() {
    return !this.isChrome() && this.check(/safari/, "isSafari");
  }
  isFirefox() {
    return this.check(/gecko/, "isFirefox");
  }
  isMac() {
    return this.check(/macintosh|mac os x/, "isMac");
  }
  isLinux() {
    return this.check(/linux/, "isLinux");
  }
  isWindows() {
    return this.check(/windows|win32/, "isWindows");
  }
  /**
   * Open file upload dialog to get local files
   *
   * @example
   *
   * ```
   * btn({
   * 	type: "button",
   * 	text: t("Attach files"),
   * 	icon: "attach_file",
   * 	handler: async () => {
   *
   * 		const files = await browser.pickLocalFiles(true);
   * 		this.mask();
   * 		const blobs = await client.uploadMultiple(files);
   * 		this.unmask();
   * 	  console.warn(blobs);
   *
   * 	}
   * })
   * ```
   *
   * @param accept eg. "image/*"
   * @param directory Allow directory upload
   * @param multiple Allow multiple files
   */
  pickLocalFiles(multiple = false, directory = false, accept = void 0) {
    return new Promise((resolve, reject) => {
      const input = document.createElement("input");
      input.style.display = "none";
      input.setAttribute("type", "file");
      input.addEventListener("change", (e) => {
        if (input.files) {
          resolve(Array.from(input.files));
        }
        input.value = "";
      });
      document.body.appendChild(input);
      if (accept) {
        input.setAttribute("accept", accept);
      }
      if (directory) {
        input.setAttribute("webkitdirectory", "true");
        input.setAttribute("directory", "true");
      }
      if (multiple) {
        input.setAttribute("multiple", "true");
      }
      input.click();
      input.remove();
    });
  }
  /**
   * Copy text to clip board
   *
   * @param {string} text
   */
  copyTextToClipboard(text) {
    const al = document.activeElement;
    if (!navigator.clipboard) {
      const textArea = document.createElement("textarea");
      textArea.value = text;
      document.body.appendChild(textArea);
      textArea.focus();
      textArea.select();
      try {
        const successful = document.execCommand("copy");
        const msg = successful ? "successful" : "unsuccessful";
        console.log("Fallback: Copying text command was " + msg);
      } catch (err) {
        console.error("Fallback: Oops, unable to copy", err);
      }
      document.body.removeChild(textArea);
      if (al && al.focus) {
        al.focus();
      }
      return;
    }
    navigator.clipboard.writeText(text).then(function() {
      console.log("Async: Copying to clipboard was successful!");
      if (al && al.focus) {
        al.focus();
      }
    }, function(err) {
      console.error("Async: Could not copy text: ", err);
    });
  }
};
var browser = new Browser();

// goui/script/util/Collection.ts
var Collection = class extends Observable {
  constructor(items = []) {
    super();
    this.items = items;
  }
  [Symbol.iterator]() {
    return this.items[Symbol.iterator]();
  }
  /**
   * Add items at the end
   *
   * @returns the index of the last added item
   */
  add(...items) {
    let index = -1;
    items.forEach((item) => {
      index = this.items.length;
      if (!this.fire("beforeadd", this, item, index)) {
        return -1;
      }
      this.items.push(item);
      this.fire("add", this, item, index);
      this.fire("datachanged", this);
    });
    return index;
  }
  /**
   * Insert items at the given index
   *
   * @param index Use negative indexes to insert from the end. For example -1 inserts before the last item.
   */
  insert(index, ...items) {
    if (index < 0) {
      index = this.count() + index;
    }
    items.forEach((item) => {
      if (!this.fire("beforeadd", this, item, index)) {
        return -1;
      }
      this.items.splice(index, 0, item);
      this.fire("add", this, item, index);
      this.fire("datachanged", this);
      index++;
    });
    return index;
  }
  /**
   * Get an item at the given index
   * @param index
   */
  get(index) {
    return this.items[index];
  }
  /**
   * Return first item
   */
  first() {
    return this.get(0);
  }
  /**
   * return the last item
   */
  last() {
    return this.get(this.count() - 1);
  }
  /**
   * Find the index of an item. Returns -1 if not found.
   * @param item
   */
  indexOf(item) {
    return this.items.indexOf(item);
  }
  /**
   * Remove items
   */
  remove(...items) {
    items.forEach((item) => {
      const index = this.indexOf(item);
      if (index == -1) {
        return false;
      } else {
        return this.removeAt(index);
      }
    });
  }
  /**
   * Remove an item
   *
   * @param index Item index
   */
  removeAt(index) {
    const item = this.get(index);
    if (!item) {
      return false;
    }
    if (!this.fire("beforeremove", this, item, index)) {
      return false;
    }
    this.items.splice(index, 1);
    this.fire("remove", this, item, index);
    this.fire("datachanged", this);
    return true;
  }
  /**
   * Count the number of items
   */
  count() {
    return this.items.length;
  }
  /**
   * Clears all items
   */
  clear() {
    while (this.count()) {
      this.removeAt(0);
    }
    return this;
  }
  /**
   * Replace the collection with an array
   *
   * @param items
   */
  replace(...items) {
    this.clear().add(...items);
  }
  /**
   * Performs the specified action for each element in an array.
   * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the array.
   * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.
   */
  forEach(callbackfn, thisArg) {
    return this.items.forEach(callbackfn, thisArg);
  }
  getArray() {
    return this.items;
  }
  /**
   * Returns the value of the first element in the array where predicate is true, and undefined
   * otherwise.
   * @param predicate find calls predicate once for each element of the array, in ascending
   * order, until it finds one where predicate returns true. If such an element is found, find
   * immediately returns that element value. Otherwise, find returns undefined.
   *
   * @returns CollectionItem | undefined
   */
  find(predicate) {
    return this.items.find(predicate);
  }
  /**
   * The filter() method creates a new array with all elements that pass the test implemented by the provided function.
   *
   * @param predicate find calls predicate once for each element of the array, in ascending
   * order, until it finds one where predicate returns true. If such an element is found, find
   * immediately returns that element value. Otherwise, find returns undefined.
   *
   * @returns CollectionItem | undefined
   */
  filter(predicate) {
    return this.items.filter(predicate);
  }
  /**
   * Check if the given item is present
   *
   * @param item
   */
  has(item) {
    return this.findIndex((v) => v == item) > -1;
  }
  /**
   * Returns the index of the first element in the array where predicate is true, and -1
   * otherwise.
   * @param predicate find calls predicate once for each element of the array, in ascending
   * order, until it finds one where predicate returns true. If such an element is found,
   * findIndex immediately returns that element index. Otherwise, findIndex returns -1.
   */
  findIndex(predicate) {
    return this.items.findIndex(predicate);
  }
};

// goui/script/util/Cookies.ts
var Cookies = class {
  /**
   * Set a cookie
   *
   * @param {string} name
   * @param {string} value
   * @param {int} maxAge Maximum age in seconds. Leave empty to clear on browser close
   * @param {string} path
   * @returns {void}
   */
  set(name, value, maxAge, path) {
    path = path || document.location.pathname;
    let maxAgeStr = "";
    if (maxAge) {
      maxAgeStr = ";Max-Age=" + maxAge;
    }
    const secure = location.protocol === "https:" ? ";secure" : "";
    document.cookie = name + "=" + encodeURIComponent(value + "") + maxAgeStr + secure + ";path=" + path + ";SameSite=Strict";
  }
  /**
   * Get a cookie
   *
   * @param {string} name
   * @returns {string}
   */
  get(name) {
    const nameEQ = name + "=";
    const ca = document.cookie.split(";");
    for (let i = 0; i < ca.length; i++) {
      let c = ca[i];
      while (c.charAt(0) == " ")
        c = c.substring(1, c.length);
      if (c.indexOf(nameEQ) == 0)
        return decodeURIComponent(c.substring(nameEQ.length, c.length));
    }
    return void 0;
  }
  /**
   * Unset a cookie
   *
   * @param {string} name
   * @returns {void}
   */
  unset(name, path) {
    path = path || document.location.pathname;
    document.cookie = name + "=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=" + path + ";SameSite=Strict";
  }
};
var cookies = new Cookies();

// goui/script/util/DateTime.ts
var SystemTimeZone = Intl.DateTimeFormat().resolvedOptions().timeZone.toLowerCase();
function pad(n) {
  return (n < 10 ? "0" : "") + n;
}
var durationRegex = /(-)?P(?:([.,\d]+)Y)?(?:([.,\d]+)M)?(?:([.,\d]+)W)?(?:([.,\d]+)D)?(?:T(?:([.,\d]+)H)?(?:([.,\d]+)M)?(?:([.,\d]+)S)?)?/;
var _DateTime = class _DateTime {
  /**
   * Constructor
   *
   * @param date Can be a date object, a unix timestamp or date string (see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date#date_time_string_format)
   */
  constructor(date = /* @__PURE__ */ new Date()) {
    /**
     * The timezone of the date
     */
    this.timezone = SystemTimeZone;
    this.date = date instanceof Date ? date : new Date(date);
  }
  static staticInit(lang) {
    const locale = new Intl.Locale(lang), dayList = ["su", "mo", "tu", "we", "th", "fr", "sa"];
    if ("weekInfo" in locale) {
      _DateTime.firstWeekDay = locale.weekInfo.firstDay;
    }
    let tmp = /* @__PURE__ */ new Date("1970-01-01"), intlDays = new Intl.DateTimeFormat(lang, { weekday: "long" });
    for (let i = 0; i < 7; i++) {
      tmp.setDate(i + 4 + _DateTime.firstWeekDay);
      const d = dayList[tmp.getDay()];
      _DateTime.dayMap.push(d);
      _DateTime.dayNames[d] = intlDays.format(tmp);
    }
    let intlMonth = new Intl.DateTimeFormat(lang, { month: "long" });
    for (let i = 0; i < 12; i++) {
      tmp.setMonth(i);
      _DateTime.monthNames[i] = intlMonth.format(tmp);
    }
  }
  /**
   * User timezone offset in minutes
   *
   * eg.
   *
   * ```
   * const browser = new Date("2021-10-21T16:00:00+00:00");
   *
   * const amsterdam = utc.toTimezone("europe/amsterdam");
   *
   * console.log(amsterdam.getUserTimezoneOffset()); // 120 (in daylight savings time)
   * ```
   */
  getTimezoneOffset() {
    const systemTZOffset = this.date.getTimezoneOffset();
    return systemTZOffset + this.getSystemTimezoneDiff();
  }
  /**
   * Convert date to timezone
   *
   * @example On computer with Amsterdam timezone
   *
   * ```
   * const browser = new Date("2021-10-21T16:00:00+00:00");
   * console.log(browser.format("c"), browser.getTimeZone());
   * const riga = browser.toTimezone("europe/riga");
   *
   * console.log(riga.format("c"), riga.getTimeZone());
   *
   * const utc = riga.toUTC();
   *
   * console.log(utc.format("c"), utc.getTimeZone());
   *
   * ```
   *
   * Output:
   * 2021-10-21T18:00:00+02:00 Europe/Amsterdam
   * 2021-10-21T19:00:00+03:00 europe/riga
   * 2021-10-21T16:00:00+00:00 UTC
   *
   *
   * @param timezone eg. europe/amsterdam
   */
  toTimezone(timezone) {
    if (this.timezone == timezone) {
      return this.clone();
    }
    const offset = this.getTimezoneOffset();
    const d = this.clone();
    d.timezone = timezone;
    const newOffset = d.getTimezoneOffset();
    d.setMinutes(d.getMinutes() - newOffset + offset);
    return d;
  }
  diff(end) {
    let monthDays = end.clone().setDate(0).getDate(), sihdmy = [0, 0, 0, 0, 0, end.getYear() - this.getYear()], it = 0, map = { getSeconds: 60, getMinutes: 60, getHours: 24, getDate: monthDays, getMonth: 12 };
    for (let i2 in map) {
      let fn = i2;
      if (sihdmy[it] + end[fn]() < this[fn]()) {
        sihdmy[it + 1]--;
        sihdmy[it] += map[fn] - this[fn]() + end[fn]();
      } else if (sihdmy[it] + end[fn]() > this[fn]()) {
        sihdmy[it] += end[fn]() - this[fn]();
      }
      it++;
    }
    const [s, i, h, d, m, y] = sihdmy;
    return "P" + (y > 0 ? y + "Y" : "") + (m > 0 ? m + "M" : "") + (d > 0 ? d + "D" : "") + (h || i || s ? "T" + (h > 0 ? h + "H" : "") + (i > 0 ? i + "M" : "") + (s > 0 ? s + "S" : "") : "");
  }
  diffInDays(other) {
    return Math.floor((Date.UTC(other.getYear(), other.date.getMonth(), other.getDate()) - Date.UTC(this.getYear(), this.date.getMonth(), this.getDate())) / 864e5);
  }
  /**
   * Create a copy of this object without reference
   */
  clone() {
    return new _DateTime(new Date(this.date));
  }
  /**
   * Convert to UTC timezone
   */
  toUTC() {
    return this.toTimezone("UTC");
  }
  static getFormatter(timezone) {
    if (!_DateTime.cache[timezone]) {
      _DateTime.cache[timezone] = new Intl.DateTimeFormat("en-US", {
        timeZone: timezone,
        dateStyle: "short",
        timeStyle: "short"
      });
    }
    return _DateTime.cache[timezone];
  }
  /**
   * Get the offset between the system timezone and the date timezone in minutes.
   *
   * For example if the computer has europe/amsterdam and the date UTF it's 60 minutes in winter time.
   *
   * @private
   */
  getSystemTimezoneDiff() {
    if (this.timezone == SystemTimeZone) {
      return 0;
    }
    const local = _DateTime.getFormatter(this.timezone).format(this.date);
    const d = new Date(local);
    const u = Math.floor(this.date.getTime() / 1e3 / 60), localU = Math.floor(d.getTime() / 1e3 / 60);
    return u - localU;
  }
  /**
   * The ISO-8601 week number of year (weeks starting on Monday)
   */
  getWeekOfYear() {
    const ms1d = 864e5, ms7d = 7 * ms1d;
    const DC3 = Date.UTC(this.getYear(), this.getMonth() - 1, this.getMonthDay() + 3) / ms1d, AWN = Math.floor(DC3 / 7), Wyr = new Date(AWN * ms7d).getUTCFullYear();
    return AWN - Math.floor(Date.UTC(Wyr, 0, 7) / ms7d) + 1;
  }
  /**
   * Get the numeric day number of the year
   */
  getDayOfYear() {
    let num = 0, d = new _DateTime(this.date.getTime()), m = this.getMonth(), i;
    for (i = 1, d.setMonthDay(1), d.setMonth(1); i < m; d.setMonth(++i)) {
      num += d.getDaysInMonth();
    }
    return num + this.getMonthDay() - 1;
  }
  /**
   * Get the year
   */
  getYear() {
    return this.date.getFullYear();
  }
  /** Gets the month. Unlike JS it's 1 - 12 */
  getMonth() {
    return this.date.getMonth() + 1;
  }
  getDate() {
    return this.date.getDate();
  }
  getMonthDay() {
    return this.getDate();
  }
  /** 0 for sunday, 1 for monday, 2 for tuesday */
  getDay() {
    return this.date.getDay();
  }
  /**
   * Like getDay but take firstWeekDay of the week into account
   * 0 = first day of the week, 6 = last day
   */
  getWeekDay() {
    if (_DateTime.firstWeekDay == 1) {
      return (this.date.getDay() || 7) - 1;
    }
    return this.date.getDay();
  }
  getHours() {
    return this.date.getHours();
  }
  getMinutes() {
    return this.date.getMinutes();
  }
  getSeconds() {
    return this.date.getSeconds();
  }
  getMilliseconds() {
    return this.date.getMilliseconds();
  }
  /** Gets the time value in milliseconds. */
  getTime() {
    return this.date.getTime();
  }
  getMinuteOfDay() {
    return this.date.getHours() * 60 + this.date.getMinutes();
  }
  /**
   * Sets the hour value in the Date object
   * @param hours A numeric value equal to the hours value.
   * @params min A numeric value equal to the minutes value.
   * @param sec A numeric value equal to the seconds value.
   */
  setHours(hours, min = this.date.getMinutes(), sec = this.date.getSeconds(), ms = this.date.getMilliseconds()) {
    this.date.setHours(hours, min, sec, ms);
    return this;
  }
  setMinutes(min) {
    this.date.setMinutes(min);
    return this;
  }
  setSeconds(s) {
    this.date.setSeconds(s);
    return this;
  }
  setMilliseconds(ms) {
    this.date.setMilliseconds(ms);
    return this;
  }
  setYear(year) {
    this.date.setFullYear(year);
    return this;
  }
  setMonth(month) {
    this.date.setMonth(month - 1);
    return this;
  }
  setDate(date) {
    this.date.setDate(date);
    return this;
  }
  setMonthDay(date) {
    return this.setDate(date);
  }
  setWeekDay(day) {
    this.date.setDate(this.date.getDate() - this.getWeekDay() + day);
    return this;
  }
  /** Jump to day in current week */
  setDay(day) {
    this.date.setDate(this.date.getDate() - this.date.getDay() + day);
    return this;
  }
  addYears(years) {
    this.date.setFullYear(this.date.getFullYear() + years);
    return this;
  }
  addMonths(months) {
    this.date.setMonth(this.date.getMonth() + months);
    return this;
  }
  addDays(days) {
    this.date.setDate(this.date.getDate() + days);
    return this;
  }
  addHours(hours) {
    this.date.setHours(this.date.getHours() + hours);
    return this;
  }
  addMinutes(minutes) {
    this.date.setMinutes(this.date.getMinutes() + minutes);
    return this;
  }
  addSeconds(seconds) {
    this.date.setSeconds(this.date.getSeconds() + seconds);
    return this;
  }
  addDuration(iso8601) {
    let p2, matches = iso8601.match(durationRegex);
    matches.shift();
    const sign = matches.shift() || "";
    for (let o of ["FullYear", "Month", "Week", "Date", "Hours", "Minutes", "Seconds"]) {
      if (p2 = matches.shift()) {
        if (o === "Week") {
          p2 *= 7;
          o = "Date";
        }
        this.date["set" + o](this.date["get" + o]() + parseInt(sign + p2));
      }
    }
    return this;
  }
  /**
   * Check if current date is in a leap year
   */
  isLeapYear() {
    const year = this.getYear();
    return !!((year & 3) == 0 && (year % 100 || year % 400 == 0 && year));
  }
  /**
   * Get the number of days in the current month
   */
  getDaysInMonth() {
    const daysInMonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];
    const m = this.getMonth() - 1;
    return m == 1 && this.isLeapYear() ? 29 : daysInMonth[m];
  }
  getGMTOffset(colon = ":") {
    const tzo = this.getTimezoneOffset();
    return (tzo > 0 ? "-" : "+") + pad(Math.floor(Math.abs(tzo) / 60)) + colon + pad(Math.abs(tzo % 60));
  }
  /**
   * Format date similar to PHP's date function
   *
   * Note: indented options are NOT supported.
   *
   * d - The day of the month (from 01 to 31)
   * D - A textual representation of a day (three letters)
   * j - The day of the month without leading zeros (1 to 31)
   * l (lowercase 'L') - A full textual representation of a day
   * N - The ISO-8601 numeric representation of a day (1 for Monday, 7 for Sunday)
   * S - The English ordinal suffix for the day of the month (2 characters st, nd, rd or th. Works well with j)
   * w - A numeric representation of the day (0 for Sunday, 6 for Saturday)
   * z - The day of the year (from 0 through 365)
   * W - The ISO-8601 week number of year (weeks starting on Monday)
   * F - A full textual representation of a month (January through December)
   * m - A numeric representation of a month (from 01 to 12)
   * M - A short textual representation of a month (three letters)
   * n - A numeric representation of a month, without leading zeros (1 to 12)
   * t - The number of days in the given month
   * L - Whether it's a leap year (1 if it is a leap year, 0 otherwise)
   * o - The ISO-8601 year number
   * Y - A four digit representation of a year
   * y - A two digit representation of a year
   * a - Lowercase am or pm
   * A - Uppercase AM or PM
   * B - Swatch Internet time (000 to 999)
   * g - 12-hour format of an hour (1 to 12)
   * G - 24-hour format of an hour (0 to 23)
   * h - 12-hour format of an hour (01 to 12)
   * H - 24-hour format of an hour (00 to 23)
   * i - Minutes with leading zeros (00 to 59)
   * s - Seconds, with leading zeros (00 to 59)
   * u - Microseconds (added in PHP 5.2.2)
   * e - The timezone identifier (Examples: UTC, GMT, Atlantic/Azores)
   * I  (capital i) - Whether the date is in daylights savings time (1 if Daylight Savings Time, 0 otherwise)
   * O - Difference to Greenwich time (GMT) in hours (Example: +0100)
   * P - Difference to Greenwich time (GMT) in hours:minutes (added in PHP 5.1.3)
   * T - Timezone abbreviations (Examples: EST, MDT)
   * Z - Timezone offset in seconds. The offset for timezones west of UTC is negative (-43200 to 50400)
   * c - The ISO-8601 date (e.g. 2013-05-05T16:34:42+00:00)
   * r - The RFC 2822 formatted date (e.g. Fri, 12 Apr 2013 12:01:05 +0200)
   * U - The seconds since the Unix Epoch (January 1 1970 00:00:00 GMT)
   */
  format(format) {
    const chars = format.split("");
    let output = "";
    for (let i = 0, l = chars.length; i < l; i++) {
      let char = chars[i];
      if (char == "\\") {
        i++;
        if (chars.length > i + 1) {
          char += chars[i];
        }
      } else if (char in _DateTime.converters) {
        char = _DateTime.converters[char](this) + "";
      }
      output += char;
    }
    return output;
  }
  static createFormatRegex(format) {
    const chars = format.split("");
    let output = "";
    for (let i = 0, l = chars.length; i < l; i++) {
      let char = chars[i];
      switch (char) {
        case "Y":
          char = "(?<Y>\\d{4})";
          break;
        case "s":
        case "i":
        case "y":
          char = "(?<" + char + ">\\d{2})";
          break;
        case "G":
        case "H":
        case "m":
        case "d":
        case "g":
        case "h":
        case "j":
        case "n":
          char = "(?<" + char + ">\\d{1,2})";
          break;
        case "a":
          char = "(?<a>am|pm)";
          break;
        case "A":
          char = "(?<a>AM|PM)";
          break;
        case "P":
          char = "(?<P>[+-](\\d{2}):(\\d{2}))";
          break;
        case "c":
          char = _DateTime.createFormatRegex("Y-m-dTH:i:sP");
          break;
        default:
          break;
      }
      output += char;
    }
    return output;
  }
  /**
   * Create date by given format
   * See Date.format()
   * Supports:
   * Y, y, m, n, d, j, H, h, G, g, i, s, a, A
   *
   * @example
   * ```
   * const date = Date.createFromFormat("2021-10-21 21:09", "Y-m-d H:i"));
   *
   * const date = console.log(Date.createFromFormat("10/12/2021 9:09am", "m/d/Y g:ia", "America/New_York));
   * ```
   */
  static createFromFormat(dateStr, format = "c", timezone) {
    const regex = new RegExp(_DateTime.createFormatRegex(format), "u");
    const result = regex.exec(dateStr);
    if (!result) {
      return null;
    }
    const date = new _DateTime();
    date.setHours(0, 0, 0, 0);
    if (timezone) {
      date.timezone = timezone;
    }
    for (let key in result.groups) {
      switch (key) {
        case "Y":
          date.setYear(parseInt(result.groups["Y"]));
          break;
        case "y":
          date.setYear(parseInt(2e3 + result.groups["y"]));
          break;
        case "n":
        case "m":
          date.setMonth(parseInt(result.groups[key]));
          break;
        case "j":
        case "d":
          date.setMonthDay(parseInt(result.groups[key]));
          break;
        case "G":
        case "H":
          date.setHours(parseInt(result.groups[key]));
          break;
        case "h":
        case "g":
          const pm = result.groups.a ? result.groups.a == "pm" : result.groups.A ? result.groups.A == "PM" : false;
          const h = pm ? parseInt(result.groups[key]) : parseInt(result.groups[key]) + 12;
          date.setHours(h);
          break;
        case "i":
          date.setMinutes(parseInt(result.groups[key]));
          break;
        case "s":
          date.setSeconds(parseInt(result.groups[key]));
          break;
      }
    }
    return date;
  }
  /**
   * Compare with given date
   *
   * Returns:
   *
   * - -1 if this date is before the given date
   * - 0 if dates are equal
   * - 1 if this date is after the given date
   *
   * @param date
   * @return number
   */
  compare(date) {
    return (this.date > date.date ? 1 : 0) - (this.date < date.date ? 1 : 0);
  }
};
_DateTime.firstWeekDay = 1;
// 1 = monday, 7 = sunday
_DateTime.dayNames = {};
// {mo: t('Monday'), tu: t('Tuesday'), ...}
_DateTime.dayMap = [];
// ['mo','tu',...] if week starts on monday else index 0 = 'su'
_DateTime.monthNames = [];
_DateTime.cache = {};
_DateTime.converters = {
  "d": (date) => pad(date.getMonthDay()),
  "D": (date) => _DateTime.dayNames[_DateTime.dayMap[date.getWeekDay()]].substring(0, 3),
  "j": (date) => date.getMonthDay(),
  "l": (date) => _DateTime.dayNames[_DateTime.dayMap[date.getWeekDay()]],
  "S": (date) => ["st", "nd", "rd"][((date.getMonthDay() + 90) % 100 - 10) % 10 - 1] || "th",
  "w": (date) => date.getDay(),
  "z": (date) => date.getDayOfYear(),
  "W": (date) => date.getWeekOfYear(),
  "F": (date) => _DateTime.monthNames[date.getMonth() - 1],
  "m": (date) => pad(date.getMonth()),
  "M": (date) => _DateTime.monthNames[date.getMonth() - 1].substring(0, 3),
  "n": (date) => date.getMonth(),
  "Y": (date) => date.getYear(),
  "y": (date) => (date.getYear() + "").substr(-2),
  "a": (date) => date.getHours() > 12 ? "pm" : "am",
  "A": (date) => date.getHours() > 12 ? "PM" : "AM",
  "g": (date) => date.getHours() % 12,
  "G": (date) => date.getHours(),
  "h": (date) => pad(date.getHours() % 12),
  "H": (date) => pad(date.getHours()),
  "i": (date) => pad(date.getMinutes()),
  "s": (date) => pad(date.getSeconds()),
  "O": (date) => date.getGMTOffset(""),
  "P": (date) => date.getGMTOffset(),
  "U": (date) => Math.floor(date.getTime() / 1e3),
  "c": (date) => date.format("Y-m-dTH:i:sP")
};
var DateTime = _DateTime;
DateTime.staticInit(navigator.language);

// goui/script/util/Element.ts
function E(tag, ...items) {
  const el = document.createElement(tag);
  el.append(...items);
  return el;
}
Object.assign(Element.prototype, {
  on(event, listener, options) {
    this.addEventListener(event, listener, options);
    return this;
  },
  un(event, listener, options) {
    this.removeEventListener(event, listener, options);
    return this;
  },
  cls(name, enable) {
    if (!name)
      return this;
    if (Array.isArray(name)) {
      name.map((n) => {
        this.cls(n, enable);
      });
      return this;
    }
    name = name;
    if (enable !== void 0) {
      name = (enable ? "+" : "-") + name;
    }
    switch (name.substring(0, 1)) {
      case "+":
        this.classList.add(name.substring(1));
        break;
      case "-":
        this.classList.remove(name.substring(1));
        break;
      case "!":
        this.classList.toggle(name.substring(1));
        break;
      default:
        this.className = name;
    }
    return this;
  },
  attr(name, value) {
    if (value === void 0) {
      return this.getAttribute(name);
    }
    this.setAttribute(name, value);
    return this;
  },
  has(clsOrAttr) {
    if (clsOrAttr.substring(0, 1) === ".") {
      return this.classList.contains(clsOrAttr.substring(1));
    }
    return this.hasAttribute(clsOrAttr);
  },
  isA(tagName) {
    return this.tagName.toLowerCase() === tagName.toLowerCase();
  },
  up(expression, until) {
    if (!expression)
      return this.parentElement;
    let curr = this;
    do {
      if (curr === until)
        return null;
      if (curr.matches(expression))
        return curr;
    } while (curr = curr.parentElement);
  }
});

// goui/script/util/Format.ts
var _Format = class _Format {
  /**
   * Format a date to a string
   *
   * @param date
   */
  static date(date) {
    if (!date) {
      return "";
    }
    if (!(date instanceof DateTime)) {
      date = new DateTime(date);
    }
    return date.toTimezone(this.timezone).format(_Format.dateFormat);
  }
  /**
   * Format a time to a string
   * @param date
   */
  static time(date) {
    if (!(date instanceof DateTime)) {
      date = new DateTime(date);
    }
    return date.toTimezone(this.timezone).format(_Format.timeFormat);
  }
  /**
   * Formats a date and time to the default format
   *
   * @see Format.dateFormat
   * @see Format.timeFormat
   * @param date
   */
  static dateTime(date) {
    if (!(date instanceof DateTime)) {
      date = new DateTime(date);
    }
    return date.toTimezone(this.timezone).format(_Format.dateFormat + " " + _Format.timeFormat.replace("g", "h").replace("G", "H"));
  }
  /**
   * Format a number to money
   *
   * @param amount
   */
  static money(amount) {
    return this.currency + " " + this.number(amount, 2);
  }
  /**
   * Format a number to a localized string
   *
   * @param value
   * @param decimals
   */
  static number(value, decimals = 2) {
    const neg = value < 0;
    if (neg) {
      value *= -1;
    }
    const dec = this.decimalSeparator, tho = this.thousandsSeparator;
    const no = value.toFixed(decimals), parts = no.split(".");
    let formatted = "";
    const length = parts[0].length;
    for (let i = length - 1, l = 0; i >= l; i--) {
      formatted = parts[0][i] + formatted;
      if (i > 0 && (length - i) % 3 == 0) {
        formatted = tho + formatted;
      }
    }
    if (decimals) {
      formatted += dec + parts[1];
    }
    if (neg) {
      formatted = "-" + formatted;
    }
    return formatted;
  }
  static duration(value, zeroPad = false) {
    if (value < 0) {
      return "";
    }
    let retStr = "";
    const hours = Math.floor(value / 60), minutes = value % 60;
    ;
    if (zeroPad && hours < 10) {
      retStr += "0";
    }
    retStr += hours + ":";
    retStr += minutes < 10 ? "0" + minutes : minutes;
    return retStr;
  }
  static shortTime(v) {
    let arV = v.split(":");
    if (arV.length !== 3) {
      return "";
    }
    return arV[0] + ":" + arV[1];
  }
  static minutes(timeStr) {
    const parts = timeStr.split(":");
    return parseInt(parts[0]) * 60 + parseInt(parts[1]);
  }
};
/**
 * Date format when using date formatting functions
 */
_Format.dateFormat = "d-m-Y";
/**
 * Time format when using time formatting functions
 */
_Format.timeFormat = "G:i";
/**
 * Timezone when using time format functions
 */
_Format.timezone = Intl.DateTimeFormat().resolvedOptions().timeZone.toLowerCase();
/**
 * Currency to use
 */
_Format.currency = "\u20AC";
/**
 * Decimal separator when using number formatting functions
 */
_Format.decimalSeparator = ".";
/**
 * Escape a HTML string for safely inserting it in the DOM tree
 * @param str
 */
_Format.escapeHTML = function(str) {
  const p2 = document.createElement("p");
  p2.innerText = str;
  return p2.innerHTML;
};
var Format = _Format;

// goui/script/util/ObjectUtil.ts
var _ObjectUtil = class _ObjectUtil {
  /**
   * Simple JSON path function
   *
   * eg.
   * const obj = {
   *   foo : {
   *     bar: {
   *       test: 1
   *     }
   *   }
   * }
   *
   * Object.path(obj, 'foo.bar.test'); // 1
   *
   * @param obj
   * @param path
   * @return The value from the path or undefined if not found
   */
  static path(obj, path) {
    const dotPos = path.indexOf(".");
    const prop = dotPos > -1 ? path.substr(0, dotPos) : path;
    const next = obj[prop];
    if (!next) {
      return next;
    }
    if (dotPos == -1) {
      return next;
    }
    if (!_ObjectUtil.isObject(next)) {
      return void 0;
    }
    return this.path(next, path.substr(dotPos + 1));
  }
  /**
   * Deep merge two key value objects
   * @param o1
   * @param o2
   */
  static merge(o1, o2) {
    if (!this.isObject(o2)) {
      return o2;
    }
    for (let key in o2) {
      if (key in o1 && this.isObject(o1[key])) {
        o1[key] = this.merge(o1[key], o2[key]);
      } else {
        o1[key] = o2[key];
      }
    }
    return o1;
  }
  /**
   * Clone an object
   *
   * @param source
   */
  static clone(source) {
    return JSON.parse(JSON.stringify(source));
  }
  //
  // private static deepClone(source: any, hash = new WeakMap()): any {
  // 	if(Array.isArray(source)) {
  // 		return source.map(item => this.deepClone(item));
  // 	}
  //
  // 	if(source instanceof Date) {
  // 		return new Date(source.getTime());
  // 	}
  //
  // 	if(source && typeof source === 'object') {
  //
  // 		if (hash.has(source)) return hash.get(source);
  //
  //
  // 		const copy = Object.getOwnPropertyNames(source).reduce((o, prop) => {
  // 			Object.defineProperty(o, prop, Object.getOwnPropertyDescriptor(source, prop)!);
  // 			o[prop] = this.deepClone((source as { [key: string]: any })[prop]);
  // 			return o;
  // 		}, Object.create(Object.getPrototypeOf(source)));
  //
  // 		hash.set(source, copy);
  //
  // 		return copy;
  // 	}
  //
  // 	return source;
  // }
};
_ObjectUtil.isObject = (obj) => {
  return Object.prototype.toString.call(obj) === "[object Object]";
};
var ObjectUtil = _ObjectUtil;

// goui/script/util/QuoteStripper.ts
var QuoteStripper = class {
  constructor(body) {
    this.body = body;
  }
  getBodyWithoutQuote() {
    if (this.quote === void 0) {
      this.split();
    }
    return this.bodyWithoutQuote;
  }
  getQuote() {
    if (this.quote === void 0) {
      this.split();
    }
    return this.quote;
  }
  split() {
    let quoteIndex = this.findByBlockQuote();
    if (quoteIndex < 1) {
      quoteIndex = this.findQuoteByHeaderBlock();
    }
    if (quoteIndex > 0) {
      this.bodyWithoutQuote = this.body.substring(0, quoteIndex);
      this.quote = this.body.substring(quoteIndex);
    } else {
      this.bodyWithoutQuote = this.body;
      this.quote = "";
    }
  }
  findByBlockQuote() {
    this.quoteIndex = this.body.indexOf("<blockquote");
    return this.quoteIndex;
  }
  splitLines() {
    if (!this.lines) {
      const br = "|BR|";
      const html = this.body.replace(/<\/(p|div|ul|ol|table|dl)>/ig, "$&" + br).replace(/<br[^>]*>/ig, "$&" + br);
      this.lines = html.split(br);
    }
    return this.lines;
  }
  /**
   * eg
   *
   * Van: Merijn Schering [mailto:mschering@intermesh.nl]
    Verzonden: donderdag 20 november 2014 16:40
    Aan: Someone
    Onderwerp: Subject
   *
   * @return int|boolean
   */
  findQuoteByHeaderBlock() {
    const lines = this.splitLines();
    const greaterThan = /^&gt;(\s|&nbsp;)/;
    const header = /^[a-z]+:(\s|&nbsp;)+.*&lt;[a-z0-9._\-+&]+@[a-z0-9.\-_]+&gt;/i;
    const maybeHeader = /^[a-z]+:(\s|&nbsp;)+.*/i;
    let pos = 0, maybePos = 0;
    for (let i = 0, c = lines.length; i < c; i++) {
      const plain = lines[i].replace(/(<([^>]+)>)/ig, "");
      if (plain.match(maybeHeader)) {
        if (!maybePos) {
          maybePos = pos;
        }
      } else {
        maybePos = 0;
      }
      if (plain.match(header)) {
        return maybePos || pos;
      }
      if (plain.match(greaterThan)) {
        return pos;
      }
      pos += lines[i].length;
    }
    return -1;
  }
};

// goui/script/util/Recurrence.ts
var _Recurrence = class _Recurrence {
  constructor(config) {
    this.occurrence = 0;
    this.indices = {
      "BYSECOND": 0,
      "BYMINUTE": 0,
      "BYHOUR": 0,
      "BYDAY": 0,
      "BYMONTH": 0,
      "BYWEEKNO": 0,
      "BYMONTHDAY": 0
    };
    var _a;
    (_a = config.rule).interval || (_a.interval = 1);
    this.rule = config.rule;
    this.dtstart = config.dtstart;
    this.current = new DateTime(+this.dtstart);
    this.occurrence++;
    this.validate(this.rule);
    if (config.ff) {
      while (this.current.date < config.ff && this.next()) {
        console.log("ff ", this.current.date);
      }
    }
  }
  dayNb(shortName) {
    return { "mo": 1, "tu": 2, "wo": 3, "th": 4, "fr": 5, "sa": 6, "su": 0 }[shortName];
  }
  nDayHas(date) {
    for (const d of this.rule.byDay) {
      if (this.dayNb(d.day) === date.getDay() && d.nthOfPeriod == 1)
        return true;
    }
    return false;
  }
  next() {
    let previous = this.last ? this.last.clone() : null;
    if (this.rule.count && this.occurrence >= this.rule.count || this.rule.until && this.current.date > this.rule.until) {
      return null;
    }
    switch (this.rule.frequency) {
      case "daily":
        this.nextDaily();
        break;
      case "weekly":
        this.nextWeekly();
        break;
      case "monthly":
        this.nextMonthly();
        break;
      case "yearly":
        this.nextYearly();
        break;
      default:
        return null;
    }
    if (this.current == previous) {
      throw new Error("Recursion isn't going anywhere");
    }
    this.last = this.current;
    if (this.rule.until && this.current.date > this.rule.until) {
      return null;
    } else {
      this.occurrence++;
      console.log(this.occurrence);
      return this.current;
    }
  }
  // private nextHourly() {
  // 	this.current.setHours(this.current.getHours() + this.rule.interval);
  // }
  nextDaily() {
    if (!this.rule.byDay) {
      this.current.addDays(this.rule.interval);
      return;
    }
    do {
      this.current.addDays(this.rule.interval);
    } while (!this.nDayHas(this.current) || !this.rule.byMonth.includes("" + this.current.getMonth()));
  }
  nextWeekly() {
    if (!this.rule.byDay) {
      this.current.addDays(this.rule.interval * 7);
      return;
    }
    do {
      this.current.addDays(1);
      if (_Recurrence.dayMap[this.current.getDay()].toLowerCase() === this.rule.firstDayOfWeek) {
        this.current.addDays(this.rule.interval * 7);
        this.current.setDay(this.dayNb(this.rule.firstDayOfWeek));
      }
    } while (!this.nDayHas(this.current));
  }
  nextMonthly() {
    this.current.addMonths(this.rule.interval);
  }
  nextYearly() {
    if (!this.rule.byMonth) {
      this.current.addYears(this.rule.interval);
      return;
    } else {
      throw new Error("Not yet supported");
    }
  }
  validate(p2) {
    if ("byYearDay" in p2 && ("byMonth" in p2 || "byWeek" in p2 || "byMonthDay" in p2 || "byDay" in p2)) {
      throw new Error("Invalid byYearday rule");
    }
    if ("byWeekNo" in p2 && "byMonthDay" in p2) {
      throw new Error("byWeekNo does not fit to byMonthDay");
    }
    if (["daily", "weekly", "monthly", "yearly"].indexOf(p2.frequency) === -1) {
      throw new Error("Invalid frequency rule");
    }
    if (p2.frequency == "monthly" && ("byYearDay" in p2 || "byWeekNo" in p2)) {
      throw new Error("Invalid monthly rule");
    }
    if (p2.frequency == "weekly" && ("byYearDay" in p2 || "byMonthDay" in p2)) {
      throw new Error("Invalid weekly rule");
    }
    if (p2.frequency != "yearly" && "byYearDay" in p2) {
      throw new Error("byYearDay may only appear in yearly rules");
    }
  }
};
_Recurrence.dayMap = ["SU", "MO", "TU", "WE", "TH", "FR", "SA"];
var Recurrence = _Recurrence;

// goui/script/util/BrowserStorage.ts
var Connection = class {
  constructor(dbName = "goui") {
    this.dbName = dbName;
    this.enabled = true;
    this.upgrading = false;
  }
  /**
   * Workaround safari / webkit bug:
   *
   *  https://bugs.webkit.org/show_bug.cgi?id=226547
   * @return {*|Promise}
   */
  idbReady() {
    const isSafari = /Safari\//.test(navigator.userAgent) && !/Chrom(e|ium)\//.test(navigator.userAgent);
    if (!isSafari || !window.indexedDB.databases) {
      return Promise.resolve(this);
    }
    let intervalId;
    return new Promise((resolve) => {
      const tryIdb = () => {
        window.indexedDB.databases().finally(() => resolve(this));
      };
      intervalId = setInterval(tryIdb, 100);
      tryIdb();
    }).finally(() => clearInterval(intervalId));
  }
  /**
   * If upgrade is pending this promise resolves when done.
   */
  upgradeReady() {
    let intervalId;
    return new Promise((resolve) => {
      const checkUpgrading = () => {
        console.log("Check upgrade");
        if (!this.upgrading) {
          resolve(this);
        }
      };
      intervalId = setInterval(checkUpgrading, 100);
      checkUpgrading();
    }).finally(() => clearInterval(intervalId));
  }
  /**
   * Connect to the database
   */
  connect() {
    return __async(this, null, function* () {
      if (this.upgrading) {
        yield this.upgradeReady();
      }
      if (!this.conn) {
        this.conn = yield this.idbReady().then(() => {
          return new Promise((resolve, reject) => {
            const openreq = window.indexedDB.open(this.dbName);
            openreq.onerror = () => {
              this.enabled = false;
              console.warn("Disabling browser storage in indexedDB because browser doesn't support it.");
              reject(openreq.error);
            };
            openreq.onsuccess = () => {
              this.version = openreq.result.version;
              openreq.result.onversionchange = (e) => {
                console.warn("Version change");
                this.conn = void 0;
                openreq.result.close();
              };
              resolve(openreq.result);
            };
            openreq.onblocked = function() {
              console.log("IndexedDB upgrade blocked");
              reject("blocked");
            };
          });
        });
      }
      return this.conn;
    });
  }
  /**
   * Prepare database for upgrade.
   *
   * When resolved it MUST set conn.upgrading to false.
   */
  upgrade() {
    return __async(this, null, function* () {
      this.upgrading = true;
      this.disconnect();
      return new Promise((resolve, reject) => {
        let version = this.version + 1;
        const openreq = window.indexedDB.open(this.dbName, version);
        openreq.onerror = () => {
          this.enabled = false;
          console.warn("Disabling browser storage in indexedDB because browser doesn't support it.");
          reject(openreq.error);
        };
        openreq.onsuccess = () => {
          this.version = openreq.result.version;
          openreq.result.onversionchange = (e) => {
            console.warn("Version change");
            this.conn = void 0;
            openreq.result.close();
          };
          this.conn = openreq.result;
        };
        openreq.onblocked = (e) => {
          console.warn("IndexedDB upgrade blocked.");
        };
        openreq.onupgradeneeded = (e) => {
          resolve(e);
        };
      });
    });
  }
  /**
   * Disconnect from the database
   */
  disconnect() {
    if (this.conn) {
      this.conn.close();
      this.conn = void 0;
    }
  }
  /**
   * Delete the entire database
   */
  deleteDatabase() {
    return __async(this, null, function* () {
      if (!this.enabled) {
        return Promise.resolve(null);
      }
      return new Promise((resolve, reject) => {
        const openreq = indexedDB.deleteDatabase(this.dbName);
        openreq.onerror = () => reject(openreq.error);
        openreq.onsuccess = () => resolve(openreq.result);
      });
    });
  }
};
var browserStoreConnection = new Connection();
var BrowserStore = class {
  constructor(storeName) {
    this.storeName = storeName;
  }
  getStore(mode) {
    return __async(this, null, function* () {
      let db = yield browserStoreConnection.connect();
      if (!db.objectStoreNames.contains(this.storeName)) {
        db = yield this.createStore();
      }
      return db.transaction(this.storeName, mode).objectStore(this.storeName);
    });
  }
  createStore() {
    return __async(this, null, function* () {
      if (browserStoreConnection.upgrading) {
        yield browserStoreConnection.upgradeReady();
        let db = yield browserStoreConnection.connect();
        if (db.objectStoreNames.contains(this.storeName)) {
          return db;
        }
      }
      return new Promise((resolve, reject) => __async(this, null, function* () {
        const e = yield browserStoreConnection.upgrade(), db = e.target.result, t2 = e.target.transaction;
        db.createObjectStore(this.storeName);
        t2.oncomplete = () => {
          browserStoreConnection.upgrading = false;
          resolve(db);
        };
        t2.onerror = (e2) => {
          browserStoreConnection.upgrading = false;
          reject(e2);
        };
      }));
    });
  }
  requestPromise(req) {
    return new Promise((resolve, reject) => {
      req.onerror = (e) => {
        reject(e);
      };
      req.onsuccess = (e) => {
        resolve(req.result);
      };
    });
  }
  /**
   * Get an item from the store
   * returns undefined if not found.
   *
   * @param key
   */
  getItem(key) {
    return __async(this, null, function* () {
      if (!browserStoreConnection.enabled) {
        return Promise.resolve(void 0);
      }
      const store2 = yield this.getStore("readonly"), req = store2.get(key + "");
      return this.requestPromise(req);
    });
  }
  /**
   * Set an item
   *
   * @param key
   * @param value
   */
  setItem(key, value) {
    return __async(this, null, function* () {
      if (!browserStoreConnection.enabled) {
        return Promise.resolve(true);
      }
      const store2 = yield this.getStore("readwrite"), req = store2.put(value, key + "");
      return this.requestPromise(req).then(() => true).catch(() => false);
    });
  }
  /**
   * Remove an item
   *
   * @param key
   */
  removeItem(key) {
    return __async(this, null, function* () {
      if (!browserStoreConnection.enabled) {
        return Promise.resolve(key);
      }
      const store2 = yield this.getStore("readwrite"), req = store2.delete(key);
      return this.requestPromise(req).then(() => key);
    });
  }
  /**
   * Clear all data from this store
   */
  clear() {
    return __async(this, null, function* () {
      if (!browserStoreConnection.enabled) {
        return Promise.resolve(null);
      }
      const store2 = yield this.getStore("readwrite"), req = store2.clear();
      return this.requestPromise(req);
    });
  }
  keys() {
    return __async(this, null, function* () {
      if (!browserStoreConnection.enabled) {
        return Promise.resolve([]);
      }
      const keys = [];
      const store2 = yield this.getStore("readonly");
      return new Promise((resolve, reject) => {
        (store2.openKeyCursor || store2.openCursor).call(store2).onsuccess = function() {
          if (!this.result) {
            resolve(keys);
            return;
          }
          keys.push(this.result.key);
          this.result.continue();
        };
      });
    });
  }
};

// goui/script/component/Component.ts
var _Component = class _Component extends Observable {
  /**
   * Component constructor
   *
   * @param tagName The tagname used for the root HTMLElement of this component
   */
  constructor(tagName = "div") {
    super();
    this.tagName = tagName;
    /**
     * True when added to the DOM tree
     *
     * @private
     */
    this._rendered = false;
    /**
     * Component item ID that can be used to lookup the Component inside a Component with Component.findItem() and
     * Component.findItemIndex();
     *
     * if stateId is given it will also be used as itemId
     */
    this._itemId = "";
    /**
     * Set arbitrary data on a component.
     *
     * Should be used with caution as this data is not typed.
     */
    this.dataSet = {};
  }
  /**
   * Component item ID that can be used to lookup the Component inside a Component with Component.findItem() and
   * Component.findItemIndex();
   *
   * if stateId is given it will also be used as itemId
   */
  get itemId() {
    return this._itemId || this.stateId || this.el.id || "";
  }
  set itemId(itemId) {
    this._itemId = itemId;
  }
  initItems() {
    this.items.on("add", (_collection, item, index) => {
      item.parent = this;
      item.fire("added", item, index);
      if (this.rendered) {
        item.render();
      }
    });
    this.items.on("remove", (_collection, item) => {
      if (item.parent) {
        item.parent = void 0;
        item.remove();
      }
    });
  }
  getState() {
    return State.get().getItem(this.stateId);
  }
  hasState() {
    return this.stateId && State.get().hasItem(this.stateId);
  }
  /**
   * Restore state of the component in this function. It's called before render in init().
   *
   * @see saveState();
   * @param state
   *
   * @protected
   */
  restoreState(state) {
  }
  /**
   * Call save start when something relevant to the state changes.
   * Implement buildState() to save relevant state properties and restore it in restoreState()
   *
   * stateId must be set on components to be stateful
   *
   * @protected
   */
  saveState() {
    if (this.stateId) {
      State.get().setItem(this.stateId, this.buildState());
    }
  }
  /**
   * Build state for the component
   *
   * @see saveState();
   * @protected
   */
  buildState() {
    return {};
  }
  /**
   * Title of the dom element
   */
  set title(title) {
    this.el.title = title;
  }
  get title() {
    return this.el.title;
  }
  /**
   * Check if the component has been rendered and added to the DOM tree
   */
  get rendered() {
    return this._rendered;
  }
  get el() {
    if (!this._el) {
      this._el = document.createElement(this.tagName);
    }
    return this._el;
  }
  /**
   * Class name to add to element
   *
   * Some common classes to add for layout:
   *
   * - hbox: Set's flex layout to horizontal boxes. Use flex: n to stretch columns
   * - vbox: As above but vertical
   * - fit: Fit the parent's size
   * - scroll: Set's autoscroll: true
   * - pad: Set common padding on the element
   * - border-(top|bottom|left|right) to add a border
   *
   * Other:
   *
   * - goui-fade-in: The component will fade in when show() is used.
   * - goui-fade-out: The component will fade out when hide is used.
   *
   */
  set cls(cls) {
    this._cls = cls;
    this.initClassName();
    this.el.className += " " + cls;
  }
  initClassName() {
    if (this.el.classList.contains("goui")) {
      return;
    }
    this.el.classList.add("goui");
    if (this.baseCls) {
      this.el.className += " " + this.baseCls;
    }
  }
  /**
   * Renders the component and it's children
   */
  internalRender() {
    this.initClassName();
    this.renderItems();
    if (this.stateId) {
      this.restoreState(this.getState());
    }
    return this.el;
  }
  /**
   * The tabindex attribute specifies the tab order of an element (when the "tab" button is used for navigating).
   */
  set tabIndex(tabIndex) {
    this.el.tabIndex = tabIndex;
  }
  get tabIndex() {
    return this.el.tabIndex;
  }
  /**
   * CSS flex value
   */
  set flex(flex) {
    this.el.style.flex = flex + "";
  }
  /**
   * CSS flex value
   */
  get flex() {
    return this.el.style.flex;
  }
  /**
   * Make it resizable
   */
  set resizable(resizable) {
    if (resizable) {
      this.el.classList.add("goui-resizable");
    } else {
      this.el.classList.remove("goui-resizable");
    }
  }
  get resizable() {
    return this.el.classList.contains("goui-resizable");
  }
  /**
   * Render the component
   *
   * @param parentEl The element this componennt will render into
   * @param insertBefore If given, the element will be inserted before this child
   */
  render(parentEl, insertBefore) {
    if (this._rendered) {
      throw new Error("Already rendered");
    }
    this.fire("beforerender", this);
    if (!parentEl) {
      if (this.renderTo) {
        parentEl = this.renderTo;
      } else {
        if (!this.parent) {
          throw new Error("No parent set for " + typeof this);
        }
        parentEl = this.parent.itemContainerEl;
        insertBefore = this.getInsertBefore();
      }
    }
    if (!insertBefore) {
      parentEl.appendChild(this.el);
    } else {
      parentEl.insertBefore(this.el, insertBefore);
    }
    this.internalRender();
    this._rendered = true;
    this.fire("render", this);
    return this.el;
  }
  /**
   * Finds the DOM node in the parent's children to insert before when rendering a child component
   *
   * @protected
   */
  getInsertBefore() {
    if (!this.parent.rendered) {
      return void 0;
    }
    const index = this.parent.items.indexOf(this);
    let refItem = void 0;
    for (let i = index + 1, l = this.parent.items.count(); i < l; i++) {
      if (this.parent.items.get(i).rendered) {
        refItem = this.parent.items.get(i).el;
        break;
      }
    }
    return refItem;
  }
  /**
   * Remove component from the component tree
   */
  remove() {
    if (!this.fire("beforeremove", this)) {
      return false;
    }
    this.internalRemove();
    this.fire("remove", this);
    return true;
  }
  internalRemove() {
    this.items.clear();
    if (this.parent) {
      const p2 = this.parent;
      this.parent = void 0;
      p2.items.remove(this);
    }
    if (this.el) {
      this.el.remove();
    }
  }
  /**
   * Hide element
   */
  set hidden(hidden) {
    if (this.el.hidden == hidden) {
      return;
    }
    const eventName = hidden ? "hide" : "show";
    if (this.fire("before" + eventName, this) === false) {
      return;
    }
    this.el.hidden = hidden;
    this.fire(eventName, this);
  }
  get hidden() {
    return this.el.hidden;
  }
  /**
   * Hide this component
   *
   * This sets the "hidden" attribute on the DOM element which set's CSS display:none.
   * You can change this to fade with css class "goui-fade-in" and "goui-fade-out"
   */
  hide() {
    this.hidden = true;
    return this.hidden;
  }
  /**
   * Show the component
   */
  show() {
    this.hidden = false;
    return !this.hidden;
  }
  /**
   * Disable component
   */
  set disabled(disabled) {
    if (!disabled) {
      this.el.removeAttribute("disabled");
    } else {
      this.el.setAttribute("disabled", "");
    }
  }
  get disabled() {
    return this.el.hasAttribute("disabled");
  }
  /**
   * Set the HTML contents of the component (innerHTML)
   */
  set html(html) {
    this.el.innerHTML = html;
  }
  get html() {
    return this.el.innerHTML;
  }
  /**
   * Set the innerText
   */
  set text(text) {
    this.el.innerText = text;
  }
  get text() {
    return this.el.innerText;
  }
  /**
   * Width in pixels
   */
  set width(width) {
    this.el.style.width = width + "px";
  }
  get width() {
    return this.el.offsetWidth || parseFloat(this.el.style.width);
  }
  /**
   * Width in pixels
   */
  set style(style) {
    Object.assign(this.el.style, style);
  }
  get style() {
    return this.el.style;
  }
  computeZIndex() {
    const z = parseInt(window.getComputedStyle(this.el).getPropertyValue("z-index"));
    if (!z) {
      return this.parent ? this.parent.computeZIndex() : 0;
    } else {
      return z;
    }
  }
  /**
   * height in pixels
   */
  set height(height) {
    this.el.style.height = height + "px";
  }
  get height() {
    return this.el.offsetHeight || parseFloat(this.el.style.height);
  }
  /**
   * Element ID
   */
  set id(id) {
    this.el.id = id;
  }
  get id() {
    return this.el.id;
  }
  /**
   * Focus the component
   *
   * @param o
   */
  focus(o) {
    this.el.focus(o);
    this.fire("focus", this, o);
  }
  isFocusable() {
    return this.el && !this.hidden && (this.el.tagName == "BUTTON" || this.el.tagName == "INPUT" || this.el.tagName == "A" || this.el.tagName == "AREA" || this.el.tabIndex > -1);
  }
  /**
   * Get the component that's next to this one
   */
  nextSibling() {
    if (!this.parent) {
      return void 0;
    }
    const index = this.parent.items.indexOf(this);
    if (index == -1) {
      return void 0;
    }
    return this.parent.items.get(index + 1);
  }
  /**
   * Get the component that's previous to this one
   */
  previousSibling() {
    if (!this.parent) {
      return void 0;
    }
    const index = this.parent.items.indexOf(this);
    if (index < 1) {
      return void 0;
    }
    return this.parent.items.get(index - 1);
  }
  /**
   * Find ancestor
   *
   * The method traverses the Component's ancestors (heading toward the document root) until it finds
   * one where the given function returns true.
   *
   * @param fn When the function returns true the item will be returned. Otherwise it will move up to the next parent.
   */
  findAncestor(fn) {
    let p2 = this.parent;
    while (p2 != void 0) {
      if (fn(p2)) {
        return p2;
      } else {
        p2 = p2.parent;
      }
    }
    return void 0;
  }
  /**
   * Find parent by instance type of the parent
   *
   * @example
   * ```
   * const form = textField.findAncestorByType(Form);
   * ```
   * @param cls
   */
  findAncestorByType(cls) {
    const p2 = this.findAncestor((Component16) => Component16 instanceof cls);
    if (p2) {
      return p2;
    } else {
      return void 0;
    }
  }
  /**
   * The child components of this component
   */
  get items() {
    if (!this._items) {
      this._items = new Collection();
      this.initItems();
    }
    return this._items;
  }
  renderItems() {
    if (this._items) {
      this._items.forEach((item) => {
        if (item.hidden) {
          item.on("show", (item2) => {
            this.renderItem(item2);
          }, { once: true });
        } else {
          this.renderItem(item);
        }
      });
    }
  }
  get itemContainerEl() {
    return this.el;
  }
  /**
   * Can be overriden to wrap the component
   *
   * @param item
   * @protected
   */
  renderItem(item) {
    item.render(this.itemContainerEl, item.getInsertBefore ? item.getInsertBefore() : void 0);
  }
  /**
   * Find the item by element ID, itemId property, Component instance or custom function
   */
  findItemIndex(predicate) {
    let fn = this.createFindPredicateFunction(predicate);
    return this.items.findIndex(fn);
  }
  /**
   * Find the item by element ID, itemId property, Component instance or custom function.
   *
   * If you want to search the component tree hierarchy use {@see findChild()}
   *
   */
  findItem(predicate) {
    let fn = this.createFindPredicateFunction(predicate);
    return this.items.find(fn);
  }
  /**
   * Cascade down the component hierarchy
   *
   * @param fn When the function returns false then the cascading will be stopped. The current Component will be finished!
   */
  cascade(fn) {
    if (fn(this) === false) {
      return this;
    }
    if (this.items) {
      for (let cmp of this.items) {
        cmp.cascade && cmp.cascade(fn);
      }
    }
    return this;
  }
  createFindPredicateFunction(predicate) {
    if (predicate instanceof Function) {
      return predicate;
    } else {
      return (item) => {
        return item === predicate || item.itemId === predicate || item.id === predicate;
      };
    }
  }
  /**
   * Find a child at any level by element ID, itemId property, Component instance or custom function.
   *
   * It cascades down the component hierarchy. See also {@see findChildByType}
   *
   */
  findChild(predicate) {
    let fn = this.createFindPredicateFunction(predicate);
    let child;
    this.cascade((item) => {
      if (fn(item)) {
        child = item;
        return false;
      }
    });
    return child;
  }
  /**
   * Find children at any level by element ID, itemId property, Component instance or custom function.
   *
   * It cascades down the component hierarchy. See also {@see findChildByType}
   *
   */
  findChildren(predicate) {
    let fn = this.createFindPredicateFunction(predicate);
    const children = [];
    this.cascade((item) => {
      if (fn(item)) {
        children.push(item);
      }
    });
    return children;
  }
  /**
   * Find child by instance type of the parent
   *
   * @example
   * ```
   * const form = textField.findAncestorByType(Form);
   * ```
   * @param cls
   */
  findChildByType(cls) {
    const p2 = this.findChild((Component16) => Component16 instanceof cls);
    if (p2) {
      return p2;
    } else {
      return void 0;
    }
  }
  /**
   * Find children by instance type of the parent
   *
   * @example
   * ```
   * const form = textField.findAncestorByType(Form);
   * ```
   * @param cls
   */
  findChildrenByType(cls) {
    return this.findChildren((Component16) => Component16 instanceof cls);
  }
  /**
   * Set attributes of the DOM element
   *
   * @param attr
   */
  set attr(attr) {
    for (let name in attr) {
      this.el.setAttribute(name, attr[name]);
    }
  }
  /**
   * Mask the component to disable user interaction
   * It creates an absolute positioned Mask
   * component. This component should have a non-static position style for this to work.
   */
  mask(delay = 300) {
    this.maskTimeout = setTimeout(() => {
      if (!this._mask) {
        this._mask = mask({ spinner: true });
        this.items.add(this._mask);
      }
      this.el.classList.add("masked");
      this._mask.show();
      this.maskTimeout = void 0;
    }, delay);
  }
  /**
   * Unmask the body
   */
  unmask() {
    if (this.maskTimeout) {
      clearTimeout(this.maskTimeout);
      this.maskTimeout = void 0;
      return;
    }
    if (this._mask) {
      this._mask.hide();
    }
    this.el.classList.remove("masked");
  }
  static uniqueID() {
    return "goui-" + ++_Component._uniqueID;
  }
};
_Component._uniqueID = 0;
var Component = _Component;
var Mask = class extends Component {
  constructor() {
    super(...arguments);
    this.baseCls = "goui-mask";
  }
  /**
   * Show loading spinner
   */
  set spinner(spinner) {
    if (spinner) {
      this.html = '<div class="spinner"><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div></div>';
    }
  }
};
var mask = (config) => createComponent(new Mask(), config);
var comp = (config, ...items) => createComponent(new Component(config == null ? void 0 : config.tagName), config, items);
var p = (config, ...items) => createComponent(new Component("p"), typeof config == "string" ? { html: config } : config, items);
var small = (config, ...items) => createComponent(new Component("small"), typeof config == "string" ? { html: config } : config, items);
var h1 = (config, ...items) => createComponent(new Component("h1"), typeof config == "string" ? { html: config } : config, items);
var h2 = (config, ...items) => createComponent(new Component("h2"), typeof config == "string" ? { html: config } : config, items);
var h3 = (config, ...items) => createComponent(new Component("h3"), typeof config == "string" ? { html: config } : config, items);
var h4 = (config, ...items) => createComponent(new Component("h4"), typeof config == "string" ? { html: config } : config, items);
var code = (config, ...items) => createComponent(new Component("code"), typeof config == "string" ? { html: config } : config, items);
var section = (config, ...items) => createComponent(new Component("section"), typeof config == "string" ? { html: config } : config, items);
var hr = (config) => createComponent(new Component("hr"), config);
var progress = (config) => createComponent(new Component("progress"), config);
var createComponent = (comp2, config, items) => {
  if (config) {
    if (config.listeners) {
      for (let key in config.listeners) {
        const eventName = key;
        if (typeof config.listeners[eventName] == "function") {
          comp2.on(eventName, config.listeners[eventName]);
        } else {
          const o = config.listeners[eventName];
          const fn = o.fn;
          delete o.fn;
          comp2.on(eventName, fn, o);
        }
      }
      delete config.listeners;
    }
    Object.assign(comp2, config);
  }
  if (items && items.length) {
    comp2.items.add(...items);
  }
  return comp2;
};

// goui/script/component/Toolbar.ts
var Toolbar = class _Toolbar extends Component {
  constructor() {
    super("menu");
    this.baseCls = "goui-toolbar";
    /**
     * Used by keyboard nav
     * @protected
     */
    this.orientation = "horizontal";
    this.focusedItemIndex = -1;
    this.setupKeyboardNav();
  }
  /**
   * Find the first menu in the tree of submenu's
   */
  findToolbar() {
    let parent = this.parent;
    if (!parent || !(parent instanceof _Toolbar)) {
      return void 0;
    }
    if (parent.orientation == "horizontal") {
      return parent;
    } else {
      return parent.findToolbar();
    }
  }
  setupKeyboardNav() {
    this.items.on("add", (collection, item, index) => {
      item.el.addEventListener("click", () => {
        this.focusedItemIndex = index;
      });
    });
    this.on("focus", () => {
      if (this.focusedItemIndex > -1) {
        this.items.get(this.focusedItemIndex).focus();
      }
    });
    this.on("hide", () => {
      this.focusedItemIndex = -1;
    });
    this.el.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "ArrowRight":
          if (this.orientation == "vertical") {
            if (!this.focusChild()) {
              const tb = this.findToolbar();
              if (tb) {
                tb.focusNext();
              }
            }
          } else {
            this.focusNext();
          }
          ev.stopPropagation();
          ev.preventDefault();
          break;
        case "ArrowDown":
          if (this.orientation == "vertical") {
            this.focusNext();
          } else {
            this.focusChild();
          }
          ev.stopPropagation();
          ev.preventDefault();
          break;
        case "ArrowLeft":
          if (this.orientation == "vertical") {
            if (!this.focusParent()) {
              const tb = this.findToolbar();
              if (tb) {
                tb.focusNext(-1);
              }
            }
          } else {
            this.focusNext(-1);
          }
          ev.stopPropagation();
          ev.preventDefault();
          break;
        case "ArrowUp":
          if (this.orientation == "vertical") {
            this.focusNext(-1);
          } else {
            this.focusParent();
          }
          ev.stopPropagation();
          ev.preventDefault();
          break;
      }
    });
  }
  focusNext(inc = 1) {
    const nextIndex = this.focusedItemIndex + inc;
    this.focusedItemIndex = Math.min(Math.max(nextIndex, 0), this.items.count() - 1);
    if (nextIndex != this.focusedItemIndex) {
      return false;
    }
    const cmp = this.items.get(this.focusedItemIndex);
    if (!cmp.isFocusable()) {
      return this.focusNext(inc);
    } else {
      cmp.focus();
      if (this.orientation == "horizontal") {
        cmp.el.click();
      }
      return true;
    }
  }
  focusChild() {
    const child = this.items.get(this.focusedItemIndex);
    if (!child || !child.menu) {
      return false;
    }
    child.menu.focusNext();
    return true;
  }
  focusParent() {
    const child = this.items.get(this.focusedItemIndex);
    if (!child) {
      return false;
    }
    if (child.parent) {
      child.parent.focus();
    }
    return true;
  }
};
var tbar = (config, ...items) => {
  const c = new Toolbar();
  if (config) {
    Object.assign(c, config);
  }
  c.items.add(...tbarItems(items));
  return c;
};
var tbarItems = (items) => {
  const l = items.length;
  if (l) {
    for (let i = 0; i < l; i++) {
      switch (items[i]) {
        case "->":
          items[i] = comp({
            flex: 1
          });
          break;
        case "-":
          items[i] = comp({ tagName: "hr" });
          break;
      }
    }
  }
  return items;
};

// goui/script/Translate.ts
var Translate = class {
  constructor() {
    this.lang = {};
    this.missing = {};
    this.defaultPkg = "core";
    this.defaultModule = "core";
  }
  setDefaultModule(pkg, module) {
    this.defaultModule = module;
    this.defaultPkg = pkg;
  }
  load(lang, pkg = "core", module = "core") {
    if (!this.lang[pkg]) {
      this.lang[pkg] = {};
    }
    if (!this.lang[pkg][module]) {
      this.lang[pkg][module] = {};
    }
    ObjectUtil.merge(this.lang[pkg][module], lang);
    this.setDefaultModule(pkg, module);
  }
  /**
   * Translate a string
   *
   * @param key Translate key. Usually the english text.
   * @param pkg The module package
   * @param module The module name
   */
  static t(key, pkg, module) {
    var _a, _b, _c;
    if (!pkg) {
      pkg = translate.defaultPkg;
    }
    if (!module) {
      module = translate.defaultModule;
    }
    if ((_c = (_b = (_a = translate.lang) == null ? void 0 : _a[pkg]) == null ? void 0 : _b[module]) == null ? void 0 : _c[key]) {
      return translate.lang[pkg][module][key];
    } else if (pkg == "core" && module == "core") {
      if (!translate.missing[pkg]) {
        translate.missing[pkg] = {};
      }
      if (!translate.missing[pkg][module]) {
        translate.missing[pkg][module] = {};
      }
      translate.missing[pkg][module][key] = key;
      return key;
    } else {
      return t(key, "core", "core");
    }
  }
};
var translate = new Translate();
var t = Translate.t;

// goui/script/component/form/Field.ts
var Field = class extends Component {
  constructor(tagName = "label") {
    super(tagName);
    this.isFormField = true;
    this.baseCls = "goui-form-field";
    this._required = false;
    this._readOnly = false;
    this._label = "";
    this.invalidMsg = "";
    this._hint = "";
    /**
     * Validate the field on blur
     */
    this.validateOnBlur = true;
    /**
     * Fires a change event if the field's value is different since it got focus
     * @protected
     */
    this.fireChangeOnBlur = true;
  }
  onFocusOut(e) {
    if (this.validateOnBlur) {
      if (!e.relatedTarget || e.relatedTarget.tagName != "BUTTON") {
        this.validate();
      }
    }
    if (this.fireChangeOnBlur && this.isDirty()) {
      this.fireChange();
    }
  }
  /**
   * Return true if the field was modified
   */
  isDirty() {
    const v = this.value;
    if (typeof v == "object") {
      return JSON.stringify(this.oldValue) != JSON.stringify(v);
    } else {
      return this.oldValue != v;
    }
  }
  onFocusIn(e) {
    this.captureValueForChange();
  }
  captureValueForChange() {
    const v = this.value;
    this.oldValue = typeof v == "object" ? structuredClone(v) : v;
  }
  internalRender() {
    const el = super.internalRender();
    this.renderControl();
    this.el.addEventListener("focusin", this.onFocusIn.bind(this));
    this.el.addEventListener("focusout", this.onFocusOut.bind(this));
    return el;
  }
  isFocusable() {
    return !this.hidden;
  }
  get el() {
    const el = super.el;
    return el;
  }
  /**
   * A wrapper DIV element that contains input and toolbar for input buttons like an expand button for a drop down
   */
  get wrap() {
    if (!this._wrap) {
      this.el.append(this._wrap = E("div").cls("+wrap"));
    }
    return this._wrap;
  }
  renderControl() {
    const control = this.createControl();
    if (control) {
      this.wrap.append(control.cls("+control"));
    }
    const label = this.createLabel();
    if (label) {
      this.wrap.append(label);
    }
    this.renderButtons();
    const hint = this.createHint();
    if (hint) {
      this.el.appendChild(hint);
    }
  }
  renderButtons() {
    if (this._buttons) {
      this.toolbar = tbar({}, ...this._buttons);
      this.toolbar.parent = this;
      this.toolbar.render(this.wrap);
      this._buttons.forEach((btn2) => {
        if (btn2.menu) {
          this.setupMenu(btn2.menu);
        }
      });
    } else {
      if (this.toolbar) {
        this.toolbar.remove();
      }
    }
  }
  /**
   * When buttons with menus are added it is handy to delay the validation on blur.
   * Because when the user will click something in the menu it will blur the field and you probably don't
   * want it to validate at that point. It's important that the menu will return focus to the field
   * and sets the value afterward.
   *
   * @param menu
   * @protected
   */
  setupMenu(menu2) {
    let origValidateOnBlur = false;
    menu2.on("beforeshow", () => {
      origValidateOnBlur = this.validateOnBlur;
      this.validateOnBlur = false;
    });
    menu2.on("hide", () => {
      this.validateOnBlur = origValidateOnBlur;
      this.focus();
    });
  }
  createControl() {
    return void 0;
  }
  /**
   * Render buttons inside the text field
   *
   * @example
   * ```
   * buttons: [
   * 				 		btn({icon: "clear", handler:(btn) => (btn.parent!.parent! as Field).value = ""})
   * 					]
   * ```
   * @param buttons
   */
  set buttons(buttons) {
    this._buttons = buttons;
    if (this.rendered) {
      this.renderButtons();
    }
  }
  get buttons() {
    return this._buttons;
  }
  createHint() {
    this.hintEl = E("div", this._hint).cls("hint");
    return this.hintEl;
  }
  createLabel() {
    this._labelEl = E("div", this.getLabelText()).cls("label");
    return this._labelEl;
  }
  getLabelText() {
    let labelText = this._label;
    if (this._required) {
      labelText += " *";
    }
    return labelText;
  }
  /**
   * Form element name which will be the key in values
   * If omitted the field won't be included in the form values.
   */
  get name() {
    return this._name || "";
  }
  /**
   * The field's name
   */
  set name(name) {
    this._name = name;
  }
  get required() {
    return this._required;
  }
  /**
   * Required or not
   */
  set required(required) {
    this._required = required;
    if (this._labelEl) {
      this._labelEl.innerHTML = this.getLabelText();
    }
    if (this.rendered) {
      this.clearInvalid();
    }
  }
  get label() {
    return this._label + "";
  }
  /**
   * The field's label
   */
  set label(label) {
    this._label = label;
    if (this._labelEl) {
      this._labelEl.innerHTML = this.getLabelText();
    }
  }
  get icon() {
    return this._icon;
  }
  /**
   * The field's label
   */
  set icon(icon) {
    this._icon = icon;
    this.createIcon();
  }
  get hint() {
    return this._hint + "";
  }
  /**
   * The field's hint text
   */
  set hint(hint) {
    this._hint = hint;
  }
  get readOnly() {
    return this._readOnly;
  }
  /**
   * Make the field read only
   */
  set readOnly(readOnly) {
    this.el.classList.toggle("readonly", readOnly);
    this._readOnly = readOnly;
  }
  /**
   * Set the field value
   *
   * @param v
   * @param useForReset When true, this value is recorded for use when reset() is used.
   */
  set value(v) {
    const old = this._value;
    this._value = v;
    if (!this.resetValue) {
      this.resetValue = v;
    }
    this.internalSetValue(v, old);
    this.fire("setvalue", this, this._value, old);
  }
  internalSetValue(v, old) {
  }
  /**
   * Helper to fire "change" event. Child classes must implement this.
   *
   * @param suppressSetValue If another helper component like a date picker just called "set value" then you can
   *  set this to true to prevent firing twice.
   */
  fireChange() {
    const v = this.value;
    console.warn("change", this, v, this.oldValue);
    this.fire("change", this, v, this.oldValue);
    this.captureValueForChange();
  }
  get value() {
    return this._value;
  }
  /**
   * Reset the field value to the initial value
   * The initial value is set by setValue() when "useForReset" is set to true.
   * @see setValue()
   */
  reset() {
    const old = this.value;
    this.value = this.resetValue;
    this.clearInvalid();
    this.fire("reset", this, this.resetValue, old);
  }
  /**
   * Set the field as invalid and set a message
   *
   * @param msg
   */
  setInvalid(msg) {
    this.invalidMsg = msg;
    if (this.rendered) {
      this.applyInvalidMsg();
    }
    this.fire("invalid", this);
  }
  /**
   * Check if the field is empty
   */
  isEmpty() {
    const v = this.value;
    return v == void 0 || v == null || v == "";
  }
  validate() {
    this.clearInvalid();
    if (this._required && this.isEmpty()) {
      this.setInvalid(t("This field is required"));
    }
    this.fire("validate", this);
  }
  /*
  
  			badInput
  :
  false
  customError
  :
  false
  patternMismatch
  :
  false
  rangeOverflow
  :
  false
  rangeUnderflow
  :
  false
  stepMismatch
  :
  false
  tooLong
  :
  false
  tooShort
  :
  false
  typeMismatch
  :
  false
  
  			 */
  setValidityState(input) {
    const validity = input.validity;
    if (validity.valid) {
      return;
    }
    if (validity.typeMismatch) {
      switch (input.type) {
        case "email":
          this.setInvalid(t("Please enter a valid e-mail address"));
          return;
        default:
          this.setInvalid(t("Please enter a valid value"));
          return;
      }
    } else if (validity.valueMissing) {
      this.setInvalid(t("This field is required"));
    } else if (validity.tooLong) {
      this.setInvalid(t("The maximum length for this field is {max}").replace("{max}", input.maxLength));
    } else if (validity.tooShort) {
      this.setInvalid(t("The minimum length for this field is {max}").replace("{max}", input.minLength));
    } else if (validity.patternMismatch) {
      this.setInvalid(t("Please match the format requested").replace("{max}", input.minLength));
    } else {
      console.warn("TODO: Implement translated message");
      this.setInvalid(input.validationMessage);
    }
  }
  applyInvalidMsg() {
    if (this.invalidMsg) {
      this.el.classList.add("invalid");
      if (this.hintEl)
        this.hintEl.innerHTML = this.invalidMsg;
    } else {
      this.el.classList.remove("invalid");
      if (this.hintEl)
        this.hintEl.innerHTML = this._hint || "";
    }
  }
  /**
   * Clears the invalid state
   */
  clearInvalid() {
    if (!this.invalidMsg) {
      return;
    }
    this.invalidMsg = "";
    if (this.rendered) {
      this.applyInvalidMsg();
    }
  }
  /**
   * Checks if the field is valid
   */
  isValid() {
    if (this.invalidMsg != "") {
      return false;
    }
    this.validate();
    if (this.invalidMsg != "") {
      console.warn("Field '" + this.name + "' is invalid: " + this.invalidMsg, this);
    }
    return this.invalidMsg == "";
  }
  createIcon() {
    if (this._icon) {
      if (!this.iconEl) {
        this.iconEl = E("i").cls("icon");
      }
      this.iconEl.innerText = this._icon;
      this.el.classList.add("with-icon");
      if (this.wrap) {
        this.wrap.insertBefore(this.iconEl, this.wrap.firstChild);
      }
      return this.iconEl;
    } else {
      if (this.iconEl) {
        this.iconEl.remove();
        this.el.classList.remove("with-icon");
      }
    }
  }
};

// goui/script/component/form/ContainerField.ts
var ContainerField = class extends Field {
  constructor(tagName = "div") {
    super(tagName);
    this.baseCls = "";
    this.hideLabel = true;
    /**
     * Not needed on container field as the fields within handle this.
     */
    this.validateOnBlur = false;
  }
  /**
   * Find all form fields
   *
   * @param nameOrItemId If given only items with matching name or itemId are returned
   */
  findFields(nameOrItemId = void 0) {
    const fields = [];
    const fn = (item) => {
      if (item == this) {
        return;
      }
      if (item.isFormField) {
        if (!nameOrItemId || (item.name == nameOrItemId || item.itemId == nameOrItemId)) {
          fields.push(item);
        }
        return false;
      }
    };
    this.cascade(fn);
    return fields;
  }
  /**
   * Find form field by name or item ID
   *
   * It cascades down the component hierarchy.
   *
   * @param nameOrItemId
   */
  findField(nameOrItemId) {
    let field;
    const fn = (item) => {
      if (item.isFormField && item.name == nameOrItemId || item.itemId == nameOrItemId) {
        field = item;
        return false;
      }
    };
    this.cascade(fn);
    return field;
  }
  set value(v) {
    for (let name in v) {
      let fields = this.findFields(name);
      fields.forEach((field) => field.setValue ? field.setValue(v[name]) : field.value = v[name]);
    }
    super.value = v;
  }
  get value() {
    const formProps = super.value || {};
    this.findFields().forEach((field) => {
      const fieldName = field.getName ? field.getName() : field.name;
      const fieldVal = field.getValue ? field.getValue() : field.value;
      if (fieldName && !field.disabled) {
        if (!formProps[fieldName]) {
          formProps[fieldName] = fieldVal;
        } else {
          formProps[fieldName] = ObjectUtil.merge(formProps[field.name], fieldVal);
        }
      }
    });
    return formProps;
  }
  validate() {
    super.validate();
    let invalid;
    this.findFields().forEach((i) => {
      if (!i.disabled && !i.isValid()) {
        invalid = i;
      }
    });
    if (invalid) {
      this.setInvalid("There's an invalid field");
    }
  }
  /**
   * Find the first invalid field
   */
  findFirstInvalid() {
    const items = this.findFields();
    for (let i = 0, l = items.length; i < l; i++) {
      if (!items[i].disabled && !items[i].isValid()) {
        return items[i];
      }
    }
    return void 0;
  }
  renderControl() {
  }
  clearInvalid() {
    super.clearInvalid();
    const items = this.findFields();
    items.forEach((field) => {
      field.clearInvalid();
    });
  }
  applyInvalidMsg() {
    if (this.invalidMsg) {
      this.el.classList.add("invalid");
    } else {
      this.el.classList.remove("invalid");
    }
  }
  /**
   * @inheritDoc
   */
  focus(o) {
    const fields = this.findFields();
    if (fields.length) {
      fields[0].focus(o);
      this.fire("focus", this, o);
    } else {
      super.focus(o);
    }
  }
};
var containerfield = (config, ...items) => createComponent(new ContainerField(), config, items);

// goui/script/component/Root.ts
var Root = class extends Component {
  constructor() {
    super();
    this.items.on("beforeadd", () => {
      this.el.classList.add("goui");
      this.el.classList.add("root");
    }, {
      once: true
    });
  }
  internalRender() {
    this.renderItems();
    return this.el;
  }
  get el() {
    if (!this._rootEl) {
      this._rootEl = document.getElementById("goui") || document.body;
    }
    return this._rootEl;
  }
  get rendered() {
    return true;
  }
};
var root = new Root();

// goui/script/component/menu/Menu.ts
var Menu = class _Menu extends Toolbar {
  constructor() {
    super();
    /**
     * Make the menu at least as wide as the component it aligns too.
     */
    this.alignToInheritWidth = false;
    /**
     * @inheritDoc
     */
    this.renderTo = root.el;
    /**
     * Remove menu when closed
     */
    this.removeOnClose = true;
    this.baseCls = "";
    this.orientation = "vertical";
  }
  internalRender() {
    const el = super.internalRender();
    this.el.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "Escape":
          this.close();
          ev.stopPropagation();
          ev.preventDefault();
          break;
      }
    });
    const onClose = () => {
      if (_Menu.openedMenu == this) {
        _Menu.openedMenu = void 0;
      }
    };
    this.on("hide", onClose);
    this.on("remove", onClose);
    return el;
  }
  /**
   * Menu can be rendered as a component in the normal flow or as a
   * floating dropdown.
   *
   * @param value
   */
  set isDropdown(value) {
    this.el.classList.toggle("goui-dropdown", value);
    this.el.classList.toggle("goui-fade-out", value);
  }
  get isDropdown() {
    return this.el.classList.contains("goui-dropdown");
  }
  /**
   * Expand menu on the left side of the parent button
   * @param expandLeft
   */
  set expandLeft(expandLeft) {
    this.el.classList.add("expand-left");
  }
  get expandLeft() {
    return this.el.classList.contains("expand-left");
  }
  renderItem(item) {
    const insertBefore = this.getInsertBefore();
    if (!insertBefore) {
      this.el.appendChild(this.wrapLI(item));
    } else {
      this.el.insertBefore(this.wrapLI(item), insertBefore);
    }
  }
  wrapLI(item) {
    const li = document.createElement("li");
    item.render(li);
    if (item instanceof Button && item.menu) {
      item.menu.render(li);
    }
    return li;
  }
  /**
   * Show aligned to the given component.
   *
   * It will align the top left of the menu top the bottom left of the component. It will also be at least as wide as
   * the given component by setting the min-width style.
   *
   * @todo avoid going out of the viewport
   * @param cmp
   */
  showFor(alignEl) {
    this.alignTo = alignEl;
    this.show();
  }
  setAlignTo() {
    if (!this.alignTo) {
      return;
    }
    const rect = this.alignTo.getBoundingClientRect();
    this.x = this.expandLeft ? rect.right - this.width : rect.x;
    this.y = rect.bottom;
    if (this.alignToInheritWidth) {
      this.el.style.minWidth = rect.width + "px";
    }
  }
  /**
   * Set X coordinate
   *
   * @param x
   */
  set x(x) {
    this.el.style.left = x + "px";
  }
  /**
   * Set Y coordinate
   *
   * @param y
   */
  set y(y) {
    this.el.style.top = y + "px";
  }
  show() {
    if (!this.parent) {
      root.items.add(this);
    }
    if (!this.rendered) {
      this.render();
    }
    this.setAlignTo();
    if (_Menu.openedMenu == this) {
      console.warn("Already open");
      return true;
    }
    if (_Menu.openedMenu) {
      _Menu.openedMenu.el.classList.remove("goui-fade-out");
      _Menu.openedMenu.close();
    }
    _Menu.openedMenu = this;
    window.addEventListener("mousedown", (ev) => {
      this.close();
    }, { once: true });
    this.el.addEventListener("mousedown", (ev) => {
      ev.stopPropagation();
    });
    this.el.classList.add("goui-fade-out");
    return super.show();
  }
  /**
   * Show menu at coordinates on the page.
   *
   * Useful for a context menu
   *
   * @param coords
   */
  showAt(coords) {
    this.x = coords.x;
    this.y = coords.y;
    this.show();
  }
  /**
   * Closes the menu.
   *
   * It will hide or remove it depending on the "removeOnClose" property.
   */
  close() {
    return this.removeOnClose ? this.remove() : this.hide();
  }
  /**
   * @inheritDoc
   */
  focus(o) {
    var _a;
    (_a = this.items.get(0)) == null ? void 0 : _a.focus(o);
  }
};
var menu = (config, ...items) => createComponent(new Menu(), config, items);

// goui/script/Router.ts
var Router = class extends Observable {
  constructor() {
    super();
    this.routes = [];
    this.suspendEvent = false;
    this.loadedPath = "";
    this.debug = false;
    this.params = [];
    window.addEventListener("hashchange", () => {
      this.start();
    }, false);
  }
  /**
   * Get the router path
   */
  getPath() {
    return window.location.hash.substr(1);
  }
  /**
   * Get the parameters evaluated from the router path
   */
  getParams() {
    return this.params;
  }
  /**
   * Set route path without executing matching routes.
   *
   * @param path
   */
  setPath(path) {
    if (path != window.location.hash) {
      this.suspendEvent = true;
      const oldPath = this.getPath();
      window.location.hash = path;
      this.fire("change", this.getPath(), oldPath);
    }
  }
  /**
   * Add a route
   *
   * The first matching route will be executed
   *
   * @example
   * ```
   * go.Router.add(/^([a-zA-Z0-9]*)\/([\d]*)$/, (entity:string, id:string) => {
   *
   * });
   * ```
   *
   * @param re eg. /^notes/(.*)$/
   * @param handler Is called with the arguments matched in the route regexp. May return Promise so the router start()
   *  promise will resolve when this promise is resolved.
   */
  add(re, handler) {
    if (this.debug) {
      console.debug("Router add: ", re);
    }
    if (typeof re == "function") {
      handler = re;
      this.defaultRoute = handler;
      return this;
    }
    const route = { re, handler };
    this.routes.push(route);
    return this;
  }
  /**
   * Start the router and run the matching route handlers
   */
  start() {
    const path = this.getPath();
    const oldPath = this.loadedPath;
    if (this.suspendEvent) {
      setTimeout(() => {
        this.suspendEvent = false;
      });
      return Promise.resolve();
    }
    this.loadedPath = path;
    for (let i = 0; i < this.routes.length; i++) {
      const args = path.match(this.routes[i].re);
      if (args) {
        if (this.debug) {
          console.debug("Router match: ", this.routes[i].re);
        }
        args.shift();
        return this.handleRoute(this.routes[i].handler, args, oldPath);
      }
    }
    return this.defaultRoute ? this.handleRoute(this.defaultRoute, [], oldPath) : Promise.resolve();
  }
  handleRoute(handler, match, oldPath) {
    for (let n = 0, l = match.length; n < l; n++) {
      match[n] = match[n] ? decodeURIComponent(match[n]) : match[n];
    }
    this.params = match;
    const result = handler.apply({}, match);
    this.fire("change", this.getPath(), oldPath);
    return result instanceof Promise ? result : Promise.resolve();
  }
  /**
   * Reload current page.
   */
  reload() {
    this.start();
  }
  /**
   * Go to the give router path
   *
   * @param path
   * @return Promise<Router>
   */
  goto(path) {
    const p2 = new Promise((resolve, reject) => {
      this.on("change", (path1, oldPath) => {
        resolve(this);
      }, { once: true });
    });
    window.location.hash = path || "";
    return p2;
  }
};
var router = new Router();

// goui/script/component/Button.ts
var Button = class _Button extends Component {
  constructor() {
    super("button");
    this.baseCls = "goui-button";
    this.type = "button";
  }
  /**
   * Find the first menu in the tree of submenu's
   */
  findTopMenu() {
    if (!(this.parent instanceof Menu)) {
      return void 0;
    }
    if (!(this.parent.parent instanceof _Button)) {
      return this.parent;
    } else {
      const next = this.parent.parent.findTopMenu();
      if (next) {
        return next;
      } else {
        return this.parent;
      }
    }
  }
  /**
   * Button type. "button" or "submit", defaults to "button".
   */
  set type(type) {
    this.el.type = type;
  }
  get type() {
    return this.el.type;
  }
  internalRender() {
    const el = super.internalRender();
    if (this.route != void 0) {
      if (this.handler) {
        throw "You can't set both handler and route on a button";
      }
      this.handler = () => {
        router.goto(this.route);
      };
    }
    if (this.menu) {
      this.menu.hide();
      if (!(this.parent instanceof Menu)) {
        this.el.addEventListener("mouseenter", this.onMenuMouseEnter.bind(this));
        this.el.addEventListener("click", this.onMenuButtonClick.bind(this));
      } else {
        this.menu.renderTo = void 0;
      }
    }
    el.addEventListener("click", (e) => {
      if (this.handler && e.button == 0 && e.detail < 2) {
        this.handler.call(this, this, e);
        const topMenu = this.findTopMenu();
        if (topMenu && topMenu.isDropdown) {
          topMenu.close();
        }
      }
      this.fire("click", this, e);
    });
    return el;
  }
  onMenuButtonClick(ev) {
    if (this._menu.hidden) {
      this.showMenu();
    } else {
      this._menu.hide();
    }
  }
  onMenuMouseEnter(ev) {
    if (Menu.openedMenu && Menu.openedMenu != this._menu && Menu.openedMenu.parent.parent === this._menu.parent.parent) {
      this.showMenu();
    }
  }
  /**
   * Add menu to this button
   */
  set menu(menu2) {
    if (menu2) {
      menu2.parent = this;
      menu2.removeOnClose = false;
      menu2.isDropdown = true;
      this.el.classList.add("has-menu");
    }
    this._menu = menu2;
  }
  get menu() {
    return this._menu;
  }
  showMenu() {
    if (!this._menu) {
      return;
    }
    if (!this._menu.hidden) {
      return;
    }
    if (this.fire("beforeshowmenu", this, this._menu) === false) {
      return;
    }
    if (!this._menu.alignTo) {
      this._menu.alignTo = this.el;
    }
    this._menu.show();
    this.fire("showmenu", this, this._menu);
  }
  internalRemove() {
    if (this.menu) {
      this.menu.remove();
    }
    super.internalRemove();
  }
  /**
   * Button text
   *
   * Set's the button text and adds a "text" css class
   *
   * This overrides the "html" property. Use html if you want something custom.
   */
  set text(text) {
    if (text) {
      this._text = text;
      this.el.classList.add("with-text");
    } else {
      this.el.classList.remove("with-text");
    }
    this.textEl.innerText = text + "";
  }
  get text() {
    return this._text + "";
  }
  /**
   * Set's the button icon and adds a "icon" css class
   */
  set icon(icon) {
    this._icon = icon;
    if (this._icon != void 0) {
      this.el.classList.add("with-icon");
    } else {
      this.el.classList.remove("with-icon");
    }
    this.iconEl.innerText = icon + "";
  }
  get icon() {
    return this._icon;
  }
  get iconEl() {
    if (!this._iconEl) {
      this._iconEl = document.createElement("i");
      this._iconEl.classList.add("icon");
      if (this._textEl) {
        this.el.insertBefore(this._iconEl, this._textEl);
      } else {
        this.el.appendChild(this._iconEl);
      }
    }
    return this._iconEl;
  }
  get textEl() {
    if (!this._textEl) {
      this._textEl = document.createElement("span");
      this._textEl.classList.add("text");
      this.el.appendChild(this._textEl);
    }
    return this._textEl;
  }
};
var btn = (config) => createComponent(new Button(), config);

// goui/script/component/form/ArrayField.ts
var ArrayField = class extends ContainerField {
  /**
   *
   * @param buildField Function that returns a new form field for an array item
   */
  constructor(buildField) {
    super("div");
    this.buildField = buildField;
  }
  set value(v) {
    super.value = v;
    this.items.clear();
    if (v) {
      v.forEach((item) => {
        this.internalAddValue(item);
      });
    }
  }
  get value() {
    const v = [];
    this.items.forEach((item) => {
      if (item instanceof Field) {
        v.push(item.value);
      }
    });
    return v;
  }
  /**
   * Add value to the values array. Also fires change event
   *
   * @param value
   */
  addValue(value = {}) {
    if (!this.oldValue) {
      this.captureValueForChange();
    }
    this.internalAddValue(value);
    this.fireChange();
    return this;
  }
  internalAddValue(value) {
    const field = this.buildField(value);
    field.value = value;
    field.items.add(
      btn({
        // style: {
        // 	alignSelf: "center"
        // },
        icon: "delete",
        title: "Delete",
        handler: (btn2) => {
          btn2.parent.remove();
          this.fireChange();
        }
      })
    );
    this.items.add(field);
  }
  reset() {
    super.reset();
    this.items.clear();
  }
};
var arrayfield = (config, ...items) => createComponent(new ArrayField(config.buildField), config, items);

// goui/script/component/form/TextField.ts
var TextField = class extends Field {
  constructor() {
    super(...arguments);
    /**
     * input type. text, password, email etc.
     */
    this.type = "text";
    /**
     * When the field is empty this will be displayed inside the field
     */
    this.placeholder = " ";
  }
  get input() {
    return this._input;
  }
  set title(title) {
    super.title = title;
    if (this._input) {
      this._input.title = this.title;
    }
  }
  focus(o) {
    var _a;
    if (!this._input) {
      super.focus(o);
    }
    (_a = this._input) == null ? void 0 : _a.focus(o);
  }
  createControl() {
    const v = this.value, name = this.name;
    this._input = document.createElement("input");
    this._input.classList.add("text");
    this._input.type = this.type;
    if (this.pattern) {
      this._input.pattern = this.pattern;
    }
    if (this.autocomplete) {
      this._input.autocomplete = this.autocomplete;
    }
    if (this.placeholder) {
      this._input.placeholder = this.placeholder;
    }
    this._input.required = this.required;
    if (name) {
      this._input.name = name;
    }
    this._input.readOnly = this.readOnly;
    if (v !== void 0) {
      this._input.value = v;
    }
    if (this.title) {
      this._input.title = this.title;
    }
    if (this.invalidMsg) {
      this.applyInvalidMsg();
    }
    return this._input;
  }
  setInvalid(msg) {
    super.setInvalid(msg);
    if (this.rendered) {
      this.applyInvalidMsg();
    }
  }
  clearInvalid() {
    super.clearInvalid();
    this.applyInvalidMsg();
  }
  setInputValue(v) {
    if (this._input) {
      this._input.value = v + "";
    }
  }
  set value(v) {
    this.setInputValue(v || "");
    super.value = v;
  }
  get value() {
    if (!this._input) {
      return super.value;
    } else {
      return this._input.value;
    }
  }
  set name(name) {
    super.name = name;
    if (this._input) {
      this._input.name = this.name;
    }
  }
  get name() {
    return super.name;
  }
  set readOnly(readOnly) {
    super.readOnly = readOnly;
    if (this._input) {
      this._input.readOnly = this.readOnly;
    }
  }
  get readOnly() {
    return super.readOnly;
  }
  validate() {
    super.validate();
    if (this._input) {
      this.setValidityState(this._input);
    }
  }
};
var textfield = (config) => createComponent(new TextField(), config);

// goui/script/component/picker/ColorPicker.ts
var ColorPicker = class _ColorPicker extends Component {
  constructor() {
    super();
    this._value = "";
    /**
     * Set color value as button text color if this menu belongs to a button
     */
    this.updateButton = true;
    this.baseCls = "goui-dropdown goui-menu-color";
    this.colors = [
      // '000000', //Black
      // 'FFFFFF', //White
      // '009BC9', //Group-Office blue
      // //'243A80', //Intermesh blue
      // '78A22F', //default secondary
      // 'FF9100',  //Default accent
      "B71C1C",
      "C62828",
      "D32F2F",
      "E53935",
      "F44336",
      // Red
      "880E4F",
      "AD1457",
      "C2185B",
      "D81B60",
      "E91E63",
      // Pink
      "4A148C",
      "6A1B9A",
      "7B1FA2",
      "8E24AA",
      "9C27B0",
      // Purple
      "1A237E",
      "283593",
      "303F9F",
      "3949AB",
      "3F51B5",
      // Indigo
      "0D47A1",
      "1565C0",
      "1976D2",
      "1E88E5",
      "2196F3",
      // Blue
      "006064",
      "00838F",
      "0097A7",
      "00ACC1",
      "00BCD4",
      // Cyan
      "004D40",
      "00695C",
      "00796B",
      "00897B",
      "009688",
      // Teal
      "1B5E20",
      "2E7D32",
      "388E3C",
      "43A047",
      "4CAF50",
      // Green
      "827717",
      "9E9D24",
      "AFB42B",
      "C0CA33",
      "CDDC39",
      // Lime
      //'F57F17', 'F9A825', 'FBC02D', 'FDD835', 'FFEB3B', // Yellow
      "FF6F00",
      "FF8F00",
      "FFA000",
      "FFB300",
      "FFC107",
      // Amber
      "E65100",
      "EF6C00",
      "F57C00",
      "FB8C00",
      "FF9800",
      // Orange
      "212121",
      "424242",
      "616161",
      "757575",
      "BDBDBD"
      // Grey
    ];
    this.items.add(
      comp(
        {},
        ...this.colors.map((color) => btn({
          itemId: color,
          cls: this.value == "#" + color ? "with-icon pressed" : "with-icon",
          listeners: {
            beforerender: (btn2) => {
              const colorDiv = document.createElement("div");
              colorDiv.style.backgroundColor = "#" + color;
              btn2.el.appendChild(colorDiv);
            }
          },
          handler: (btn2) => {
            this.value = btn2.itemId;
            this.fire("select", this, btn2.itemId);
          }
        }))
      )
    );
  }
  //picker interfae
  getText() {
    return this.value.substring(1);
  }
  setValue(val) {
    this.value = val;
  }
  static rgb2hex(str) {
    const rgb = str.match(/^rgb\((\d+),\s*(\d+),\s*(\d+)\)$/i);
    if (!rgb) {
      return "";
    }
    return "#" + ((1 << 24) + (parseInt(rgb[1]) << 16) + (parseInt(rgb[2]) << 8) + parseInt(rgb[3])).toString(16).slice(1);
  }
  /**
   * Color hex value eg. #000000
   */
  set value(color) {
    const hex = _ColorPicker.rgb2hex(color);
    if (hex) {
      color = hex.toUpperCase();
    } else {
      color = color.toUpperCase();
    }
    this._value = color;
    if (!this.rendered) {
      return;
    }
    if (color == "") {
      color = "auto";
    }
    this.items.forEach((btn2) => {
      btn2.el.classList.toggle("pressed", btn2.itemId == color);
    });
  }
  get value() {
    return this._value;
  }
};
var colorpicker = (config) => createComponent(new ColorPicker(), config);

// goui/script/component/picker/DatePicker.ts
var DatePicker = class extends Component {
  constructor() {
    super();
    this.enableRangeSelect = false;
    this.baseCls = "datepicker";
    this.showWeekNbs = true;
    this.value = new DateTime();
    this.now = new DateTime();
    this.el.cls("+preload");
    this.el.append(
      this.menu = E(
        "header",
        E("button", E("i", "chevron_left").cls("icon")).cls(["goui-button", "nav"], true).on("click", (_) => {
          this.moveMonth(-1);
        }),
        this.monthEl = E("button").cls("goui-button").on("click", (_) => {
          this.months.cls("!active");
          this.years.cls("-active");
        }),
        this.yearEl = E("button").cls("goui-button").on("click", (_) => {
          this.years.cls("!active");
          this.months.cls("-active");
          this.years.scrollTop = this.years.scrollHeight / 2 - this.years.clientHeight / 2;
        }),
        E("button", E("i", "chevron_right").cls("icon")).cls(["goui-button", "nav"], true).on("click", (_) => {
          this.moveMonth(1);
        })
      ),
      this.grid = E("div").cls("+cards"),
      E(
        "div",
        this.years = E("ul").on("click", ({ target }) => {
          const year = +target.textContent;
          if (!isNaN(year)) {
            this.value.setYear(year);
            this.internalRender();
            this.years.cls("-active");
          }
        }),
        this.months = E("ul").on("click", ({ target }) => {
          if (target.dataset.nb) {
            this.value.setMonth(+target.dataset.nb);
            this.internalRender();
            this.months.cls("-active");
          }
        })
      ).cls(["cards", "top"], true)
    );
    for (let i = -100; i < 100; i++) {
      this.years.append(E("li", this.value.getYear() + i + "").cls("now", i === 0));
    }
    this.months.replaceChildren(...DateTime.monthNames.map(
      (name, i) => E("li", name).cls("now", i === this.now.getMonth() - 1).attr("data-nb", i + 1)
    ));
    this.internalRender();
  }
  moveMonth(amount) {
    this.grid.cls("reverse", amount > 0);
    this.grid.offsetHeight;
    this.value.addMonths(amount);
    this.internalRender();
  }
  internalRender() {
    var _a, _b, _c, _d, _e;
    const el = super.internalRender();
    this.monthEl.textContent = this.value.format("M");
    this.yearEl.textContent = this.value.format("Y");
    (_a = this.years.querySelector(".selected")) == null ? void 0 : _a.cls("-selected");
    let y = this.value.getYear() - this.now.getYear() + 100;
    (_b = this.years.children[y]) == null ? void 0 : _b.cls("+selected");
    (_c = this.months.querySelector(".selected")) == null ? void 0 : _c.cls("-selected");
    (_d = this.months.querySelector('[data-nb="' + this.value.getMonth() + '"]')) == null ? void 0 : _d.cls("+selected");
    const cal = E("div").cls("active");
    const dl2 = E("dl"), itr = this.value.clone().setMonthDay(1).setWeekDay(0), weekNbs = E("ol");
    dl2.append(...Object.values(DateTime.dayNames).map((s) => E("dt", s.substring(0, 2))));
    for (let i = 0; i < 42; i++) {
      if (this.showWeekNbs && i % 7 == 0) {
        weekNbs.append(E("li", itr.format("W")));
      }
      let disabled = false;
      if (this.minDate && this.minDate.getTime() > itr.getTime()) {
        disabled = true;
      }
      if (this.maxDate && this.maxDate.getTime() < itr.getTime()) {
        disabled = true;
      }
      dl2.append(
        E("dd", itr.format("j")).attr("data-date", itr.format("Y-m-d")).cls("disabled", disabled).cls("off", itr.format("Ym") !== this.value.format("Ym")).cls("today", itr.format(`Ymd`) === this.now.format(`Ymd`))
      );
      itr.addDays(1);
    }
    if (this.showWeekNbs) {
      cal.append(weekNbs);
    }
    this.setupDraggability(dl2);
    cal.append(dl2);
    const old = (_e = this.grid.firstElementChild) == null ? void 0 : _e.cls("-active");
    this.grid.prepend(cal);
    setTimeout(function() {
      el.cls("-preload");
      old == null ? void 0 : old.remove();
    }, 250);
    return el;
  }
  setValue(start, end) {
    if (start.format("Ym") != this.value.format("Ym")) {
      this.value = start;
      this.internalRender();
    }
    const startEl = this.grid.querySelector('dd[data-date="' + start.format("Y-m-d") + '"]'), endEl = !end ? startEl : this.grid.querySelector('dd[data-date="' + end.format("Y-m-d") + '"]');
    if (startEl && endEl) {
      this.markSelected(startEl, endEl);
    }
  }
  markSelected(start, end) {
    this.el.querySelectorAll(".selected").forEach((e) => e.cls(["selected", "tail", "head"], false));
    end.cls("+tail");
    start.cls("+head");
    let itr = start;
    while (itr && itr != end.nextElementSibling) {
      itr.cls("+selected");
      itr = itr.nextElementSibling;
    }
  }
  setupDraggability(dl2) {
    let anchor, start, end;
    const onMouseMove = ({ target }) => {
      if (!this.enableRangeSelect) {
        return false;
      }
      if (target.isA("DD") && target != end) {
        [start, end] = anchor.compareDocumentPosition(target) & 2 ? [target, anchor] : [anchor, target];
        this.markSelected(start, end);
      }
    }, onMouseUp = (_e) => {
      dl2.un("mousemove", onMouseMove);
      window.removeEventListener("mouseup", onMouseUp);
      if (!this.enableRangeSelect || start == end) {
        this.fire("select", this, new DateTime(start.attr("data-date")));
      } else {
        this.fire("select-range", this, new DateTime(start.attr("data-date")), new DateTime(end.attr("data-date")));
      }
    };
    dl2.on("mousedown", ({ target }) => {
      if (target.isA("dd")) {
        anchor = start = end = target;
        dl2.querySelectorAll("dd.selected").forEach((e) => e.cls(["selected", "tail", "head"], false));
        target.cls(["selected", "tail", "head"], true);
        dl2.on("mousemove", onMouseMove);
        window.addEventListener("mouseup", onMouseUp);
      }
    });
  }
};
var datepicker = (config) => createComponent(new DatePicker(), config);

// goui/script/component/form/SelectField.ts
var SelectField = class extends Field {
  constructor() {
    super(...arguments);
    this.baseCls = "goui-form-field select";
    this.options = [];
    this.valueField = "value";
    this.textRenderer = (record) => record.name;
  }
  createControl() {
    const v = this.value;
    this.input = document.createElement("select");
    this.input.name = this.name;
    if (this.required) {
      this.input.setAttribute("required", "");
    }
    this.drawOptions();
    this.el.appendChild(this.input);
    return this.input;
  }
  getInput() {
    return this.input;
  }
  drawOptions() {
    if (!this.input)
      return;
    this.input.innerHTML = "";
    (this.store ? this.store.items : this.options).forEach((o) => {
      var _a;
      const opt = new Option();
      if (o[this.valueField]) {
        opt.value = o[this.valueField];
      }
      opt.innerHTML = this.textRenderer(o);
      (_a = this.input) == null ? void 0 : _a.appendChild(opt);
    });
    if (this._value) {
      this.value = this._value;
    }
  }
  set value(v) {
    if (this.input) {
      this.input.value = v;
    }
    super.value = v;
  }
  get value() {
    if (!this.input) {
      return super.value;
    } else {
      return this.input.value;
    }
  }
  set name(name) {
    super.name = name;
    if (this.input) {
      this.input.name = this.name;
    }
  }
  get name() {
    return super.name;
  }
  validate() {
    super.validate();
    if (this.input)
      this.setValidityState(this.input);
  }
};
var select = (config) => createComponent(new SelectField(), config);

// goui/script/component/menu/ColorMenu.ts
var ColorMenu = class _ColorMenu extends Menu {
  constructor() {
    super();
    this._value = "";
    /**
     * Set color value as button text color if this menu belongs to a button
     */
    this.updateButton = true;
    this.baseCls = "goui-dropdown goui-menu-color";
    this.colors = [
      "000000",
      //Black
      "FFFFFF",
      //White
      "009BC9",
      //Group-Office blue
      //'243A80', //Intermesh blue
      "78A22F",
      //default secondary
      "FF9100",
      //Default accent
      "B71C1C",
      "C62828",
      "D32F2F",
      "E53935",
      "F44336",
      // Red
      "880E4F",
      "AD1457",
      "C2185B",
      "D81B60",
      "E91E63",
      // Pink
      "4A148C",
      "6A1B9A",
      "7B1FA2",
      "8E24AA",
      "9C27B0",
      // Purple
      "311B92",
      "4527A0",
      "512DA8",
      "5E35B1",
      "673AB7",
      // Deep purple
      "1A237E",
      "283593",
      "303F9F",
      "3949AB",
      "3F51B5",
      // Indigo
      "0D47A1",
      "1565C0",
      "1976D2",
      "1E88E5",
      "2196F3",
      // Blue
      "01579B",
      "0277BD",
      "0288D1",
      "039BE5",
      "03A9F4",
      // Light blue
      "006064",
      "00838F",
      "0097A7",
      "00ACC1",
      "00BCD4",
      // Cyan
      "004D40",
      "00695C",
      "00796B",
      "00897B",
      "009688",
      // Teal
      "1B5E20",
      "2E7D32",
      "388E3C",
      "43A047",
      "4CAF50",
      // Green
      "33691E",
      "558B2F",
      "689F38",
      "7CB342",
      "8BC34A",
      // Light Green
      "827717",
      "9E9D24",
      "AFB42B",
      "C0CA33",
      "CDDC39",
      // Lime
      "F57F17",
      "F9A825",
      "FBC02D",
      "FDD835",
      "FFEB3B",
      // Yellow
      "FF6F00",
      "FF8F00",
      "FFA000",
      "FFB300",
      "FFC107",
      // Amber
      "E65100",
      "EF6C00",
      "F57C00",
      "FB8C00",
      "FF9800",
      // Orange
      "212121",
      "424242",
      "616161",
      "757575",
      "BDBDBD"
      // Grey
    ];
    this.items.add(btn({
      itemId: "auto",
      text: "Auto",
      cls: this.value == "" ? "pressed" : "",
      handler: () => {
        this.value = "";
        this.fire("select", this, "");
      }
    }));
    this.colors.forEach((color) => {
      this.items.add(btn({
        itemId: "#" + color,
        cls: this.value == color ? "with-icon pressed" : "with-icon",
        listeners: {
          beforerender: (btn2) => {
            const colorDiv = document.createElement("div");
            colorDiv.style.backgroundColor = "#" + color;
            btn2.el.appendChild(colorDiv);
          }
        },
        handler: (btn2) => {
          this.value = btn2.itemId;
          this.fire("select", this, btn2.itemId);
        }
      }));
    });
  }
  static rgb2hex(str) {
    const rgb = str.match(/^rgb\((\d+),\s*(\d+),\s*(\d+)\)$/i);
    if (!rgb) {
      return "";
    }
    return "#" + ((1 << 24) + (parseInt(rgb[1]) << 16) + (parseInt(rgb[2]) << 8) + parseInt(rgb[3])).toString(16).slice(1);
  }
  /**
   * Color hex value eg. #000000
   */
  set value(color) {
    const hex = _ColorMenu.rgb2hex(color);
    if (hex) {
      color = hex.toUpperCase();
    } else {
      color = color.toUpperCase();
    }
    this._value = color;
    if (this.updateButton) {
      this.parent.el.style.color = color;
    }
    if (!this.rendered) {
      return;
    }
    if (color == "") {
      color = "auto";
    }
    this.items.forEach((btn2) => {
      btn2.el.classList.toggle("pressed", btn2.itemId == color);
    });
  }
  get value() {
    return this._value;
  }
};
var colormenu = (config) => createComponent(new ColorMenu(), config);

// goui/script/component/form/DateField.ts
var DateField = class extends TextField {
  constructor() {
    super();
    this.inputFormat = "d-m-Y";
    this.outputFormat = "Y-m-d";
    let origValidateOnBlur = true;
    this.picker = datepicker({ showWeekNbs: false });
    this.buttons = [
      this.pickerButton = btn({
        icon: "calendar_today",
        menu: menu(
          {
            alignTo: this.wrap,
            alignToInheritWidth: false
          },
          this.picker
        )
      })
    ];
    this.pattern = DateTime.createFormatRegex(this.inputFormat);
    this.title = "Incorrect date format";
  }
  createControl() {
    const input = super.createControl();
    this.picker.on("select", (_, val) => {
      this.date = val;
      this.pickerButton.menu.hide();
      this.clearInvalid();
      super.value = val.format(this.inputFormat);
    });
    this.pickerButton.menu.on("show", () => {
      let dt = new DateTime();
      if (this.minDate) {
        this.picker.minDate = this.minDate;
        if (this.minDate.getTime() > dt.getTime()) {
          dt = this.minDate;
          dt.setHours(0);
          dt.setMinutes(0);
        }
      }
      if (this.maxDate) {
        this.picker.maxDate = this.maxDate;
        if (this.maxDate.getTime() < dt.getTime()) {
          dt = this.maxDate;
          dt.setHours(23);
          dt.setMinutes(59);
        }
      }
      this.picker.setValue(this.getValueAsDateTime() || dt);
    });
    return input;
  }
  validate() {
    super.validate();
    const v = super.value;
    if (!v) {
      return;
    }
    const dv = DateTime.createFromFormat(v, this.inputFormat);
    if (!dv) {
      this.setInvalid(t("{date} is not a valid date. The format for dates is {format}").replace("{date}", v).replace("{format}", this.inputFormat));
    } else {
      if (this.maxDate && dv.getTime() > this.maxDate.getTime()) {
        this.setInvalid(t("The date in this field must be before {maxDate}.").replace("{maxDate}", this.maxDate.format(this.inputFormat)));
      }
      if (this.minDate && dv.getTime() < this.minDate.getTime()) {
        this.setInvalid(t("The date in this field must be after {minDate}.").replace("{minDate}", this.minDate.format(this.inputFormat)));
      }
    }
  }
  set value(v) {
    if (v === void 0 || v === null) {
      if (this.timefield) {
        this.timefield.value = "";
      }
      super.value = "";
      return;
    }
    const [date, time] = v.split("T");
    const d = DateTime.createFromFormat(date + "", this.outputFormat);
    if (!d) {
      throw new Error("Invalid date format " + v);
    }
    if (this.timefield && time) {
      this.timefield.value = time.substring(0, 5);
    }
    this.date = d;
    this.picker.setValue(d);
    super.value = d.format(this.inputFormat);
  }
  get value() {
    const date = this.getValueAsDateTime();
    if (!date) {
      return void 0;
    }
    const timeFormat = this.timefield ? "TH:i" : "";
    return date.format(this.inputFormat + timeFormat);
  }
  getValueAsDateTime() {
    let v = super.value, timeFormat = "";
    if (this.timefield) {
      v += "T" + this.timefield.value;
      timeFormat = "TH:i";
    }
    let date;
    if (!v || !(date = DateTime.createFromFormat(v, this.inputFormat + timeFormat))) {
      return void 0;
    }
    return date;
  }
};
var datefield = (config) => createComponent(new DateField(), config);

// goui/script/Notifier.ts
var Notifier = class {
  /**
   * Show an error toast
   *
   * @param msg any string or object with a message property
   * @param timeout Timeout in seconds before it automatically disappears. It also dissappears on any mouseclick
   */
  static error(msg, timeout = 0) {
    console.error(msg);
    if (msg instanceof Object && msg.message) {
      msg = msg.message;
    } else if (typeof msg != "string") {
      msg = msg + "";
    }
    return new Message(msg, "error", timeout);
  }
  /**
   * Show success toast
   *
   * @param msg
   * @param timeout
   */
  static success(msg, timeout = 3e3) {
    return new Message(msg, "success", timeout);
  }
  /**
   * Show a notice toast
   *
   * @param msg
   * @param timeout
   */
  static notice(msg, timeout = 3e3) {
    return new Message(msg, "notice", timeout);
  }
  /**
   * Show a warning toast
   * @param msg
   * @param timeout
   */
  static warning(msg, timeout = 3e3) {
    return new Message(msg, "warning", timeout);
  }
};
var Message = class {
  constructor(msg, type, timeout = 3e3) {
    const alert2 = comp(
      {
        cls: "goui-alert " + type
      },
      comp({
        tagName: "span",
        text: msg
      })
    );
    root.items.add(alert2);
    if (timeout) {
      this.timeout = window.setTimeout(() => {
        alert2.remove();
      }, timeout);
    }
    setTimeout(() => {
      document.body.addEventListener("click", () => {
        alert2.remove();
        if (this.timeout) {
          clearTimeout(this.timeout);
        }
      }, { once: true });
    });
  }
};

// goui/script/component/form/Form.ts
var Form = class extends ContainerField {
  constructor() {
    super("form");
    this.baseCls = "goui-form";
    this.hideLabel = true;
  }
  internalRender() {
    const el = super.internalRender();
    this.el.noValidate = true;
    el.addEventListener("submit", (event) => {
      event.preventDefault();
      this.submit();
    });
    el.addEventListener("reset", (e) => {
      e.preventDefault();
      this.reset();
    });
    el.addEventListener("keydown", (e) => {
      if ((e.ctrlKey || e.metaKey) && e.key == "Enter") {
        e.preventDefault();
        if (document.activeElement && "blur" in document.activeElement) {
          document.activeElement.blur();
        }
        this.submit();
      }
    });
    return el;
  }
  /**
   * Get all form values
   */
  getValues() {
    return this.value;
  }
  /**
   * Set form field values
   *
   * @param v
   */
  setValues(v) {
    this.value = v;
  }
  /**
   * @inheritDoc
   */
  reset() {
    this.findFields().forEach((field) => {
      field.reset();
    });
  }
  /**
   * Validates the form and submits it using the handler function passed with the config.
   */
  submit() {
    return __async(this, null, function* () {
      const el = this.el;
      this.clearInvalid();
      if (this.isValid()) {
        el.cls(["+valid", "-invalid"]);
        let handlerResponse = void 0;
        if (this.handler) {
          try {
            handlerResponse = yield this.handler(this);
          } catch (e) {
            el.cls(["-valid", "+invalid"]);
            const msg = typeof e == "string" ? e : e.message;
            Notifier.error(msg);
            return;
          }
        }
        this.fire("submit", this, handlerResponse);
      } else {
        el.cls(["-valid", "+invalid"]);
        const invalid = this.findFirstInvalid();
        if (invalid) {
          invalid.focus();
        }
        this.setInvalid(t("You have errors in your form. The invalid fields are marked."));
      }
    });
  }
  setInvalid(msg) {
    super.setInvalid(msg);
    Notifier.error(msg);
  }
};
var form = (config, ...items) => createComponent(new Form(), config, items);

// goui/script/component/form/NumberField.ts
var NumberField = class extends TextField {
  constructor() {
    super(...arguments);
    this.type = "number";
    this.decimals = 2;
  }
  validate() {
    super.validate();
    const v = super.value;
    if (v && isNaN(v)) {
      this.setInvalid("Incorrect number format");
    }
    if (this.max !== void 0 && v > this.max) {
      this.setInvalid("Number is too big");
    }
    if (this.min !== void 0 && v < this.min) {
      this.setInvalid("Number is too small");
    }
  }
  set value(v) {
    if (isNaN(v)) {
      throw new Error("Invalid number");
    } else if (v) {
      super.value = +v.toFixed(this.decimals);
    }
  }
  get value() {
    const v = super.value;
    return v === void 0 || isNaN(v) ? void 0 : +(+v).toFixed(this.decimals);
  }
  createControl() {
    super.createControl();
    this._input.cls("+number");
    if (this.min !== void 0) {
      this._input.attr("min", this.min);
    }
    if (this.max !== void 0) {
      this._input.attr("max", this.max);
    }
    if (this.decimals) {
      this._input.attr("step", "0.".padEnd(this.decimals + 1, "0") + "1");
    }
    return this._input;
  }
};
var numberfield = (config) => createComponent(new NumberField(), config);

// goui/script/component/CardContainer.ts
var CardContainer = class extends Component {
  constructor() {
    super();
    this.baseCls = "cards";
    this.items.on("beforeadd", (card, item) => {
      item.hide();
      item.on("show", (comp2) => {
        const index = this.findItemIndex(comp2);
        this.activeItem = index;
      });
      item.el.classList.add("card-container-item");
    });
  }
  internalRender() {
    this.setCardVisibilities();
    const el = super.internalRender();
    return el;
  }
  setCardVisibilities() {
    this.items.forEach((item, index) => {
      if (index == this.activeItem) {
        item.show();
      } else {
        item.hide();
      }
    });
  }
  /**
   * The active card index. Defaults to 0 if not given.
   */
  set activeItem(ref) {
    let index;
    if (ref instanceof Component) {
      index = this.findItemIndex(ref);
    } else {
      index = ref;
    }
    const old = this._activeItem;
    this._activeItem = index;
    if (old !== index) {
      this.fire("cardchange", this, index, old);
    }
    this.setCardVisibilities();
  }
  /**
   * The active card index. Defaults to 0 if not given.
   */
  get activeItem() {
    if (this._activeItem == void 0 && this.items.count()) {
      this._activeItem = 0;
    }
    return this._activeItem;
  }
  focus(o) {
    if (this.activeItem) {
      const activeItem = this.items.get(this.activeItem);
      if (activeItem) {
        activeItem.focus(o);
        return;
      }
    }
    super.focus(o);
  }
};
var cards = (config, ...items) => createComponent(new CardContainer(), config, items);

// goui/script/component/picker/RecurrencePicker.ts
var _RecurrencePicker = class _RecurrencePicker extends CardContainer {
  constructor(startDate) {
    super();
    this.baseCls = "recurrencepicker";
    this.weekOfMonth = 1;
    this.startDate = startDate;
    this.width = 450;
    const weeklyOptions = checkboxgroup({
      itemId: "weeklyOptions",
      label: "Weekdays",
      options: [0, 1, 2, 3, 4, 5, 6].map((i) => {
        return { label: DateTime.dayNames[DateTime.dayMap[i]].substring(0, 2), name: DateTime.dayMap[i] };
      })
    });
    this.menu = comp({}, ...this.quickMenuItems());
    this.form = form(
      {},
      comp(
        { cls: "flow" },
        comp({ text: t("Every") }),
        numberfield({
          decimals: 0,
          name: "interval",
          itemId: "interval",
          min: 1,
          width: 50,
          value: 1,
          listeners: {
            "change": (me, newVal, oldVal) => {
              if (oldVal == 1 && newVal != 1 || oldVal != 1 && newVal == 1) {
                this.form.findField("frequency").drawOptions();
              }
            }
          }
        }),
        select({
          name: "frequency",
          itemId: "frequency",
          width: 120,
          textRenderer: (r) => this.form.findChild("interval").value == 1 ? r.text : r.plural,
          options: Object.keys(_RecurrencePicker.frequencies).map((k) => ({
            value: k,
            text: _RecurrencePicker.frequencies[k][0],
            plural: _RecurrencePicker.frequencies[k][1]
          })),
          listeners: {
            "change": (me, v) => {
              this.changeFrequency(v);
            }
          }
        }),
        select({
          hidden: true,
          disabled: true,
          name: "monthlyType",
          itemId: "monthlyOptions",
          label: t("at the"),
          width: 120,
          value: "byMonthDay",
          options: [
            { value: "byMonthDay", name: this.startDate.format("jS") },
            { value: "byDay", name: this.getSuffix() + " " + this.startDate.format("l") }
          ]
        }),
        weeklyOptions,
        textfield({
          hidden: true,
          name: "byDay",
          listeners: {
            "change": (fld, val) => {
              for (let j = 0; j < 7; j++) {
                const cb = weeklyOptions.items.get(j);
                cb.value = val.indexOf(cb.name) !== -1;
              }
            }
          }
        }),
        comp(
          { cls: "flow" },
          comp({ html: t("Ends") }),
          select({
            width: 100,
            name: "endsRadio",
            value: "forever",
            textRenderer: (r) => r.text,
            options: [
              { text: t("Never"), value: "forever" },
              { text: t("After"), value: "count" },
              { text: t("At"), value: "until" }
            ],
            listeners: {
              "change": (me, v) => {
                const c = this.form.findField("count"), u = this.form.findField("until");
                c.hidden = c.disabled = v != "count";
                u.hidden = u.disabled = v != "until";
              }
            }
          })
        ),
        this.endAtCtr = comp(
          {},
          numberfield({
            itemId: "repeatCount",
            name: "count",
            hidden: true,
            max: 1e3,
            width: 80,
            value: 13,
            decimals: 0,
            hint: t("times")
            // should be suffix
          }),
          datefield({
            itemId: "endDate",
            name: "until",
            width: 160,
            hidden: true,
            required: false
          })
        )
      ),
      tbar(
        {},
        btn({
          text: t("Back"),
          handler: (b) => this.activeItem = 0
        }),
        comp({ flex: 1 }),
        btn({
          text: t("Ok"),
          handler: (btn2) => {
            this.setValue(this.createCustomRule(this.form.value));
            this.parent.close();
            this.activeItem = 0;
          }
        })
      )
    );
    this.items.add(this.menu, this.form);
  }
  quickMenuItems() {
    return [
      btn({
        text: t("Not recurring"),
        handler: (_) => this.setValue(null)
      }),
      comp({ tagName: "hr" }),
      btn({
        text: t("Daily"),
        handler: (_) => this.setValue({ frequency: "daily" })
      }),
      btn({
        text: t("Weekly") + " " + t("at ") + this.startDate.format("l"),
        handler: (_) => this.setValue({ frequency: "weekly" })
      }),
      btn({
        text: t("Monthly") + " " + t("at day") + " " + this.startDate.format("j"),
        handler: (_) => this.setValue({ frequency: "monthly", byMonthDay: [+this.startDate.format("j")] })
      }),
      btn({
        text: t("Monthly") + " " + t("at the") + " " + this.getSuffix() + " " + this.startDate.format("l"),
        handler: (_) => this.setValue({
          frequency: "monthly",
          byDay: [{ day: DateTime.dayMap[this.startDate.getWeekDay()], nthOfPeriod: this.weekOfMonth }]
        })
      }),
      btn({
        text: t("Annually") + " " + t("at ") + this.startDate.format("j F"),
        handler: (_) => this.setValue({ frequency: "yearly" })
      }),
      btn({
        text: t("Each working day"),
        handler: (_) => this.setValue({
          frequency: "weekly",
          byDay: [{ day: "mo" }, { day: "tu" }, { day: "we" }, { day: "th" }, { day: "fr" }]
        })
      }),
      comp({ tagName: "hr" }),
      btn({
        text: t("Customize") + "...",
        handler: () => {
          var _a;
          this.changeFrequency(((_a = this.rule) == null ? void 0 : _a.frequency) || "yearly");
          this.activeItem = 1;
        }
      })
    ];
  }
  createCustomRule(values) {
    const rule = { frequency: values.frequency };
    if (values.interval != 1)
      rule.interval = values.interval;
    if (values.until)
      rule.until = values.until;
    if (values.count)
      rule.count = values.count;
    if (values.monthlyType) {
      switch (values.monthlyType) {
        case "byMonthDay":
          rule.byMonthDay = [parseInt(this.startDate.format("j"))];
          break;
        case "byDay":
          rule.byDay = [{
            day: DateTime.dayMap[this.startDate.getWeekDay()],
            nthOfPeriod: this.weekOfMonth
          }];
          break;
      }
    }
    return rule;
  }
  setStartDate(date) {
    this.startDate = date.clone();
    for (var i = 0, m = date.getMonth(); m == date.getMonth(); date = date.addDays(-7)) {
      i++;
    }
    this.weekOfMonth = i;
    this.menu.items.clear().add(...this.quickMenuItems());
  }
  changeFrequency(f) {
    const record = _RecurrencePicker.frequencies[f];
    this.form.findChild("frequency").value = f;
    var repeat = this.endAtCtr.items.get(0);
    if (repeat.disabled) {
      repeat.value = record[2];
    }
    var until = this.endAtCtr.items.get(1);
    if (until.disabled) {
      var add = record[3].split("-");
      add[1] == "d" ? this.startDate.addDays(parseInt(add[0])) : this.startDate.addYears(parseInt(add[0]));
      until.value = this.startDate.format("Y-m-d");
    }
    ["weekly", "monthly"].map((p2) => {
      const el = this.form.findChild(p2 + "Options");
      el.hidden = p2 != f;
      el.disabled = p2 != f;
    });
  }
  getSuffix(week) {
    week = week || this.weekOfMonth;
    switch (week) {
      case 1:
        return t("first");
      case 2:
        return t("second");
      case 3:
        return t("third");
      case 4:
        return t("fourth");
      default:
        return t("last");
    }
  }
  setValue(rrule) {
    if (this.rule == rrule)
      return;
    this.rule = rrule;
    this.fire("select", this, rrule);
    const form2 = this.form;
    if (rrule && rrule.frequency) {
      form2.value = rrule;
      this.changeFrequency(rrule.frequency);
      if (rrule.until) {
        form2.findField("endsRadio").value = "until";
        form2.value = { until: rrule.until };
      }
      if (rrule.count) {
        form2.findField("endsRadio").value = "count";
        form2.value = { count: rrule.count };
      }
      if (rrule.byDay) {
        form2.findField("monthlyOptions").value = "byDay";
      }
      if (rrule.byMonthDay) {
        form2.findField("monthlyOptions").value = "byMonthDay";
      }
    }
  }
};
_RecurrencePicker.frequencies = {
  "daily": [t("day"), t("days"), 30, "30-d", t("Daily")],
  "weekly": [t("week"), t("weeks"), 13, "91-d", t("Weekly")],
  "monthly": [t("month"), t("months"), 12, "1-y", t("Monthly")],
  "yearly": [t("year"), t("years"), 5, "5-y", t("Annually")]
};
var RecurrencePicker = _RecurrencePicker;

// goui/script/component/picker/ListPicker.ts
var ListPicker = class extends Component {
  constructor(list2) {
    super();
    this.list = list2;
    this.list.rowSelectionConfig = { multiSelect: false };
    this.items.add(list2);
    this.list.on("rowclick", (table2, rowIndex, row, ev) => {
      this.onSelect();
    });
    this.list.on("hide", () => {
      this.list.rowSelection.clear();
    });
    this.list.store.on("datachanged", () => {
      if (this.list.store.count() > 0) {
        this.list.rowSelection.selected = [0];
      }
    }, { buffer: 0 });
    this.list.el.addEventListener("mousedown", (ev) => {
      ev.stopPropagation();
    });
    this.list.el.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "Enter":
          ev.preventDefault();
          this.onSelect();
          break;
      }
    });
  }
  focus(o) {
    this.list.focus(o);
  }
  onSelect() {
    const selected = this.list.rowSelection.selected;
    if (selected.length) {
      this.fire("select", this, this.list.store.get(selected[0]));
    }
  }
};
var listpicker = (config) => createComponent(new ListPicker(config.list), config);

// goui/script/component/form/AutocompleteField.ts
var AutocompleteField = class extends TextField {
  /**
   *
   * @param list The table to use for suggestions
   * @param buffer Buffer typing in the input in ms
   */
  constructor(list2, buffer = 300) {
    super();
    this.list = list2;
    this.buffer = buffer;
    this.autocomplete = "off";
    this.baseCls += " autocomplete";
    this.picker = listpicker({
      list: list2
    });
    this.picker.on("select", (tablePicker, record) => {
      tablePicker.list.findAncestorByType(Menu).hide();
      this.focus();
      this.value = this.pickerRecordToValue(this, record);
    });
    this.menu = menu(
      {
        height: 300,
        cls: "scroll",
        listeners: {
          hide: (menu2) => {
            if (menu2.rendered) {
              const textfield2 = menu2.findAncestorByType(TextField);
              textfield2.focus();
            }
          }
        }
      },
      this.picker
    );
    this.menuButton = btn({
      icon: "expand_more",
      type: "button",
      handler: () => {
        this.fire("autocomplete", this, "");
      },
      menu: this.menu
    });
  }
  /**
   * Method that transforms a record from the TablePicker store to a value for this field.
   * This is not necessarily a text value. In conjunction with {@see valueToTextField()} this
   * could also be an ID of an object for example.
   *
   * @param field
   * @param record
   */
  pickerRecordToValue(field, record) {
    return record.id;
  }
  /**
   * This method transforms the value in to a text representation for the input field
   *
   * @param field
   * @param value
   */
  valueToTextField(field, value) {
    return __async(this, null, function* () {
      return "";
    });
  }
  setInputValue(v) {
    this.valueToTextField(this, v).then((v2) => {
      if (this._input) {
        super.setInputValue(v2);
      } else {
        this.on("render", () => {
          super.setInputValue(v2);
        }, { once: true });
      }
    });
  }
  get value() {
    return this._value;
  }
  set value(v) {
    super.value = v;
  }
  internalRender() {
    this.buttons = this.buttons || [];
    this.buttons.push(this.menuButton);
    const el = super.internalRender();
    this.menu.alignTo = this.wrap;
    this.menu.alignToInheritWidth = true;
    this._input.addEventListener("input", FunctionUtil.buffer(this.buffer, this.onInput.bind(this)));
    this._input.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "Enter":
          if (!this.menu.hidden) {
            ev.preventDefault();
            this.picker.onSelect();
          }
          break;
        case "ArrowDown":
          ev.preventDefault();
          this.menuButton.showMenu();
          this.list.focus();
          break;
        case "Escape":
          if (!this.menu.hidden) {
            this.menu.hide();
            ev.preventDefault();
            ev.stopPropagation();
            this.focus();
          }
          break;
      }
    });
    return el;
  }
  onInput(ev) {
    this.menuButton.showMenu();
    this.fire("autocomplete", this, this._input.value);
  }
};
var autocomplete = (config) => createComponent(new AutocompleteField(config.list), config);

// goui/script/component/form/CheckboxField.ts
var CheckboxField = class extends Field {
  /**
   *
   * @param type Render the checkbox as a checkbox, switch or toggle button
   */
  constructor(type = "box") {
    super();
    this.type = type;
    this.baseCls = "goui-form-field check";
    // we handle it with the native change event here
    this.fireChangeOnBlur = false;
    this.type = type;
    this.el.cls(type, true);
  }
  applyTitle() {
    if (this.title) {
      this.input.title = this.title;
    }
  }
  focus(o) {
    var _a;
    if (!this.input) {
      super.focus(o);
    }
    (_a = this.input) == null ? void 0 : _a.focus(o);
  }
  createLabel() {
    const old = this.label;
    this.label = "";
    const span = super.createLabel();
    this.label = old;
    return span;
  }
  createControl() {
    const el = E("input").on("change", () => {
      this.fireChange();
    });
    el.type = "checkbox";
    el.required = this.required;
    el.name = this.name;
    el.readOnly = this.readOnly;
    el.checked = !!this.value;
    this.input = el;
    if (this.invalidMsg) {
      this.applyInvalidMsg();
    }
    const control = E(
      "div",
      this.input
    );
    if (this.label) {
      const lbl = E("span", this.label).cls("box-label");
      if (this._cls) {
        lbl.cls("+" + this._cls);
      }
      control.append(lbl);
    }
    return control;
  }
  set readOnly(readOnly) {
    super.readOnly = readOnly;
    if (this.input) {
      this.input.readOnly = this.readOnly;
    }
  }
  get readOnly() {
    return super.readOnly;
  }
  setInvalid(msg) {
    super.setInvalid(msg);
    if (this.rendered) {
      this.applyInvalidMsg();
    }
  }
  applyInvalidMsg() {
    super.applyInvalidMsg();
    this.input.setCustomValidity(this.invalidMsg);
    if (this.input.offsetParent) {
      this.input.reportValidity();
    }
    if (this.invalidMsg != "") {
      this.input.addEventListener("input", () => {
        this.clearInvalid();
      }, { once: true });
    }
  }
  clearInvalid() {
    super.clearInvalid();
    this.applyInvalidMsg();
  }
  getInput() {
    return this.input;
  }
  set value(v) {
    if (this.input) {
      this.input.checked = v;
    }
    super.value = v;
  }
  get value() {
    if (!this.input) {
      return super.value;
    } else {
      return this.input.checked;
    }
  }
  set name(name) {
    super.name = name;
    if (this.input) {
      this.input.name = this.name;
    }
  }
  get name() {
    return super.name;
  }
  validate() {
    super.validate();
    if (!this.input.validity.valid) {
      this.setInvalid(this.input.validationMessage);
    }
  }
};
var checkbox = (config) => createComponent(new CheckboxField((config == null ? void 0 : config.type) || "box"), config);

// goui/script/component/form/ColorField.ts
var ColorField = class extends Field {
  constructor() {
    super();
    this.picker = new ColorPicker();
    this.buttons = [
      this.pickerButton = btn({
        icon: "expand_more",
        menu: menu(
          {
            alignTo: this.el,
            alignToInheritWidth: true
          },
          this.picker
        )
      })
    ];
  }
  createControl() {
    this.colorDot = E("div").cls("+color-dot");
    if (this.value) {
      this.colorDot.style.backgroundColor = "#" + this.value;
    }
    this.el.cls("+no-floating-label");
    this.picker.on("select", (colorPicker, val) => {
      this.pickerButton.menu.hide();
      this.clearInvalid();
      this.focus();
      this.value = val;
    });
    return this.colorDot;
  }
  setInvalid(msg) {
    super.setInvalid(msg);
    if (this.rendered) {
      this.applyInvalidMsg();
    }
  }
  clearInvalid() {
    super.clearInvalid();
    this.applyInvalidMsg();
  }
  set value(v) {
    if (this.colorDot) {
      this.colorDot.style.backgroundColor = "#" + v;
    }
    super.value = v;
  }
  get value() {
    return super.value;
  }
};
var colorfield = (config) => createComponent(new ColorField(), config);

// goui/script/component/form/DisplayField.ts
var DisplayField = class extends Field {
  constructor(tagName = "div") {
    super(tagName);
    this.baseCls = "";
  }
  renderControl() {
  }
};
var displayfield = (config, ...items) => createComponent(new DisplayField(config == null ? void 0 : config.tagName), config, items);

// goui/script/component/form/Fieldset.ts
var Fieldset = class extends Component {
  constructor() {
    super("fieldset");
    this.baseCls = "goui-fieldset";
  }
  internalRender() {
    const el = super.internalRender();
    if (this.legend) {
      const l = document.createElement("h3");
      l.innerHTML = this.legend;
      l.classList.add("legend");
      el.insertBefore(l, el.firstChild);
    }
    return el;
  }
};
var fieldset = (config, ...items) => createComponent(new Fieldset(), config, items);

// goui/script/component/form/HtmlField.ts
var _HtmlField = class _HtmlField extends Field {
  constructor() {
    super("div");
    this.baseCls = "goui-form-field goui-html-field";
    /**
     * Toolbar items to enable.
     *
     * If you can't use inline css then use:
     *
     * ```
     * [
     * 		"bold", "italic", "underline",
     * 		"-",
     * 		"foreColor", "removeFormat",
     * 		"-",
     * 		"insertOrderedList",
     * 		"insertUnorderedList",
     * 		"-",
     * 		"indent",
     * 		"outdent",
     * 		"-",
     * 		"image"
     * 	]
     * ```
     */
    this.toolbarItems = [
      "bold",
      "italic",
      "underline",
      "-",
      "foreColor",
      "backColor",
      "removeFormat",
      "-",
      "justifyLeft",
      "justifyCenter",
      "justifyRight",
      "-",
      "insertOrderedList",
      "insertUnorderedList",
      "-",
      "indent",
      "outdent",
      "-",
      "image"
    ];
    this.commands = {
      bold: { icon: "format_bold", title: "Bold" },
      italic: { icon: "format_italic", title: "Italic" },
      underline: { icon: "format_underlined", title: "Underline" },
      strikeThrough: { icon: "format_strikethrough", title: "Strikethrough" },
      foreColor: {
        title: "Text color",
        icon: "format_color_text",
        menu: colormenu({
          updateButton: false,
          listeners: {
            select: (menu2, color) => {
              this.execCmd("foreColor", color || "#000000");
            },
            beforeshow: (menu2) => {
              menu2.value = this.tbar.findItem("foreColor").el.style.color || "";
            }
          }
        }),
        applyFn: () => {
        },
        updateFn: (btn2) => {
          const s = document.getSelection();
          if (!s || !s.anchorNode) {
            btn2.el.style.color = "";
            return;
          }
          let el = s.anchorNode;
          while (!(el instanceof HTMLElement) && el != null) {
            el = el.parentElement;
          }
          if (!el) {
            btn2.el.style.color = "";
            return;
          }
          const closest = el.closest("[color]");
          if (!closest) {
            btn2.el.style.color = "";
            return;
          }
          btn2.el.style.color = closest.getAttribute("color") + "";
        }
      },
      backColor: {
        icon: "format_color_fill",
        title: "Background color",
        menu: colormenu({
          updateButton: false,
          listeners: {
            select: (menu2, color) => {
              this.execCmd("backColor", color || "#ffffff");
            },
            beforeshow: (menu2) => {
              menu2.value = this.tbar.findItem("backColor").el.style.color || "";
            }
          }
        }),
        applyFn: () => {
        },
        updateFn: (btn2) => {
          const s = document.getSelection();
          let bgcolor = s ? this.closestStyle(s.anchorNode, "backgroundColor") : void 0;
          btn2.el.style.backgroundColor = bgcolor || "";
        }
      },
      justifyLeft: { icon: "format_align_left", title: "Align left" },
      justifyCenter: { icon: "format_align_center", title: "Align center" },
      justifyRight: { icon: "format_align_right", title: "Align right" },
      insertOrderedList: { icon: "format_list_numbered", title: "Numbered list" },
      insertUnorderedList: { icon: "format_list_bulleted", title: "Bullet list" },
      indent: { icon: "format_indent_increase", title: "Indent" },
      outdent: { icon: "format_indent_decrease", title: "Outdent" },
      image: {
        icon: "image",
        title: "Image",
        applyFn: () => {
          const inp = document.createElement("input");
          inp.setAttribute("type", "file");
          inp.accept = "image/*";
          inp.multiple = true;
          inp.onchange = () => {
            if (!inp.files) {
              this.editor.focus();
              return;
            }
            Array.from(inp.files).forEach((file) => {
              this.handleImage(file);
            });
          };
          window.addEventListener("focus", () => {
            this.editor.focus();
          }, { once: true });
          inp.click();
        }
      },
      removeFormat: { icon: "format_clear", title: "Remove formatting" }
      // emoji: {icon: 'insert_emoticon'},
      // html: {icon: 'code'}
    };
    this.lineIndex = 0;
    this.lineSequence = "";
    this.value = "";
    document.execCommand("useCSS", false, "true");
    document.execCommand("styleWithCSS", false, "false");
  }
  // noinspection JSUnusedGlobalSymbols
  getEditor() {
    if (!this.editor) {
      throw new Error("Render first");
    }
    return this.editor;
  }
  closestStyle(el, styleProp) {
    while (el = el.parentElement) {
      if (el == this.editor)
        return void 0;
      if (el.style[styleProp]) {
        return el.style[styleProp];
      }
    }
    return void 0;
  }
  updateToolbar() {
    const t2 = this.tbar;
    for (const cmd of this.toolbarItems) {
      if (cmd == "-")
        continue;
      const config = this.commands[cmd];
      if (config.updateFn) {
        config.updateFn.call(this, t2.findItem(cmd));
      } else {
        t2.findItem(cmd).el.classList.toggle("pressed", document.queryCommandState(cmd));
      }
    }
    this.fire("updatetoolbar", this);
  }
  execCmd(cmd, value) {
    document.execCommand(cmd, false, value);
    const t2 = this.tbar, config = this.commands[cmd];
    if (config) {
      if (config.updateFn) {
        config.updateFn.call(this, t2.findItem(cmd));
      } else {
        t2.findItem(cmd).el.classList.toggle("pressed", document.queryCommandState(cmd));
      }
    }
    this.editor.focus();
  }
  /**
   * Inserts an HTML string at the insertion point (deletes selection). Requires a valid HTML string as a value argument.
   *
   * @param html
   */
  insertHtml(html) {
    document.execCommand("insertHTML", false, html);
  }
  // noinspection JSUnusedGlobalSymbols
  internalRemove() {
    if (this.tbar) {
      this.tbar.remove();
    }
    super.internalRemove();
  }
  getToolbar() {
    if (this.tbar) {
      return this.tbar;
    }
    this.tbar = tbar({
      cls: "frame html-field-toolbar"
    });
    for (const cmd of this.toolbarItems) {
      if (cmd == "-") {
        this.tbar.items.add(comp({ tagName: "hr" }));
      } else {
        const config = this.commands[cmd];
        this.tbar.items.add(btn({
          itemId: cmd,
          icon: config.icon,
          menu: config.menu,
          title: config.title,
          handler: (btn2) => {
            if (!config.applyFn) {
              this.execCmd(btn2.itemId);
            } else {
              config.applyFn.call(this, btn2);
            }
          }
        }));
      }
    }
    root.items.add(this.tbar);
    return this.tbar;
  }
  showToolbar() {
    const rect = this.el.getBoundingClientRect();
    this.getToolbar().show();
    const h = this.getToolbar().height;
    const style = this.getToolbar().el.style;
    const maxX = window.innerWidth - this.getToolbar().width;
    style.left = Math.min(maxX, rect.x + window.scrollX) + "px";
    style.top = window.scrollY + rect.top - h - 8 + "px";
  }
  hideToolbar() {
    this.getToolbar().hide();
  }
  createControl() {
    const el = this.el;
    el.addEventListener("click", () => {
      this.editor.focus();
    });
    const v = this.value;
    this.editor = document.createElement("div");
    this.editor.contentEditable = "true";
    this.editor.classList.add("editor");
    this.editor.classList.add("text");
    if (this.placeholder) {
      this.editor.dataset.placeholder = this.placeholder;
    }
    if (v) {
      this.editor.innerHTML = v;
    }
    el.appendChild(this.editor);
    const selectionChangeFn = FunctionUtil.buffer(300, () => this.updateToolbar());
    this.editor.addEventListener("focus", () => {
      document.addEventListener("selectionchange", selectionChangeFn);
      this.showToolbar();
    });
    this.editor.addEventListener("blur", (e) => {
      document.removeEventListener("selectionchange", selectionChangeFn);
      if (!(e.relatedTarget instanceof HTMLElement) || !this.getToolbar().el.contains(e.relatedTarget)) {
        this.hideToolbar();
      }
    });
    this.editor.addEventListener("keydown", (ev) => {
      this.onKeyDown(ev);
    });
    this.editor.addEventListener("drop", (ev) => {
      this.onDrop(ev);
    });
    this.editor.addEventListener("dragover", (ev) => {
      ev.preventDefault();
    });
    this.editor.addEventListener("paste", (ev) => {
      this.onPaste(ev);
    });
    return this.editor;
  }
  // private isChildOfEditor(el:Node): boolean {
  // 	let p = el.parentNode;
  // 	if(p == this.editor) {
  // 		return true;
  // 	}
  //
  // 	if(!p || p == document.body) {
  // 		return false;
  // 	}
  //
  // 	return this.isChildOfEditor(p);
  //
  // }
  set value(v) {
    if (this.editor) {
      this.editor.innerHTML = v;
    }
    super.value = v;
  }
  get value() {
    if (!this.editor) {
      return super.value;
    } else {
      return this.editor.innerHTML;
    }
  }
  focus(o) {
    this.editor.focus(o);
  }
  static removeCharsFromCursorPos(count) {
    const sel = window.getSelection();
    const range = sel.getRangeAt(0);
    const clone = range.cloneRange();
    clone.setStart(range.startContainer, range.startOffset);
    clone.setEnd(range.startContainer, range.startOffset + count);
    clone.deleteContents();
  }
  onKeyDown(ev) {
    this.clearInvalid();
    if (ev.key == "Enter" || ev.key == "Backspace") {
      this.lineIndex = 0;
      this.lineSequence = "";
    } else if (this.lineIndex < 3 && ev.key.length == 1) {
      this.lineIndex++;
      this.lineSequence += ev.key;
    }
    if (browser.isWebkit() && ev.shiftKey && ev.key == "Enter" && (document.queryCommandState("insertorderedlist") || document.queryCommandState("insertunorderedlist"))) {
      ev.stopPropagation();
      this.execCmd("InsertHtml", browser.isFirefox() ? "<br />" : "<br /><br />");
      this.focus();
    } else if (ev.key == "Tab") {
      ev.preventDefault();
      if (document.queryCommandState("insertorderedlist") || document.queryCommandState("insertunorderedlist")) {
        this.execCmd(ev.shiftKey ? "outdent" : "indent");
      } else {
        this.execCmd("InsertText", "	");
      }
      this.focus();
    } else if (ev.key == " ") {
      if (this.lineSequence == "1. ") {
        this.execCmd("insertOrderedList");
        _HtmlField.removeCharsFromCursorPos(2);
        ev.preventDefault();
      } else if (this.lineSequence == "- ") {
        this.execCmd("insertUnorderedList");
        _HtmlField.removeCharsFromCursorPos(1);
        ev.preventDefault();
      }
    } else if (browser.isMac()) {
      if (ev.ctrlKey) {
        let cmd;
        switch (ev.key) {
          case "b":
            cmd = "bold";
            break;
          case "i":
            cmd = "italic";
            break;
          case "u":
            cmd = "underline";
            break;
        }
        if (cmd) {
          this.execCmd(cmd);
          ev.preventDefault();
        }
      }
    }
  }
  onDrop(e) {
    if (!e.dataTransfer || !e.dataTransfer.files) {
      return;
    }
    e.preventDefault();
    this.editor.focus();
    Array.from(e.dataTransfer.files).forEach((file) => {
      if (file.type.match(/^image\//)) {
        this.handleImage(file);
      } else {
        this.fire("attach", this, file);
      }
    });
  }
  /**
   * Generate unique ID
   */
  static imgUID() {
    return "img-" + ++_HtmlField._uid;
  }
  handleImage(file) {
    let imgEl;
    if (file.type.match(/^image\//)) {
      const objectURL = URL.createObjectURL(file), uid = _HtmlField.imgUID();
      const img = `<img id="${uid}" src="${objectURL}" alt="${file.name}" />`;
      this.insertHtml(img);
      imgEl = document.getElementById(uid);
      imgEl.removeAttribute("id");
      imgEl.addEventListener("load", () => {
        const width = imgEl.offsetWidth, height = imgEl.offsetHeight;
        imgEl.setAttribute("style", `max-width: 100%;height:auto;aspect-ratio: ${width} / ${height};`);
      });
      this.fire("insertimage", this, file, imgEl);
    }
  }
  onPaste(e) {
    if (!e.clipboardData || !e.clipboardData.files) {
      return;
    }
    const files = Array.from(e.clipboardData.files);
    files.forEach((file) => {
      if (file.type.match(/^image\//)) {
        this.handleImage(file);
      } else {
        this.fire("attach", this, file);
      }
      e.preventDefault();
    });
  }
};
_HtmlField._uid = 0;
var HtmlField = _HtmlField;
var htmlfield = (config) => createComponent(new HtmlField(), config);

// goui/script/component/form/MapField.ts
var MapField = class extends Field {
  constructor(buildField) {
    super("div");
    this.buildField = buildField;
    this.rows = [];
    this.baseCls = "";
    this._nextKey = 1;
  }
  renderControl() {
  }
  set value(v) {
    super.value = v;
    this.items.clear();
    if (v) {
      for (const key in v) {
        this.internalAdd(v[key], key);
      }
    }
  }
  get value() {
    const v = {};
    this.rows.forEach((row) => {
      v[row.key] = row.field.value;
    });
    return v;
  }
  /**
   * Add value to the map.
   *
   * Also fires change event
   *
   * @param data
   * @param key
   */
  add(data, key) {
    if (!this.oldValue) {
      this.captureValueForChange();
    }
    this.internalAdd(data, key);
    this.fireChange();
  }
  internalAdd(data, key) {
    const row = {
      key: key === void 0 ? this.nextKey() : key,
      field: this.buildField(data),
      isNew: true
    };
    row.field.items.add(btn({
      icon: "delete",
      handler: (button) => {
        this.rows.splice(this.rows.indexOf(row), 1);
        row.field.remove();
      }
    }));
    row.field.value = data;
    this.rows.push(row);
    this.items.add(row.field);
  }
  nextKey() {
    return "_new_" + this._nextKey++;
  }
  reset() {
    super.reset();
    this.rows = [];
    this.items.clear();
  }
};
var mapfield = (config, ...items) => createComponent(new MapField(config.buildField), config, items);

// goui/script/component/form/RadioField.ts
var RadioField = class extends Field {
  constructor(type = "box") {
    super("div");
    this.options = [];
    this.inputs = {};
    this.baseCls = "goui-form-field radiogroup";
    this.type = type;
    this.el.cls(type, true);
  }
  createLabel() {
    return;
  }
  createControl() {
    if (!this.name && !this.itemId) {
      this.itemId = "radio-" + Component.uniqueID();
    }
    const label = E("h3").cls("legend");
    label.innerText = this.label;
    this.el.insertBefore(label, this.wrap);
    const radio2 = E("div").cls("radio");
    this.options.forEach((o) => {
      const btn2 = E("input").on("change", () => {
        this.fireChange();
      });
      btn2.type = "radio";
      btn2.name = this.name || this.itemId;
      btn2.readOnly = this.readOnly;
      if (o.value) {
        btn2.value = o.value;
        if (this._value == o.value) {
          btn2.checked = true;
        }
        this.inputs[o.value] = btn2;
      }
      radio2.append(E(
        "label",
        btn2,
        E("span", o.text).cls("box-label")
      ).cls("control"));
    });
    if (this.invalidMsg) {
      this.applyInvalidMsg();
    }
    return radio2;
  }
  set value(v) {
    super.value = v;
    if (v && v in this.inputs)
      this.inputs[v].checked = true;
  }
  get value() {
    for (let v in this.inputs) {
      if (this.inputs[v].checked) {
        return v;
      }
    }
    return void 0;
  }
};
var radio = (config) => createComponent(new RadioField((config == null ? void 0 : config.type) || "box"), config);

// goui/script/component/form/RangeField.ts
var RangeField = class extends TextField {
  constructor() {
    super(...arguments);
    this.type = "range";
    this.decimals = 0;
    this.min = 0;
    this.max = 100;
  }
  validate() {
    super.validate();
    const v = super.value;
    if (v && isNaN(v)) {
      this.setInvalid("Incorrect number format");
    }
    if (this.max !== void 0 && v > this.max) {
      this.setInvalid("Number is too big");
    }
    if (this.min !== void 0 && v < this.min) {
      this.setInvalid("Number is too small");
    }
  }
  set value(v) {
    if (isNaN(v)) {
      throw new Error("Invalid number");
    } else if (v) {
      super.value = +v.toFixed(this.decimals);
    }
  }
  get value() {
    const v = super.value;
    return v === void 0 || isNaN(v) ? void 0 : +(+v).toFixed(this.decimals);
  }
  createControl() {
    super.createControl();
    this.el.cls("+range");
    if (this.min !== void 0) {
      this._input.attr("min", this.min);
    }
    if (this.max !== void 0) {
      this._input.attr("max", this.max);
    }
    if (this.decimals) {
      this._input.attr("step", "0.".padEnd(this.decimals + 1, "0") + "1");
    }
    return this._input;
  }
};
var rangefield = (config) => createComponent(new RangeField(), config);

// goui/script/component/form/RecurrenceField.ts
var RecurrenceField = class extends Field {
  constructor() {
    super();
    this.picker = new RecurrencePicker(new DateTime());
    this.buttons = [
      this.pickerButton = btn({
        icon: "expand_more",
        menu: menu(
          {
            alignTo: this.el,
            alignToInheritWidth: true
          },
          this.picker
        )
      })
    ];
  }
  createControl() {
    const input = E("input").attr("type", "text").attr("readOnly", true).cls("text");
    this.picker.on("select", (_, val) => {
      this.pickerButton.menu.hide();
      this.clearInvalid();
      this.focus();
      this.value = val;
      input.value = this.toText(val);
    });
    return input;
  }
  setStartDate(date) {
    this.picker.setStartDate(date);
  }
  toText(rule) {
    const rr = rule;
    if (!rr || !rr.frequency) {
      return t("Not recurring");
    }
    const record = RecurrencePicker.frequencies[rr.frequency];
    if (!record) {
      return "Unsupported frequency: " + rr.frequency;
    }
    let str = record[4];
    if (rr.interval) {
      str = t("Every") + " " + rr.interval + " " + record[rr.interval > 1 ? 1 : 0];
    }
    if (rr.byDay) {
      let days = [], workdays = rr.byDay.length === 5;
      for (var i = 0; i < rr.byDay.length; i++) {
        if (rr.byDay[i].day == "sa" || rr.byDay[i].day == "su") {
          workdays = false;
        }
        var nthDay = "";
        if (rr.byDay[i].nthOfPeriod) {
          nthDay = t("the") + " " + this.getSuffix(rr.byDay[i].nthOfPeriod) + " ";
        }
        days.push(nthDay + DateTime.dayNames[rr.byDay[i].day]);
      }
      if (workdays) {
        days = [t("Workdays")];
      }
      str += " " + t("at ") + days.join(", ");
    }
    if (rr.byMonthDay) {
      str += " " + t("at day") + " " + rr.byMonthDay.join(", ");
    }
    if (rr.count) {
      str += ", " + rr.count + " " + t("times");
    }
    if (rr.until) {
      str += ", " + t("until") + " " + new DateTime(rr.until).format("d-m-Y");
    }
    return str;
  }
  getSuffix(week) {
    switch (week) {
      case 1:
        return t("first");
      case 2:
        return t("second");
      case 3:
        return t("third");
      case 4:
        return t("fourth");
      default:
        return t("last");
    }
  }
};
var recurrencefield = (config) => createComponent(new RecurrenceField(), config);

// goui/script/component/form/TextareaField.ts
var TextAreaField = class extends TextField {
  constructor() {
    super(...arguments);
    this.baseCls = "goui-form-field textarea";
  }
  createControl() {
    const v = this.value;
    this._input = document.createElement("textarea");
    if (this.autocomplete) {
      this._input.autocomplete = this.autocomplete;
    }
    if (this.placeholder) {
      this._input.placeholder = this.placeholder;
    }
    this._input.required = this.required;
    this._input.name = this.name;
    if (v) {
      this._input.value = v;
    }
    return this._input;
  }
};
var textarea = (config) => createComponent(new TextAreaField(), config);

// goui/script/component/form/CheckboxGroup.ts
var CheckboxGroup = class extends Component {
  constructor() {
    super("fieldset");
    this.label = "";
    this.baseCls = "checkbox-group";
  }
  set options(options) {
    this.items.replace(...options.map((o) => {
      o.type = "button";
      return checkbox(o);
    }));
  }
  createLabel() {
    return;
  }
  renderItems() {
    const label = E("h3").cls("legend");
    label.innerText = this.label;
    this.el.appendChild(label);
    return super.renderItems();
  }
  get itemContainerEl() {
    if (!this._itemContainerEl) {
      this._itemContainerEl = E("div").cls("goui group");
      this.el.appendChild(this._itemContainerEl);
    }
    return this._itemContainerEl;
  }
};
var checkboxgroup = (config) => createComponent(new CheckboxGroup(), config);

// goui/script/component/form/ChipsField.ts
var ChipsField = class extends Field {
  constructor() {
    super(...arguments);
    this.baseCls = "goui-form-field chips";
    /**
     * Function that transforms the user text input to a chip.
     *
     * If it returns an empty value no chip will be rendered
     *
     * @param text
     */
    this.textInputToValue = (text) => __async(this, null, function* () {
      return text;
    });
    /**
     * Renders a value to the chip component
     * @param value
     */
    this.chipRenderer = (chip, value) => __async(this, null, function* () {
      chip.items.get(0).text = value;
    });
    this.selectedIndex = -1;
  }
  get editor() {
    return this._editor;
  }
  createControl() {
    this.items.on("datachanged", (collection) => {
      this.el.classList.toggle("has-chips", this.items.count() > 1);
    });
    this.chipsContainer = document.createElement("div");
    this.chipsContainer.classList.add("control-wrap");
    this.wrap.append(this.chipsContainer);
    this._editor = comp({
      id: Component.uniqueID(),
      attr: {
        contentEditable: "true"
      },
      cls: "editor"
    });
    this.el.setAttribute("for", this._editor.id);
    this._editor.el.addEventListener("keydown", (ev) => {
      this.onEditorKeyDown(ev);
    });
    this.el.addEventListener("click", (ev) => {
      if (this.selectedIndex == -1) {
        this._editor.focus();
      }
    });
    this._editor.el.addEventListener("focus", (ev) => {
      this.clearSelection();
    });
    this._editor.render(this.chipsContainer);
    this.items.add(this._editor);
    return void 0;
  }
  onEditorKeyDown(ev) {
    this.clearInvalid();
    switch (ev.key) {
      case "Enter":
        ev.preventDefault();
        const chip = this.createChip();
        this.textInputToValue(this._editor.text).then((value) => {
          if (!value) {
            return;
          }
          return this.chipRenderer(chip, value).then(() => {
            this.items.insert(-1, chip);
            this.value.push(value);
          });
        });
        this._editor.text = "";
        break;
      case "ArrowRight":
        if (this.selectedIndex > -1) {
          this.select(this.selectedIndex + 1);
        }
        break;
      case "ArrowLeft":
        if (this.selectedIndex == -1) {
          if (this._editor.html == "") {
            ev.preventDefault();
            this.select(this.items.count() - 2);
          }
        } else {
          this.select(this.selectedIndex - 1);
        }
        break;
      case "Delete":
      case "Backspace":
        ev.stopPropagation();
        if (this._editor.html == "") {
          if (this.selectedIndex == -1) {
            ev.preventDefault();
            this.select(this.items.count() - 2);
          } else {
            this.items.get(this.selectedIndex).remove();
            this.value.splice(this.selectedIndex, 1);
            this.select(this.selectedIndex);
          }
        }
        break;
    }
  }
  createChip() {
    const chip = comp(
      {
        cls: "chip hbox"
      },
      comp({}),
      btn({
        title: t("Remove"),
        icon: "cancel",
        handler: (btn2) => {
          this.captureValueForChange();
          const index = this.items.indexOf(chip);
          chip.remove();
          this.value.splice(index, 1);
          this.fireChange();
        }
      })
    );
    chip.el.addEventListener("click", () => {
      this.select(this.items.indexOf(chip));
    });
    return chip;
  }
  get itemContainerEl() {
    return this.chipsContainer;
  }
  clearSelection() {
    if (this.selectedIndex > -1) {
      const item = this.items.get(this.selectedIndex);
      if (item) {
        item.el.classList.remove("selected");
      }
      this.selectedIndex = -1;
    }
  }
  select(index) {
    if (index < 0) {
      return;
    }
    this.clearSelection();
    if (index > this.items.count() - 2) {
      this._editor.focus();
      return;
    }
    this.selectedIndex = index;
    this.items.get(this.selectedIndex).el.classList.add("selected");
  }
  internalSetValue(v, old) {
    if (this.rendered) {
      for (let i = 0; i < this.items.count() - 1; i++) {
        this.items.removeAt(i);
      }
      while (this.items.count() > 1) {
        this.items.removeAt(0);
      }
      this.renderValue();
    }
  }
  internalRender() {
    const el = super.internalRender();
    this.renderValue();
    return el;
  }
  renderValue() {
    if (!this.value) {
      return;
    }
    this.value.forEach((v) => {
      const chip = this.createChip();
      this.chipRenderer(chip, v).then(() => {
        this.items.insert(-1, chip);
      });
    });
  }
  focus(o) {
    this.editor.focus(o);
  }
};
var chips = (config) => createComponent(new ChipsField(), config);

// goui/script/component/DraggableComponent.ts
var DraggableComponent = class extends Component {
  constructor(tagName = "div") {
    super(tagName);
    this._dragConstrainTo = window;
    /**
     * Enable dragging
     */
    this._draggable = true;
    this.onDragHandleClick = (e) => {
      e.stopPropagation();
    };
    this.onDragHandleMouseDown = (e) => {
      const target = e.target;
      if (target != this.el && (target.tagName == "BUTTON" || target.closest("BUTTON"))) {
        return;
      }
      if (e.button != 0) {
        return;
      }
      e.preventDefault();
      this.focus();
      const el = this.el, rect = el.getBoundingClientRect();
      if (this.setPosition === void 0) {
        const cmpStyle = getComputedStyle(el);
        this.setPosition = cmpStyle.position == "absolute" || cmpStyle.position == "fixed";
      }
      this.dragData = {
        startOffsetLeft: el.offsetLeft,
        startOffsetTop: el.offsetTop,
        grabOffsetLeft: e.clientX - rect.x,
        grabOffsetTop: e.clientY - rect.y,
        x: e.clientX,
        y: e.clientY,
        startX: e.clientX,
        startY: e.clientY,
        data: {}
      };
      if (this.fire("dragstart", this, this.dragData, e) !== false) {
        this.onDragStart(e);
      }
    };
    if (this.baseCls) {
      this.baseCls += " draggable";
    } else {
      this.baseCls = "draggable";
    }
    this.on("render", () => {
      if (this._draggable) {
        this.toggleDraggable(this._draggable);
      }
    }, { once: true });
  }
  /**
   * Enable or disable dragging
   */
  set draggable(draggable2) {
    this._draggable = draggable2;
    if (this.rendered) {
      this.toggleDraggable(draggable2);
    }
  }
  toggleDraggable(draggable2) {
    const dragHandle = this.getDragHandle();
    if (!dragHandle) {
      return;
    }
    if (draggable2) {
      dragHandle.classList.add("goui-draghandle");
      dragHandle.addEventListener("click", this.onDragHandleClick);
      dragHandle.addEventListener("mousedown", this.onDragHandleMouseDown);
    } else {
      dragHandle.classList.remove("goui-draghandle");
      dragHandle.removeEventListener("click", this.onDragHandleClick);
      dragHandle.removeEventListener("mousedown", this.onDragHandleMouseDown);
    }
  }
  get draggable() {
    return this._draggable;
  }
  /**
   * Returns the DOM element that can be grabbed to drag the component
   * @protected
   */
  getDragHandle() {
    return this.el;
  }
  /**
   * Constrain dragging to this element
   * @param el
   * @param pad Supply paddings
   */
  dragConstrainTo(el, pad2) {
    this._dragConstrainTo = el;
    this._dragConstrainPad = pad2;
  }
  calcConstrainBox(el, pad2) {
    let box = {
      left: 0,
      right: 0,
      bottom: 0,
      top: 0
    };
    if (el instanceof Window) {
      box.right = window.innerWidth;
      box.bottom = window.innerHeight;
    } else {
      const rect = el.getBoundingClientRect();
      box.left = rect.left;
      box.right = rect.right;
      box.bottom = rect.bottom;
      box.top = rect.top;
    }
    if (pad2) {
      if (pad2.left)
        box.left += pad2.left;
      if (pad2.right)
        box.right -= pad2.right;
      if (pad2.top)
        box.top += pad2.top;
      if (pad2.bottom)
        box.bottom -= pad2.bottom;
    }
    return box;
  }
  onDragStart(e) {
    e.preventDefault();
    this.constrainBox = this.calcConstrainBox(this._dragConstrainTo, this._dragConstrainPad);
    const onDrag = FunctionUtil.onRepaint((e2) => {
      this.onDrag(e2);
    });
    document.addEventListener("mousemove", onDrag);
    document.addEventListener("mouseup", (e2) => {
      document.removeEventListener("mousemove", onDrag);
      this.fire("drop", this, this.dragData, e2);
    }, { once: true });
  }
  onDrag(e) {
    const d = this.dragData;
    d.x = e.clientX;
    d.y = e.clientY;
    this.constrainCoords();
    if (this.setPosition) {
      this.el.style.top = d.startOffsetTop + d.y - d.startY + "px";
      this.el.style.left = d.startOffsetLeft + d.x - d.startX + "px";
    }
    this.fire("drag", this, this.dragData, e);
  }
  constrainCoords() {
    if (!this.constrainBox) {
      return;
    }
    const maxTop = this.constrainBox.bottom - this.el.offsetHeight + this.dragData.grabOffsetTop;
    const maxLeft = this.constrainBox.right - this.el.offsetWidth + this.dragData.grabOffsetLeft;
    this.dragData.y = Math.max(this.constrainBox.top + this.dragData.grabOffsetTop, Math.min(this.dragData.y, maxTop));
    this.dragData.x = Math.max(this.constrainBox.left + this.dragData.grabOffsetLeft, Math.min(this.dragData.x, maxLeft));
    return;
  }
  /**
   * Constrain the component to the given element
   *
   * @param el
   * @param pad
   */
  constrainTo(el, pad2) {
    const constraints = this.calcConstrainBox(el, pad2);
    let maxTop = constraints.bottom - this.el.offsetHeight, maxLeft = constraints.right - this.el.offsetWidth, minTop = 0, minLeft = 0;
    if (this.el.offsetTop > maxTop) {
      console.warn("Contraining to top " + maxTop);
      this.el.style.top = maxTop + "px";
    } else if (this.el.offsetTop < minTop) {
      console.warn("Contraining to top " + minTop);
      this.el.style.top = minTop + "px";
    }
    if (this.el.offsetLeft > maxLeft) {
      console.warn("Contraining to left " + maxLeft);
      this.el.style.left = maxLeft + "px";
    } else if (this.el.offsetLeft < minLeft) {
      console.warn("Contraining to left " + minLeft);
      this.el.style.left = minLeft + "px";
    }
  }
};
var draggable = (config, ...items) => createComponent(new DraggableComponent(config == null ? void 0 : config.tagName), config, items);

// goui/script/component/Window.ts
var Window2 = class _Window extends DraggableComponent {
  constructor() {
    super();
    this.baseCls = "goui-window";
    /**
     * Maximize the window
     */
    this.maximized = false;
    /**
     * Enable tool to maximize window
     */
    this.maximizable = false;
    /**
     * Enable tool to close window
     */
    this.closable = true;
    /**
     * Enable tool to collapse window
     */
    this.collapsible = false;
    /**
     * Make the window modal so the user can only interact with this window.
     */
    this.modal = false;
    this.resizable = true;
    this.width = 400;
    this.hidden = true;
  }
  // /**
  //  * Focus first item if possible.
  //  * @param o
  //  */
  // public focus(o?: FocusOptions) {
  //
  // 	const first = this.items.first();
  // 	if(first) {
  // 		return first.focus(o);
  // 	} else {
  // 		return super.focus(o);
  // 	}
  // }
  initMaximizeTool() {
    const maximizeBtn = btn({
      icon: "fullscreen",
      title: t("Maximize"),
      handler: () => {
        this.isMaximized() ? this.unmaximize() : this.maximize();
      }
    });
    this.on("maximize", () => {
      maximizeBtn.icon = "fullscreen_exit";
      maximizeBtn.title = t("Restore");
    });
    this.on("unmaximize", () => {
      maximizeBtn.icon = "fullscreen";
      maximizeBtn.title = t("Maximize");
    });
    return maximizeBtn;
  }
  applyTitle() {
  }
  getDragHandle() {
    return this.getHeader().el;
  }
  getHeader() {
    if (!this.header) {
      this.header = tbar(
        {
          cls: "header"
        },
        this.titleCmp = comp({
          tagName: "h3",
          html: this.title
        }),
        "->"
      );
      if (this.maximizable) {
        this.header.items.add(this.initMaximizeTool());
      }
      if (this.collapsible) {
        this.header.items.add(btn({
          cls: "collapse-btn",
          icon: "",
          // set empty so 'collapsed class can set it class can set it
          handler: () => {
            this.collapsed = !this.collapsed;
          }
        }));
      }
      if (this.closable) {
        this.header.items.add(btn({
          icon: "close",
          handler: () => {
            this.close();
          }
        }));
      }
    }
    this.header.parent = this;
    return this.header;
  }
  set collapsed(collapsed) {
    this.el.classList.toggle("collapsed", collapsed);
  }
  get collapsed() {
    return this.el.classList.contains("collapsed");
  }
  internalRender() {
    const header = this.getHeader();
    header.render();
    const el = super.internalRender();
    this.on("drop", () => {
      this.saveState();
    });
    if (this.maximized) {
      this.maximize();
    }
    el.setAttribute("tabindex", "-1");
    this.el.addEventListener("keydown", (e) => {
      if (e.key == "Escape") {
        this.close();
      }
    });
    if (this.resizable) {
      this.observerResize();
    }
    if (this.modal) {
      el.classList.add("modal");
    }
    return el;
  }
  observerResize() {
    const saveState = FunctionUtil.buffer(200, () => {
      this.saveState();
    });
    let init = false;
    this.resizeObserver = new ResizeObserver((entries) => {
      if (init) {
        saveState();
      } else {
        init = true;
      }
    });
    this.resizeObserver.observe(this.el);
  }
  buildState() {
    if (this.isMaximized()) {
      const s = this.getState();
      s.maximized = true;
      return s;
    } else {
      return {
        width: this.el.offsetWidth,
        height: this.el.offsetHeight,
        left: this.el.offsetLeft,
        top: this.el.offsetTop
      };
    }
  }
  restoreState(s) {
    if (s.width)
      this.width = s.width;
    if (s.height)
      this.height = s.height;
    if (s.top != void 0) {
      this.el.style.top = s.top + "px";
      if (s.left != void 0) {
        this.el.style.left = s.left + "px";
      }
    } else {
      this.center();
    }
    if (s.maximized) {
      this.maximized = true;
    }
  }
  /**
   * Open the window by rendering it into the DOM body element
   * Use show()
   * @deprecated
   */
  open() {
    return this.show();
  }
  show() {
    this.focussedBeforeOpen = document.activeElement || void 0;
    let ret;
    if (!this.rendered) {
      root.items.add(this);
      if (this.modal) {
        this.modalOverlay = comp({
          cls: "goui-window-modal-overlay goui-goui-fade-in goui-goui-fade-out",
          hidden: true
        });
        this.modalOverlay.el.style.zIndex = parseInt(getComputedStyle(this.el).zIndex).toString();
        root.items.insert(-1, this.modalOverlay);
        this.disableBodyScroll();
        this.modalOverlay.el.addEventListener("click", (ev) => {
          this.focus();
        });
        this.modalOverlay.show();
      }
      ret = super.show();
      if (!this.hasState()) {
        this.shrinkToFit();
        this.center();
      } else {
        this.constrainTo(window);
      }
    } else {
      ret = super.show();
    }
    this.focus();
    return ret;
  }
  shrinkToFit() {
    if (this.el.offsetHeight > window.innerHeight) {
      this.el.style.height = window.innerHeight * 0.9 + "px";
    }
    if (this.el.offsetWidth > window.innerWidth) {
      this.el.style.width = window.innerWidth * 0.9 + "px";
    }
  }
  disableBodyScroll() {
    if (window.getComputedStyle(document.body).overflow != "hidden") {
      document.body.style.top = `-${window.scrollY}px`;
      document.body.style.position = "fixed";
    }
  }
  enableBodyScroll() {
    const scrollY = document.body.style.top, scrollTo = parseInt(scrollY || "0") * -1;
    document.body.style.position = "";
    document.body.style.top = "";
    document.documentElement.style.scrollBehavior = "auto";
    window.scrollTo({
      top: scrollTo,
      behavior: "auto"
    });
    document.documentElement.style.scrollBehavior = "";
  }
  /**
   * @inheritDoc
   */
  remove() {
    if (this.focussedBeforeOpen instanceof HTMLElement) {
      this.focussedBeforeOpen.focus();
    }
    if (this.resizeObserver) {
      this.resizeObserver.disconnect();
    }
    if (this.modalOverlay) {
      this.enableBodyScroll();
      this.modalOverlay.remove();
    }
    this.header.remove();
    return super.remove();
  }
  /**
   * Close the window by removing it
   */
  close() {
    this.fire("close", this);
    this.remove();
  }
  /**
   * Center the window in the screen
   */
  center() {
    this.el.style.top = (window.innerHeight - this.height) / 2 + "px";
    this.el.style.left = (window.innerWidth - this.width) / 2 + "px";
    return this;
  }
  /**
   * Returns true if the window is maximized
   */
  isMaximized() {
    return this.el.classList.contains("maximized");
  }
  /**
   * Grow window to the maximum of the viewport
   */
  maximize() {
    this.collapsed = false;
    this.el.classList.add("maximized");
    this.fire("maximize", this);
    return this;
  }
  /**
   * Make the window smaller than the viewport
   */
  unmaximize() {
    this.el.classList.remove("maximized");
    this.fire("unmaximize", this);
    return this;
  }
  static error(msg) {
    return _Window.alert(msg, t("Error") + " - " + new DateTime().format("Y-m-d H:i:s"));
  }
  /**
   * Show modal alert window
   *
   */
  static alert(text, title = t("Alert")) {
    if (text.message) {
      console.error(text);
      text = text.message;
    }
    return new Promise((resolve, reject) => {
      win(
        {
          modal: true,
          title,
          listeners: {
            close: () => {
              resolve();
            }
          }
        },
        comp({
          flex: 1,
          cls: "scroll pad",
          html: text
        })
      ).show();
    });
  }
  /**
   * Prompt the user for a text input value.
   *
   * Returns a promise with the input or undefined if the user cancelled
   *
   */
  static prompt(text, inputLabel, defaultValue = "", title = t("Please enter")) {
    return new Promise((resolve, reject) => {
      let cancelled = true;
      const w = win(
        {
          modal: true,
          title,
          listeners: {
            focus: () => {
              w.items.get(0).focus();
            },
            close: () => {
              if (cancelled) {
                resolve(void 0);
              }
            }
          }
        },
        form(
          {
            handler: (form2) => {
              resolve(form2.getValues()["input"]);
              cancelled = false;
              w.close();
            }
          },
          fieldset(
            {},
            comp({
              tagName: "p",
              html: text
            }),
            textfield({
              label: inputLabel,
              name: "input",
              required: true,
              value: defaultValue
            })
          ),
          tbar(
            {},
            comp({
              flex: 1
            }),
            btn({
              type: "submit",
              text: "Ok"
            })
          )
        )
      );
      w.show();
    });
  }
  /**
   * Ask the user for confirmation
   */
  static confirm(text, title = t("Please confirm")) {
    return new Promise((resolve, reject) => {
      const yesBtn = btn({
        text: t("Yes"),
        cls: "filled primary",
        handler: () => {
          resolve(true);
          w.close();
        }
      }), w = win(
        {
          modal: true,
          title,
          closable: false,
          listeners: {
            focus: () => {
              yesBtn.focus();
            }
          }
        },
        comp({
          cls: "pad",
          html: text
        }),
        tbar(
          {},
          "->",
          btn({
            text: t("No"),
            handler: () => {
              resolve(false);
              w.close();
            }
          }),
          yesBtn
        )
      );
      w.show();
    });
  }
};
var win = (config, ...items) => createComponent(new Window2(), config, items);

// goui/script/component/form/DataSourceForm.ts
var DataSourceForm = class extends Form {
  constructor(dataSource) {
    super();
    this.dataSource = dataSource;
    this.handler = (form1) => __async(this, null, function* () {
      var _a;
      try {
        this._value = {};
        let v = this.value;
        if (this.currentId) {
          v.id = this.currentId;
        }
        this.fire("beforesave", this, v);
        let data;
        if (this.currentId) {
          data = yield this.dataSource.update(v);
        } else {
          data = yield this.dataSource.create(v);
        }
        if (data) {
          this.fire("save", this, data);
        }
      } catch (e) {
        if (e.type == "invalidProperties") {
          this.handleServerValidation(e);
          return;
        }
        console.log(t("Error"), e);
        if (this.fire("saveerror", this, e) !== false) {
          void Window2.error((_a = e.message) != null ? _a : t("Unknown error"));
        }
      } finally {
        this.unmask();
      }
    });
  }
  handleServerValidation(error) {
    for (const propertyName in error.validationErrors) {
      const field = this.findField(propertyName);
      if (!field) {
        continue;
      }
      field.setInvalid(error.validationErrors[propertyName].description);
    }
    const invalid = this.findFirstInvalid();
    if (invalid) {
      invalid.focus();
    }
    this.setInvalid(t("You have errors in your form. The invalid fields are marked."));
  }
  create(data) {
    this.reset();
    this.currentId = "_new_";
    this.fire("load", this, data);
    if (data) {
      this.setValues(data);
    }
  }
  reset() {
    super.reset();
    delete this.currentId;
  }
  /**
   * Load an entity into the form
   *
   * @param id
   */
  load(id) {
    return __async(this, null, function* () {
      this.mask();
      try {
        this.currentId = id;
        let entity = yield this.dataSource.single(id);
        if (!entity) {
          throw "Failed to load entity with id " + id;
        }
        this.fire("load", this, entity);
        this.value = entity;
      } catch (e) {
        alert(t("Error") + " " + e);
        console.log(t("Error"), e);
        if (this.fire("loaderror", this, e) !== false) {
          void Window2.error(e.message);
        }
      } finally {
        this.unmask();
      }
    });
  }
};
var datasourceform = (config, ...items) => createComponent(new DataSourceForm(config.dataSource), config, items);

// goui/script/component/form/AutocompleteChips.ts
var AutocompleteChips = class extends ChipsField {
  /**
   * Constructor
   *
   * @param list The drop down list or {@see Table}.
   * @param buffer
   */
  constructor(list2, buffer = 300) {
    super();
    this.list = list2;
    this.buffer = buffer;
    //disable create
    this.textInputToValue = (text) => __async(this, null, function* () {
      return false;
    });
    this.menu = menu(
      {
        cls: "goui-dropdown scroll",
        removeOnClose: false,
        height: 300,
        listeners: {
          hide: this.onMenuHide.bind(this)
        }
      },
      list2
    );
    this.menuButton = btn({
      icon: "expand_more",
      type: "button",
      handler: (button, ev) => {
        this.fire("autocomplete", this, "");
      },
      menu: this.menu
    });
    this.initList();
  }
  initList() {
    this.list.el.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "Enter":
          ev.preventDefault();
          this.menu.hide();
          break;
      }
    });
    if (!this.list.rowSelection) {
      this.list.rowSelectionConfig = { multiSelect: false };
    }
    if (!this.list.rowSelection.multiSelect) {
      this.list.on("rowclick", () => {
        this.menu.hide();
      });
      this.list.store.on("datachanged", () => {
        this.list.rowSelection.selected = [0];
      });
    } else {
      this.list.store.on("datachanged", () => {
        this.list.rowSelection.clear();
        this.list.store.data.forEach((record, index) => {
          if (this.isPickerRecordInValue(record)) {
            this.list.rowSelection.add(index);
          }
        });
      });
    }
  }
  isPickerRecordInValue(record) {
    if (!this.value || !this.value.length) {
      return false;
    }
    if (!this.valuesToCompare) {
      this.valuesToCompare = this.value.map((i) => JSON.stringify(i));
      setTimeout(() => {
        this.valuesToCompare = void 0;
      });
    }
    const v = JSON.stringify(this.pickerRecordToValue(this, record));
    return this.valuesToCompare.indexOf(v) > -1;
  }
  /**
   * Method that transforms a record from the TablePicker store to a value for this field.
   * This is not necessarily a text value. In conjunction with {@see valueToTextField()} this
   * could also be an ID of an object for example.
   *
   * @param field
   * @param record
   */
  pickerRecordToValue(field, record) {
    return record.id;
  }
  internalRender() {
    this.buttons = this.buttons || [];
    this.buttons.push(this.menuButton);
    const el = super.internalRender();
    this.menu.alignTo = this.wrap;
    this.menu.alignToInheritWidth = true;
    this.editor.el.addEventListener("input", FunctionUtil.buffer(this.buffer, this.onInput.bind(this)));
    this.editor.el.addEventListener("keydown", (ev) => {
      switch (ev.key) {
        case "Enter":
          if (!this.menu.hidden) {
            ev.preventDefault();
            this.menu.hide();
          }
          break;
        case "ArrowDown":
          ev.preventDefault();
          this.menuButton.showMenu();
          this.list.focus();
          break;
        case "Escape":
          if (!this.menu.hidden) {
            this.menu.hide();
            ev.preventDefault();
            ev.stopPropagation();
            this.focus();
          }
          break;
      }
    });
    return el;
  }
  onInput(ev) {
    console.warn("onInput");
    this.menuButton.showMenu();
    this.fire("autocomplete", this, this.editor.el.innerText);
  }
  onMenuHide() {
    if (!this.menu.rendered) {
      return;
    }
    const keys = {};
    const newValues = this.list.rowSelection.selected.map((index) => {
      const record = this.list.store.get(index);
      return this.pickerRecordToValue(this, record);
    });
    this.list.rowSelection.clear();
    this.editor.el.innerText = "";
    this.focus();
    if (this.list.rowSelection.multiSelect) {
      this.value = newValues;
    } else {
      if (!this.value) {
        this.value = [];
      }
      this.value = this.value.concat(newValues);
    }
  }
};
var autocompletechips = (config) => createComponent(new AutocompleteChips(config.list), config);

// goui/script/data/Store.ts
var Store = class extends Collection {
  constructor() {
    super(...arguments);
    // private static stores: Record<string, Store> = {};
    //
    // /**
    //  * If the ID is given the store can be fetched with
    //  */
    // public id?: string;
    this._loading = false;
    /**
     * Sort the data on field and direction
     */
    this.sort = [];
  }
  /**
   * True when the store is loading
   */
  get loading() {
    return this._loading;
  }
  /**
   * Load a set of records
   *
   * @param records
   * @param append
   */
  loadData(records, append = true) {
    append ? this.add(...records) : this.replace(...records);
    this.fire("load", this, records, append);
  }
  set data(records) {
    this.loadData(records);
  }
  get data() {
    return this.getArray();
  }
  /**
   * Reload the data from the source
   */
  reload() {
    return __async(this, null, function* () {
      return this.load();
    });
  }
  /**
   * Returns the loaded records. If append is true it only returns the new records.
   * Override this function for new store types.
   *
   * @param append
   * @protected
   */
  internalLoad(append) {
    this.loadData(ArrayUtil.multiSort(this.items, this.sort), append);
    return Promise.resolve(this.items);
  }
  /**
   * Loads the data from source into the store
   *
   * @param append
   */
  load(append = false) {
    this._loading = true;
    this.fire("beforeload", this, append);
    return this.internalLoad(append).finally(() => {
      this._loading = false;
    });
  }
  /**
   * Load the next set of records when paging.
   * Doesn't do anything in the array store but can be implemented in async stores.
   */
  loadNext(append = false) {
    return Promise.resolve([]);
  }
  /**
   * Load the next set of records when paging.
   * Doesn't do anything in the array store but can be implemented in async stores.
   */
  loadPrevious() {
    return Promise.resolve([]);
  }
  hasNext() {
    return false;
  }
  hasPrevious() {
    return false;
  }
};
var store = (config) => createComponent(new Store(), config);

// goui/script/data/DataSourceStore.ts
var DataSourceStore = class extends Store {
  constructor(dataSource) {
    super();
    this.dataSource = dataSource;
    this.queryParams = {};
    this.hasMore = false;
    // public properties?: string[] = [];
    /**
     * True when loaded at least once.
     */
    this.loaded = false;
    /**
     * Reload when the datasource changes
     */
    this.monitorChanges = true;
    /**
     * Builds record from entity
     * @param entity
     */
    this.buildRecord = (entity) => __async(this, null, function* () {
      return entity;
    });
    this.dataSource.on("change", (dataSource2, changes) => {
      this.onChange(dataSource2, changes);
    });
  }
  /**
   * Reloads the store when the datasource changes
   *
   * @protected
   */
  onChange(DataSource, changes) {
    return __async(this, null, function* () {
      if (this.loaded && this.monitorChanges) {
        void this.reload();
      }
    });
  }
  internalLoad(append = false) {
    return __async(this, null, function* () {
      const queryParams = structuredClone(this.queryParams);
      queryParams.sort = this.sort;
      if (queryParams.limit) {
        queryParams.limit++;
      }
      const queryResponse = yield this.dataSource.query(queryParams);
      if (this.queryParams.limit) {
        this.hasMore = queryResponse.ids.length > this.queryParams.limit;
        if (this.hasMore) {
          queryResponse.ids.pop();
        }
      }
      const getResponse = yield this.dataSource.get(queryResponse.ids);
      const entities = yield this.fetchRelations(getResponse.list), records = yield Promise.all(entities.map(this.buildRecord));
      this.loadData(records, append);
      this.loaded = true;
      return records;
    });
  }
  fetchRelations(records) {
    return __async(this, null, function* () {
      if (!this.relations) {
        return records;
      }
      let relationName;
      const promises = [];
      for (relationName in this.relations) {
        const rel = this.relations[relationName];
        let id;
        for (let i = 0, l = records.length; i < l; i++) {
          id = ObjectUtil.path(records[i], rel.path);
          if (id) {
            promises.push(rel.dataSource.single(id).then((e) => {
              if (e) {
                records[i][relationName] = e;
              }
            }));
          }
        }
      }
      return Promise.all(promises).then(() => records);
    });
  }
  reload() {
    this.queryParams.position = 0;
    return super.reload();
  }
  loadNext(append = false) {
    if (!this.queryParams.limit) {
      throw new Error("Limit must be set for pagination");
    }
    this.queryParams.position = this.queryParams.position || 0;
    this.queryParams.position += this.queryParams.limit;
    return this.load(append);
  }
  loadPrevious() {
    if (!this.queryParams.limit) {
      throw new Error("Limit and position must be set!");
    }
    this.queryParams.position = this.queryParams.position || 0;
    this.queryParams.position -= this.queryParams.limit;
    return this.load(false);
  }
  hasNext() {
    return this.hasMore;
  }
  hasPrevious() {
    return this.queryParams.position > 0;
  }
  /**
   * Load more data when this element is scrolled down
   *
   * @param el
   */
  addScrollLoader(el) {
    const onScroll = () => {
      const pixelsLeft = el.scrollHeight - el.scrollTop - el.offsetHeight;
      if (pixelsLeft < 100) {
        if (!this.loading && this.hasNext()) {
          void this.loadNext(true);
        }
      }
    };
    el.addEventListener("scroll", onScroll, { passive: true });
    this.on("load", (store2, records, append) => {
      setTimeout(() => {
        onScroll();
      });
    });
    return onScroll;
  }
};
var datasourcestore = (config) => createComponent(new DataSourceStore(config.dataSource), config);

// goui/script/data/AbstractDataSource.ts
var CommitErrorType = /* @__PURE__ */ ((CommitErrorType2) => {
  CommitErrorType2[CommitErrorType2["forbidden"] = 0] = "forbidden";
  CommitErrorType2[CommitErrorType2["overQuota"] = 1] = "overQuota";
  CommitErrorType2[CommitErrorType2["tooLarge"] = 2] = "tooLarge";
  CommitErrorType2[CommitErrorType2["rateLimit"] = 3] = "rateLimit";
  CommitErrorType2[CommitErrorType2["notFound"] = 4] = "notFound";
  CommitErrorType2[CommitErrorType2["invalidPatch"] = 5] = "invalidPatch";
  CommitErrorType2[CommitErrorType2["willDestroy"] = 6] = "willDestroy";
  CommitErrorType2[CommitErrorType2["invalidProperties"] = 7] = "invalidProperties";
  CommitErrorType2[CommitErrorType2["singleton"] = 8] = "singleton";
  CommitErrorType2[CommitErrorType2["requestTooLarge"] = 9] = "requestTooLarge";
  CommitErrorType2[CommitErrorType2["stateMismatch"] = 10] = "stateMismatch";
  return CommitErrorType2;
})(CommitErrorType || {});
var AbstractDataSource = class extends Observable {
  constructor(id) {
    super();
    this.id = id;
    /**
     * Store data in the browser storage so it will persist across sessions
     */
    this.persist = true;
    /**
     * Extra parameters to send to the Foo/set
     */
    this.commitBaseParams = {};
    this.data = {};
    this.creates = {};
    this.updates = {};
    this.destroys = {};
    this.getIds = {};
    this._createId = 0;
    this.delayedCommit = FunctionUtil.buffer(0, () => {
      void this.commit();
    });
    this.delayedGet = FunctionUtil.buffer(0, () => {
      void this.doGet();
    });
  }
  /**
   * Get the local server state ID of the store
   * @protected
   */
  getState() {
    return __async(this, null, function* () {
      if (!this._state && this.persist) {
        this._state = yield this.browserStore.getItem("__state__");
      }
      return this._state;
    });
  }
  /**
   * Set's the local server state ID
   *
   * Setting it to undefined will reset the store.
   *
   * @param state
   * @protected
   */
  setState(state) {
    return __async(this, null, function* () {
      this._state = state;
      console.warn("set state " + this.id + ": " + state);
      if (!this.persist) {
        return;
      }
      return this.browserStore.setItem("__state__", state);
    });
  }
  clearCache() {
    this.data = {};
    return this.browserStore.clear();
  }
  /**
   * Get the browser storage object to save state to the browser
   * @private
   */
  get browserStore() {
    if (!this._browserStore) {
      this._browserStore = new BrowserStore("ds-" + this.id);
    }
    return this._browserStore;
  }
  /**
   * Get entities from the store
   *
   * It will return a list of entities ordered by the requested ID's
   *
   * @param ids
   */
  get(ids) {
    return __async(this, null, function* () {
      const promises = [], order = {};
      if (ids == void 0) {
        ids = (yield this.query()).ids;
      }
      ids.forEach((id, index) => {
        order[id] = index++;
        promises.push(this.single(id));
      });
      let entities = yield Promise.all(promises);
      const response = {
        list: [],
        notFound: [],
        state: yield this.getState()
      };
      entities.forEach((e, index) => {
        if (e === void 0) {
          response.notFound.push(ids[index]);
        } else {
          response.list.push(e);
        }
      });
      response.list = response.list.sort(function(a, b) {
        return order[a.id] - order[b.id];
      });
      return response;
    });
  }
  add(data) {
    console.debug("Adding " + this.id + ": " + data.id);
    this.data[data.id] = data;
    if (!this.persist) {
      return Promise.resolve(data);
    }
    return this.browserStore.setItem(data.id, data).then(() => data);
  }
  remove(id) {
    console.debug("Removing " + this.id + ": " + id);
    delete this.data[id];
    if (!this.persist) {
      return Promise.resolve(id);
    }
    return this.browserStore.removeItem(id).then(() => id);
  }
  /**
   * Get a single entity.
   *
   * Multiple calls will be buffered and returned together on the next event loop. This way multiple calls can
   * be joined together in a single HTTP request to the server.
   *
   * @param id
   */
  single(id) {
    return __async(this, null, function* () {
      const p2 = new Promise((resolve, reject) => {
        if (!this.getIds[id]) {
          this.getIds[id] = {
            resolves: [resolve],
            rejects: [reject]
          };
        } else {
          this.getIds[id].resolves.push(resolve);
          this.getIds[id].rejects.push(reject);
        }
      });
      this.delayedGet();
      return p2;
    });
  }
  returnGet(id) {
    let r;
    while (r = this.getIds[id].resolves.shift()) {
      r.call(this, structuredClone(this.data[id]));
    }
  }
  /**
   * Does the actual getting of entities. First checks if present in this onbject, otherwise it will be requested
   * from the remote source.
   *
   * @protected
   */
  doGet() {
    return __async(this, null, function* () {
      const unknownIds = [];
      for (let id in this.getIds) {
        if (this.data[id]) {
          this.returnGet(id);
        } else if (this.persist) {
          const data = yield this.browserStore.getItem(id);
          if (data) {
            this.data[id] = data;
            this.returnGet(id);
          } else {
            unknownIds.push(id);
          }
        } else {
          unknownIds.push(id);
        }
      }
      if (!unknownIds.length) {
        return;
      }
      this.internalGet(unknownIds).then((response) => this.checkState(response.state, response)).then((response) => {
        var _a;
        response.list.forEach((e) => {
          this.add(e);
          this.returnGet(e.id);
        });
        (_a = response.notFound) == null ? void 0 : _a.forEach((id) => {
          let r;
          while (r = this.getIds[id].resolves.shift()) {
            r.call(this, void 0);
          }
        });
      });
    });
  }
  /**
   * Create entity
   *
   * Multiple calls will be joined together in a single call on the next event loop
   *
   * @param data
   * @param createId The create ID to use when committing this entity to the server
   */
  create(data, createId) {
    if (createId === void 0) {
      createId = this.createID();
    }
    const p2 = new Promise((resolve, reject) => {
      this.creates[createId] = {
        data,
        resolve,
        reject
      };
    }).finally(() => {
      delete this.creates[createId];
    });
    this.delayedCommit();
    return p2;
  }
  /**
   * Reset the data source.
   *
   * Clears all data and will resync
   */
  reset() {
    return __async(this, null, function* () {
      return this.setState(void 0);
    });
  }
  /**
   * Update an entity
   *
   * Multiple calls will be joined together in a single call on the next event loop
   *
   * @param data
   */
  update(data) {
    const p2 = new Promise((resolve, reject) => {
      this.updates[data.id] = {
        data,
        resolve,
        reject
      };
    }).finally(() => {
      delete this.updates[data.id];
    });
    this.delayedCommit();
    return p2;
  }
  createID() {
    return "_new_" + ++this._createId;
  }
  /**
   * Destroy an entity
   *
   * Multiple calls will be joined together in a single call on the next event loop
   *
   * @param id
   */
  destroy(id) {
    const p2 = new Promise((resolve, reject) => {
      this.destroys[id] = {
        resolve,
        reject
      };
    }).finally(() => {
      delete this.destroys[id];
    });
    this.delayedCommit();
    return p2;
  }
  /**
   * Fetch updates from remote
   */
  updateFromServer() {
    return __async(this, null, function* () {
      let hasMoreChanges = true, hasAChange = false;
      const allChanges = {
        created: [],
        updated: [],
        destroyed: [],
        oldState: "",
        newState: ""
      }, promises = [];
      try {
        while (hasMoreChanges) {
          const state = yield this.getState();
          if (state === void 0) {
            return;
          }
          if (!allChanges.oldState) {
            allChanges.oldState = state;
          }
          const changes = yield this.internalRemoteChanges(state);
          if (changes.created) {
            for (let id of changes.created) {
              promises.push(this.remove(id));
              allChanges.created.push(id);
              hasAChange = true;
            }
          }
          if (changes.updated) {
            for (let id of changes.updated) {
              promises.push(this.remove(id));
              allChanges.updated.push(id);
              hasAChange = true;
            }
          }
          if (changes.destroyed) {
            for (let id of changes.destroyed) {
              promises.push(this.remove(id));
              allChanges.destroyed.push(id);
              hasAChange = true;
            }
          }
          yield Promise.all(promises);
          yield this.setState(changes.newState);
          allChanges.newState = changes.newState;
          hasMoreChanges = !!changes.hasMoreChanges;
        }
      } catch (e) {
        console.error(this.id + " Error while updating from server. Resetting data source.");
        console.error(e);
        yield this.reset();
      }
      if (hasAChange) {
        this.fire("change", this, allChanges);
      }
    });
  }
  /**
   * Commit pending changes to remote
   */
  commit() {
    return __async(this, null, function* () {
      const params = Object.assign({
        create: {},
        update: {},
        destroy: [],
        ifInState: yield this.getState()
      }, this.commitBaseParams);
      for (let id in this.creates) {
        params.create[id] = this.creates[id].data;
      }
      for (let id in this.updates) {
        params.update[id] = this.updates[id].data;
      }
      for (let id in this.destroys) {
        params.destroy.push(id);
      }
      this.internalCommit(params).then((response) => __async(this, null, function* () {
        if (response.created) {
          for (let clientId in response.created) {
            let data = Object.assign(params.create ? params.create[clientId] || {} : {}, response.created[clientId] || {});
            this.add(data).then(() => this.creates[clientId].resolve(data));
          }
        }
        if (response.notCreated) {
          for (let clientId in response.notCreated) {
            this.creates[clientId].reject(response.notCreated[clientId]);
          }
        }
        if (response.updated) {
          for (let serverId in response.updated) {
            if (!this.data[serverId]) {
              continue;
            }
            let data = params.update && params.update[serverId] ? Object.assign(this.data[serverId], params.update[serverId]) : this.data[serverId];
            data = Object.assign(data, response.updated[serverId] || {});
            this.add(data).then((data2) => this.updates[serverId].resolve(data2));
          }
        }
        if (response.notUpdated) {
          for (let serverId in response.notUpdated) {
            this.updates[serverId].reject(response.notUpdated[serverId]);
          }
        }
        if (response.destroyed) {
          for (let i = 0, l = response.destroyed.length; i < l; i++) {
            this.remove(response.destroyed[i]).then((id) => this.destroys[id].resolve(id));
          }
        }
        if (response.notDestroyed) {
          for (let serverId in response.notDestroyed) {
            this.destroys[serverId].reject(response.notDestroyed[serverId]);
          }
        }
        yield this.setState(response.newState);
        this.fire("change", this, {
          created: response.created ? Object.keys(response.created) : [],
          updated: response.updated ? Object.keys(response.updated) : [],
          destroyed: response.destroyed || [],
          oldState: response.oldState,
          newState: response.newState
        });
      })).catch((e) => {
        for (let clientId in this.creates) {
          this.creates[clientId].reject(e);
        }
        for (let clientId in this.updates) {
          this.updates[clientId].reject(e);
        }
        for (let clientId in this.destroys) {
          this.destroys[clientId].reject(e);
        }
      }).finally(() => {
        this.creates = {};
        this.updates = {};
        this.destroys = {};
      });
    });
  }
  /**
   * Query the server for a list of entity ID's
   *
   * It takes filters and sort parameters.
   *
   * @link https://jmap.io/spec-core.html#query
   */
  query(params = {}) {
    return this.internalQuery(params).then((r) => {
      return this.checkState(r.queryState, r);
    });
  }
  /**
   * Check's if we are up-to-date with the server and fetches updates if needed.
   *
   * If no state is returned by the data source this function will ignore states and the data source should then
   * always refresh data.
   *
   * @param serverState
   * @param retVal
   * @private
   */
  checkState(serverState, retVal) {
    return __async(this, null, function* () {
      let state = yield this.getState();
      if (!state && serverState) {
        yield this.setState(serverState);
        state = serverState;
      }
      if (serverState != state) {
        return this.updateFromServer().then(() => retVal);
      } else {
        return Promise.resolve(retVal);
      }
    });
  }
};

// goui/script/data/RestDataSource.ts
var RestDataSource = class extends AbstractDataSource {
  /**
   * Constructor
   *
   * @param uri The base URI of the REST API. for example https://groupoffice.com/api
   * @param id The Data source ID. Will be appended to the base uri above. for Example
   *  if the ID is "users" the uri will be: "https://groupoffice.com/api/users"
   */
  constructor(uri, id) {
    id = id || "restDataSource";
    super(id);
    this.uri = uri;
    if (uri.substring(-1, 1) != "/") {
      this.uri = uri + "/";
    }
  }
  read(id, path = "", options = {}) {
    if (typeof id === "number") {
      path += "/" + id;
    } else {
      if (path != "" && path.substring(-1, 1) != "/") {
      }
    }
    return this.request(path, options);
  }
  doRequest(path = "", options = {}) {
    return this.request(path, options);
  }
  request(path = "", options = {}) {
    const baseOpts = {
      mod: "cors",
      method: "GET"
    };
    Object.assign(baseOpts, options);
    if (path != "" && path.substring(0, 1) != "") {
      path = "/" + path;
    }
    return fetch(this.uri + path, baseOpts).then((response) => {
      if (response.status != 200) {
        throw response.statusText;
      }
      return response.json();
    });
  }
  internalCommit(params) {
    const response = {
      updated: {},
      created: {},
      destroyed: []
    }, promises = [];
    for (let id in params.update) {
      promises.push(
        this.request(id, {
          method: "PUT",
          body: JSON.stringify(params.update[id])
        }).then((data) => {
          if (!response.updated) {
            response.updated = {};
          }
          response.updated[id] = data;
        }).catch((reason) => {
          if (!response.notUpdated) {
            response.notUpdated = {};
          }
          response.notUpdated[id] = reason;
        })
      );
    }
    for (let id in params.create) {
      promises.push(
        this.request("", {
          method: "POST",
          body: JSON.stringify(params.create[id])
        }).then((data) => {
          if (!response.created) {
            response.created = {};
          }
          response.created[id] = data;
        }).catch((reason) => {
          if (!response.notCreated) {
            response.notCreated = {};
          }
          response.notCreated[id] = reason;
        })
      );
    }
    for (let id of params.destroy) {
      promises.push(
        this.request(id, {
          method: "DELETE"
        }).then(() => {
          if (!response.destroyed) {
            response.destroyed = [];
          }
          response.destroyed.push(id);
        }).catch((reason) => {
          if (!response.notDestroyed) {
            response.notDestroyed = {};
          }
          response.notDestroyed[id] = reason;
        })
      );
    }
    return Promise.all(promises).then(() => {
      return response;
    });
  }
  internalGet(ids) {
    return __async(this, null, function* () {
      const promises = [];
      ids.forEach((id) => {
        promises.push(
          this.request(id).then((data) => {
            return data.data;
          })
        );
      });
      const users = yield Promise.all(promises);
      const r = {
        list: []
      };
      users.forEach((u) => {
        r.list.push(u);
      });
      return r;
    });
  }
  internalQuery(params) {
    return __async(this, null, function* () {
      const response = yield this.request();
      response.data.forEach((r) => {
        if (!this.data[r.id]) {
          this.add(r);
        }
      });
      const ids = response.data.map((r) => r.id);
      return {
        ids,
        total: response.total
      };
    });
  }
  /**
   * No sync implemented for REST. States are ignored and data must always be refreshed.
   * @param state
   * @protected
   */
  internalRemoteChanges(state) {
    return __async(this, null, function* () {
      return {};
    });
  }
};

// goui/script/component/table/MultiSelectToolbar.ts
var MultiSelectToolbar = class extends Toolbar {
  constructor(table2) {
    super();
    this.table = table2;
    this.table.rowSelection.on("selectionchange", (tableRowSelect) => {
      const l = tableRowSelect.selected.length;
      this.hidden = l < 2;
      if (!this.hidden) {
        this.label.text = l + " " + t("selected");
      }
    });
    this.label = comp({ tagName: "h3" });
    this.backBtn = btn({
      title: t("Back"),
      icon: "chevron_left",
      handler: () => {
        this.table.rowSelection.clear();
      }
    });
    this.items.add(this.backBtn, this.label);
    this.hidden = true;
    this.cls = "multiselect";
  }
};
var mstbar = (config, ...items) => {
  const c = new MultiSelectToolbar(config.table);
  if (config) {
    Object.assign(c, config);
  }
  c.items.add(...tbarItems(items));
  return c;
};

// goui/script/component/table/RowSelect.ts
var RowSelect = class extends Observable {
  constructor(list2) {
    super();
    this.list = list2;
    this._selected = [];
    /**
     * Last selected index used for multi selection with shift
     * @private
     */
    this.lastIndex = -1;
    this.multiSelect = true;
    this.hasKeyUpListener = false;
    this.list.on("beforerender", (me) => {
      const tableEl = me.el;
      tableEl.cls("+rowselect");
      tableEl.addEventListener("keydown", (e) => {
        this.onKeyDown(e);
      });
      me.on("rowmousedown", (table2, index, row, e) => {
        this.onRowMouseDown(table2, index, e);
      });
      tableEl.addEventListener("focus", (e) => {
        if (!this.selected.length && this.list.store.get(0)) {
          this.selected = [0];
        }
      });
    });
  }
  clear() {
    this.selected = [];
  }
  selectAll() {
    const selected = [];
    for (let i = 0, c = this.list.store.count(); i < c; i++) {
      selected.push(i);
    }
    this.selected = selected;
  }
  /**
   * Get selected indexes
   *
   * Note that this is a copy and can't be edited directly. You have to set the selected property with a changed array.
   */
  get selected() {
    return [...this._selected];
  }
  set selected(newSelection) {
    this.setSelected(newSelection);
  }
  onRowMouseDown(_list, index, e) {
    let selection = this.selected;
    if (e.shiftKey && this.multiSelect) {
      const start = Math.min(index, this.lastIndex);
      const end = Math.max(index, this.lastIndex);
      for (let i = start; i <= end; i++) {
        if (selection.indexOf(i) === -1)
          selection.push(i);
      }
    } else if ((e.ctrlKey || e.metaKey) && this.multiSelect) {
      const currentIndex = selection.indexOf(index);
      if (currentIndex > -1) {
        selection.splice(currentIndex, 1);
      } else {
        selection.push(index);
      }
    } else {
      selection = [index];
    }
    this.selected = selection;
  }
  add(index) {
    if (this._selected.indexOf(index) > -1) {
      return;
    }
    this._selected.push(index);
    this.fire("rowselect", this, index);
    this.fire("selectionchange", this);
  }
  /**
   * Remove an index from the selection
   *
   * @param index
   * @param silent Don't fire events
   */
  remove(index, silent = false) {
    const selectedIndex = this._selected.indexOf(index);
    if (selectedIndex == -1) {
      return;
    }
    this._selected.splice(selectedIndex, 1);
    if (!silent) {
      this.fire("rowdeselect", this, index);
      this.fire("selectionchange", this);
    }
  }
  /**
   * Set selected indexes
   *
   * @param newSelection
   * @param silent Suspends 'selectionchange' event
   */
  setSelected(newSelection, silent = false) {
    const old = this._selected;
    this._selected = newSelection;
    const deselect = ArrayUtil.diff(old, newSelection);
    const select2 = ArrayUtil.diff(newSelection, old);
    deselect.forEach((i) => {
      this.fire("rowdeselect", this, i);
    });
    select2.forEach((i, index) => {
      this.fire("rowselect", this, i);
    });
    if (newSelection.length) {
      this.lastIndex = newSelection[0];
    }
    const change = select2.length > 0 || deselect.length > 0;
    if (!silent && change) {
      this.fire("selectionchange", this);
    }
    return change;
  }
  onKeyDown(e) {
    if (e.key == "Escape") {
      this.clear();
      return;
    }
    if (e.key == "Shift") {
      this.shiftStartIndex = this.lastIndex;
    }
    if (e.key != "ArrowDown" && e.key != "ArrowUp") {
      return;
    }
    e.preventDefault();
    let index = 0, change = false;
    if (e.key == "ArrowDown") {
      if (this.lastIndex == this.list.store.count() - 1) {
        return;
      }
      index = this.lastIndex + 1;
    } else if (e.key == "ArrowUp") {
      if (this.lastIndex == 0) {
        return;
      }
      index = this.lastIndex - 1;
    }
    if (e.shiftKey && this.multiSelect) {
      const selected = this.selected;
      if (e.key == "ArrowDown" && index > this.shiftStartIndex || e.key == "ArrowUp" && index < this.shiftStartIndex) {
        if (selected.indexOf(index) == -1) {
          selected.push(index);
        }
      } else {
        const removeIndex = selected.indexOf(this.lastIndex);
        if (removeIndex > -1) {
          selected.splice(removeIndex, 1);
        }
      }
      change = this.setSelected(selected, true);
    } else {
      change = this.setSelected([index], true);
      this.list.focusRow(index);
    }
    if (change && !this.hasKeyUpListener) {
      this.hasKeyUpListener = true;
      this.list.el.addEventListener("keyup", () => {
        this.fire("selectionchange", this);
        this.hasKeyUpListener = false;
      }, { once: true });
    }
    this.lastIndex = index;
  }
};
var rowselect = (config) => createComponent(new RowSelect(config.list), config);

// goui/script/DragData.ts
var dragData = {};

// goui/script/component/List.ts
var dropPin = comp({
  cls: "drop-pin",
  hidden: true
});
root.items.add(dropPin);
var List = class extends Component {
  constructor(store2, renderer, tagName = "ul") {
    super(tagName);
    this.store = store2;
    this.renderer = renderer;
    /**
     * Shown when the list is empty.
     */
    this.emptyStateHtml = `<div class="goui-empty-state"><i class="icon">article</i><p>${t("Nothing to show")}</p></div>`;
    /**
     * Allow items to be dragged
     */
    this.draggable = false;
    /**
     * Allow to drop between items
     */
    this.dropBetween = false;
    /**
     * Allow to drop on items
     */
    this.dropOn = false;
    this.itemTag = "li";
    this.tabIndex = 0;
    const bf = new BufferedFunction(100, this.mask.bind(this));
    store2.on("beforeload", () => {
      bf.buffer();
    });
    store2.on("load", () => {
      if (bf.pending()) {
        bf.cancel();
      } else {
        this.unmask();
      }
    });
  }
  // protected fragment?: DocumentFragment;
  /**
   * Row selection object
   * @param rowSelectionConfig
   */
  set rowSelectionConfig(rowSelectionConfig) {
    if (typeof rowSelectionConfig != "boolean") {
      rowSelectionConfig.list = this;
      this.rowSelect = rowselect(rowSelectionConfig);
    } else {
      this.rowSelect = rowselect({ list: this });
    }
  }
  get rowSelection() {
    return this.rowSelect;
  }
  get el() {
    return super.el;
  }
  internalRender() {
    const el = super.internalRender();
    this.initNavigateEvent();
    el.on("mousedown", (e) => {
      this.onMouseEvent(e, "rowmousedown");
    }).on("dblclick", (e) => {
      this.onMouseEvent(e, "rowdblclick");
    }).on("click", (e) => {
      this.onMouseEvent(e, "rowclick");
    }).on("contextmenu", (e) => {
      this.onMouseEvent(e, "rowcontextmenu");
    });
    this.renderBody();
    this.initStore();
    return el;
  }
  initStore() {
    this.store.on("remove", (collection, item, index) => {
      var _a;
      const rows = this.getRowElements();
      (_a = rows[index]) == null ? void 0 : _a.remove();
      if (this.rowSelection) {
        this.rowSelection.remove(index, true);
      }
    });
    this.store.on("add", (collection, item, index) => {
      this.onRecordAdd(collection, item, index);
    });
  }
  onRecordAdd(collection, item, index) {
    const container = this.renderGroup(item);
    if (index == collection.count() - 1) {
      container.append(this.renderRow(item, index));
    } else {
      const before = container.children[index];
      container.insertBefore(this.renderRow(item, index), before);
    }
  }
  getRowElements() {
    return Array.from(this.el.querySelectorAll(":scope > .data"));
  }
  initNavigateEvent() {
    this.on("rowmousedown", (list2, storeIndex, row, ev) => {
      if (!ev.shiftKey && !ev.ctrlKey) {
        this.fire("navigate", this, storeIndex);
      }
    });
    if (this.rowSelection) {
      this.el.addEventListener("keydown", (ev) => {
        if (!ev.shiftKey && !ev.ctrlKey && (ev.key == "ArrowDown" || ev.key == "ArrowUp")) {
          const selected = this.rowSelect.selected;
          if (selected.length) {
            const storeIndex = selected[0];
            this.fire("navigate", this, storeIndex);
          }
        }
      });
    }
  }
  renderEmptyState() {
    this.emptyStateEl = E("li");
    this.emptyStateEl.hidden = this.store.count() > 0;
    this.emptyStateEl.innerHTML = this.emptyStateHtml;
    this.el.appendChild(this.emptyStateEl);
  }
  clearRows() {
    this.el.innerHTML = "";
  }
  renderBody() {
    this.renderRows(this.store.items);
    if (this.rowSelect) {
      this.rowSelect.on("rowselect", (rowSelect, storeIndex) => {
        const tr = this.getRowElements()[storeIndex];
        if (!tr) {
          console.error("No row found for selected index: " + storeIndex + ". Maybe it's not rendered yet?");
          return;
        }
        tr.classList.add("selected");
      });
      this.rowSelect.on("rowdeselect", (rowSelect, storeIndex) => {
        const tr = this.getRowElements()[storeIndex];
        if (!tr) {
          console.error("No row found for selected index: " + storeIndex + ". Maybe it's not rendered yet?");
          return;
        }
        tr.classList.remove("selected");
      });
    }
  }
  focusRow(index) {
    const tr = this.getRowElements()[index];
    if (tr) {
      tr.focus();
    }
  }
  renderRows(records) {
    for (let i = 0, l = records.length; i < l; i++) {
      const container = this.renderGroup(records[i]), row = this.renderRow(records[i], i);
      if (this.rowSelection && this.rowSelection.selected.indexOf(i) > -1) {
        row.cls("+selected");
      }
      container.append(row);
    }
    this.fire("renderrows", this, records);
  }
  renderGroup(record) {
    return this.el;
  }
  renderRow(record, storeIndex) {
    const row = E(this.itemTag).cls("+data").attr("tabindex", "0");
    if (this.draggable) {
      row.draggable = true;
      row.ondragstart = this.onNodeDragStart.bind(this);
    }
    if (this.rowSelection && this.rowSelection.selected.indexOf(storeIndex) > -1) {
      row.classList.add("selected");
    }
    this.bindDropEvents(row);
    const r = this.renderer(record, row, this, storeIndex);
    if (typeof r === "string") {
      row.innerHTML = r;
    } else if (Array.isArray(r)) {
      r.forEach((c) => {
        c.parent = this;
        c.render(row);
      });
    }
    return row;
  }
  // private onScroll() {
  // 	const el = this.el;
  // 	const pixelsLeft = el.scrollHeight - el.scrollTop - el.offsetHeight;
  //
  // 	if (pixelsLeft < 100) {
  // 		if (!this.store.loading && this.store.hasNext()) {
  // 			this.store.loadNext(true).finally(() => {
  // 				this.fire("scrolleddown", this);
  // 			});
  // 		}
  // 	}
  // }
  onMouseEvent(e, type) {
    const row = this.findRowByEvent(e), index = row ? this.getRowElements().indexOf(row) : -1;
    if (index !== -1) {
      this.fire(type, this, index, row, e);
    }
  }
  findRowByEvent(e) {
    return e.target.closest(".data");
  }
  bindDropEvents(row) {
    row.ondrop = this.onNodeDrop.bind(this);
    row.ondragend = this.onNodeDragEnd.bind(this);
    row.ondragover = this.onNodeDragOver.bind(this);
    row.ondragenter = this.onNodeDragEnter.bind(this);
    row.ondragleave = this.onNodeDragLeave.bind(this);
  }
  onNodeDragStart(e) {
    e.stopPropagation();
    const row = e.target;
    e.dataTransfer.setData("text/plain", "goui");
    dragData.row = row;
    dragData.cmp = this;
    dragData.storeIndex = this.getRowElements().indexOf(row);
    dragData.record = this.store.get(dragData.storeIndex);
    root.el.cls("+dragging");
  }
  onNodeDragEnd(e) {
    root.el.cls("-dragging");
  }
  onNodeDragEnter(e) {
    const dropRow = this.findDropRow(e);
    if (this.dropAllowed(e, dropRow)) {
      e.stopPropagation();
      e.preventDefault();
      dropRow.cls("+drag-over");
      this.onNodeDragEnterAllowed(e, dropRow);
    }
  }
  onNodeDragEnterAllowed(e, dropRow) {
  }
  onNodeDragLeave(e) {
    const dropRow = this.findDropRow(e);
    if (this.dropAllowed(e, dropRow)) {
      e.stopPropagation();
      e.preventDefault();
      this.clearOverClasses(dropRow);
      this.onNodeDragLeaveAllowed(e, dropRow);
    }
  }
  onNodeDragLeaveAllowed(e, dropRow) {
  }
  findDropRow(e) {
    return e.target.closest("LI");
  }
  clearOverClasses(dropRow) {
    dropRow.cls("-drag-over");
    dropRow.classList.remove("before");
    dropRow.classList.remove("after");
    dropRow.classList.remove("on");
    dropPin.hidden = true;
  }
  onNodeDragOver(e) {
    if (!this.dropOn && !this.dropBetween) {
      return;
    }
    const dropRow = this.findDropRow(e);
    if (this.dropAllowed(e, dropRow)) {
      e.stopPropagation();
      e.preventDefault();
      const pos = this.getDropPosition(e);
      dropRow.classList.toggle("before", "before" == pos);
      dropRow.classList.toggle("after", "after" == pos);
      dropRow.classList.toggle("on", "on" == pos);
      dropPin.hidden = pos == "on" || pos == void 0;
      const dropRowRect = dropRow.getBoundingClientRect();
      dropPin.el.style.top = (pos == "before" ? dropRowRect.y : dropRowRect.y + dropRowRect.height - 1) + "px";
      dropPin.el.style.left = dropRowRect.x + "px";
      dropPin.el.style.width = dropRowRect.width + "px";
    }
  }
  dropAllowed(e, dropRow) {
    return this.fire("dropallowed", this, e, dropRow, dragData);
  }
  getDropPosition(e) {
    if (!this.dropBetween) {
      return this.dropOn ? "on" : void 0;
    }
    const betweenZone = 6;
    if (e.offsetY < betweenZone) {
      return "before";
    } else if (e.offsetY > e.target.offsetHeight - betweenZone) {
      return "after";
    } else {
      return this.dropOn ? "on" : void 0;
    }
  }
  onNodeDrop(e) {
    const dropPos = this.getDropPosition(e);
    if (!dropPos) {
      return;
    }
    e.preventDefault();
    e.stopPropagation();
    const dropRow = this.findDropRow(e), dropIndex = this.getRowElements().indexOf(dropRow);
    this.clearOverClasses(dropRow);
    this.fire("drop", this, e, dropRow, dropIndex, dropPos, dragData);
  }
};
var list = (config) => createComponent(new List(config.store, config.renderer), config);

// goui/script/component/table/Table.ts
var Table = class extends List {
  /**
   *
   * @param store Store to provide data
   * @param columns The table columns
   */
  constructor(store2, columns) {
    super(store2, (record, row, me, storeIndex) => {
      for (let c of this.columns) {
        c.parent = this;
        if (c.hidden) {
          continue;
        }
        const td = document.createElement("td");
        if (c.align) {
          td.style.textAlign = c.align;
        }
        if (c.cls) {
          td.classList.add(...c.cls.split(" "));
        }
        let value = c.property ? ObjectUtil.path(record, c.property) : void 0;
        if (c.renderer) {
          const r = c.renderer(value, record, td, this, storeIndex);
          if (typeof r === "string") {
            td.innerHTML = r;
          } else if (r instanceof Component) {
            r.parent = this;
            r.render(td);
          } else {
            r.then((s) => {
              if (s instanceof Component) {
                s.parent = this;
                s.render(td);
              } else {
                td.innerHTML = s;
              }
            });
          }
        } else {
          td.innerText = value ? value : "";
        }
        row.append(td);
      }
    }, "table");
    this.store = store2;
    this.columns = columns;
    /**
     * Make the table fits its container in width by setting min-width: 100%
     * Defaults to true
     */
    this.fitParent = false;
    /**
     * Show headers
     */
    this.headers = true;
    /**
     * Group renderer function
     */
    this.groupByRenderer = (groupBy) => groupBy != null ? groupBy : t("None");
    this.minCellWidth = 30;
    this.baseCls = "goui-table";
    this.itemTag = "tr";
    this.colsAreFixed = false;
  }
  initStore() {
    super.initStore();
    this.store.on("remove", (collection, item, index) => {
      if (this.groupBy) {
        this.el.querySelectorAll("tbody").forEach((tbody) => {
          if (tbody.children.length == 1) {
            tbody.remove();
          }
        });
      }
    });
  }
  getRowElements() {
    return Array.from(this.el.getElementsByClassName("data"));
  }
  internalRemove() {
    if (this.columnMenu) {
      this.columnMenu.remove();
    }
    return super.internalRemove();
  }
  restoreState(state) {
    if (state.sort) {
      this.store.sort = state.sort;
    }
    if (state.columns) {
      for (let id in state.columns) {
        let col = this.findColumnById(id);
        if (col) {
          Object.assign(col, state.columns[id]);
        }
      }
    }
  }
  /**
   * Find column by "property" property.
   *
   * It's the property path of the data linked to the column
   *
   * @param id
   */
  findColumnById(id) {
    return this.columns.find((col) => {
      return col.id == id;
    });
  }
  buildState() {
    const cols = {};
    this.columns.forEach((c) => {
      cols[c.id] = {
        width: c.width,
        hidden: c.hidden
      };
    });
    return {
      sort: this.store.sort,
      columns: cols
    };
  }
  renderBody() {
    if (this.fitParent) {
      this.el.style.minWidth = "100%";
    }
    if (this.headers) {
      this.renderHeaders();
    } else {
      this.renderColGroup();
    }
    super.renderBody();
  }
  rerender() {
    const el = this.el;
    this.groupEl = void 0;
    el.innerHTML = "";
    this.renderBody();
  }
  showColumnMenu(ev) {
    ev.preventDefault();
    if (!this.columnMenu) {
      this.columnMenu = menu({
        isDropdown: true,
        removeOnClose: false
      });
      this.columns.forEach((c) => {
        if (c.header && c.hidable) {
          this.columnMenu.items.add(checkbox({
            label: c.header,
            name: c.id,
            value: !c.hidden,
            listeners: {
              change: (field) => {
                c.hidden = !field.value;
                this.saveState();
                this.rerender();
              }
            }
          }));
        }
      });
    }
    this.columnMenu.showAt(ev);
  }
  createColumnSplitter(h, header, colIndex) {
    if (h.resizable) {
      const splitter2 = draggable({
        tagName: "hr",
        setPosition: false,
        parent: this
      });
      splitter2.on("dragstart", (cmp, dragData2) => {
        if (!this.colsAreFixed) {
          this.fixColumnWidths();
        }
        dragData2.data.startWidth = header.offsetWidth;
        splitter2.dragConstrainTo(this.headersRow, {
          left: this.calcTableWidth(colIndex) + this.minCellWidth,
          right: -1e4
        });
      });
      splitter2.on("drag", (cmp, dragData2) => {
        const w = dragData2.data.startWidth + dragData2.x - dragData2.startX;
        header.style.width = w + "px";
        h.width = w;
        this.el.style.width = this.calcTableWidth() + "px";
      });
      splitter2.on("drop", () => {
        this.saveState();
      });
      return splitter2;
    } else {
      return comp({ tagName: "hr" });
    }
  }
  renderColGroup() {
    const colGroup = document.createElement("colgroup");
    let index = -1;
    for (let h of this.columns) {
      index++;
      if (h.hidden) {
        continue;
      }
      const col = document.createElement("col");
      if (h.width) {
        col.style.width = h.width + "px";
      }
      if (h.align) {
        col.style.textAlign = h.align;
      }
      if (h.cls) {
        col.classList.add(...h.cls.split(" "));
      }
      colGroup.appendChild(col);
    }
    this.el.appendChild(colGroup);
    return colGroup;
  }
  renderHeaders() {
    const thead = document.createElement("thead");
    this.headersRow = document.createElement("tr");
    this.headersRow.addEventListener("contextmenu", (ev) => {
      this.showColumnMenu(ev);
    });
    let index = -1;
    for (let h of this.columns) {
      index++;
      if (h.hidden) {
        continue;
      }
      const header = document.createElement("th");
      if (h.width) {
        header.style.width = h.width + "px";
      }
      if (h.align) {
        header.style.textAlign = h.align;
      }
      if (h.cls) {
        header.classList.add(...h.cls.split(" "));
      }
      if (h.headerRenderer) {
        const r = h.headerRenderer(h, header, this);
        if (typeof r === "string") {
          header.innerHTML = r;
        } else if (r instanceof Component) {
          r.render(header);
        }
      } else {
        header.innerHTML = h.header || "";
      }
      h.headerEl = header;
      if (h.resizable) {
        const splitter2 = this.createColumnSplitter(h, header, index);
        splitter2.render(header);
      } else {
        comp({ tagName: "hr" }).render(header);
      }
      if (h.sortable && h.property) {
        header.addEventListener("click", () => {
          this.onSort(h.property, header);
        });
        const sort = this.store.sort;
        if (sort.length) {
          sort.forEach((comparator) => {
            if (h.property == comparator.property) {
              header.classList.add("sorted");
              header.classList.add(comparator.isAscending || comparator.isAscending === void 0 ? "asc" : "desc");
            }
          });
        }
      }
      this.headersRow.appendChild(header);
    }
    thead.appendChild(this.headersRow);
    this.el.appendChild(thead);
    return this.headersRow;
  }
  onSort(dataIndex, header) {
    this.fire("sort", this, dataIndex);
    const s = this.store.sort;
    let sortIndex = s.length - 1 || 0;
    if (s[sortIndex]) {
      if (s[sortIndex].property == dataIndex) {
        s[sortIndex].isAscending = !s[sortIndex].isAscending;
      } else {
        s[sortIndex].property = dataIndex;
      }
    } else {
      s[sortIndex] = {
        isAscending: true,
        property: dataIndex
      };
    }
    this.headersRow.childNodes.forEach((node) => {
      let th = node;
      if (th == header) {
        th.classList.add("sorted");
        th.classList.remove(s[sortIndex].isAscending ? "desc" : "asc");
        th.classList.add(s[sortIndex].isAscending ? "asc" : "desc");
      } else {
        th.classList.remove("sorted");
        th.classList.remove("asc");
        th.classList.remove("desc");
      }
    });
    this.store.reload().catch((reason) => {
      Notifier.error(reason);
    });
    this.saveState();
  }
  /**
   * When resizing columns we must calculate absolute pixel widths
   *
   * @private
   */
  fixColumnWidths() {
    this.columns.forEach((col) => {
      if (!col.hidden) {
        col.width = col.headerEl.offsetWidth;
        col.headerEl.style.width = col.width + "px";
      }
    });
    this.el.style.minWidth = "";
    this.el.style.width = this.calcTableWidth() + "px";
  }
  /**
   * Returns the sum of column widths
   *
   * @param untilColumnIndex Calc width until this column
   */
  calcTableWidth(untilColumnIndex = -1) {
    return this.columns.reduce((previousValue, nextValue, currentIndex) => {
      if (nextValue.hidden || untilColumnIndex > -1 && currentIndex >= untilColumnIndex) {
        return previousValue;
      }
      return previousValue + nextValue.width;
    }, 0);
  }
  renderGroup(record) {
    if (!this.groupBy) {
      if (!this.groupEl) {
        this.groupEl = document.createElement("tbody");
        this.el.append(this.groupEl);
      }
      return this.groupEl;
    }
    const groupBy = ObjectUtil.path(record, this.groupBy);
    if (!this.groupEl || groupBy != this.lastGroup) {
      const tr = document.createElement("tr");
      tr.classList.add("group");
      const th = document.createElement("th");
      th.colSpan = this.columns.length;
      const r = this.groupByRenderer(groupBy, record, th, this);
      if (typeof r === "string") {
        th.innerHTML = r;
      } else if (r instanceof Component) {
        r.render(th);
      } else if (r instanceof Promise) {
        r.then((s) => {
          if (s instanceof Component) {
            s.render(th);
          } else {
            th.innerHTML = s;
          }
        });
      } else {
        console.warn("Renderer returned invalid value: ", r);
      }
      tr.append(th);
      this.groupEl = document.createElement("tbody");
      this.groupEl.append(tr);
      this.el.append(this.groupEl);
      this.lastGroup = groupBy;
    }
    return this.groupEl;
  }
  clearRows() {
    if (!this.el) {
      return;
    }
    this.groupEl = void 0;
    this.el.querySelectorAll("tbody").forEach((tbody) => tbody.remove());
  }
  findDropRow(e) {
    return e.target.closest("TR");
  }
};
var table = (config) => createComponent(new Table(config.store, config.columns), config);

// goui/script/component/table/TableColumns.ts
var TableColumn = class extends Observable {
  /**
   *
   * The column ID. Also used for 'property'
   */
  constructor(id) {
    super();
    this.id = id;
    /**
     * Make the column resizable by the user
     */
    this.resizable = false;
    /**
     * Make it sortable by the user
     */
    this.sortable = false;
    /**
     * Text alignment
     */
    this.align = "left";
    /**
     * Hide the column. It can be enabled by the user via the context menu.
     */
    this.hidden = false;
    /**
     * Enable this column in the enabled columns menu
     */
    this.hidable = true;
    this.property = id;
  }
};
var column = (config) => createComponent(new TableColumn(config.id), config);
var DateTimeColumn = class extends TableColumn {
  constructor() {
    super(...arguments);
    this.renderer = (date) => {
      return Format.dateTime(date);
    };
    //argh!? https://stackoverflow.com/questions/43121661/typescript-type-inference-issue-with-string-literal
    this.align = "right";
    this.width = 190;
  }
};
var datetimecolumn = (config) => createComponent(new DateTimeColumn(config.id), config);
var DateColumn = class extends TableColumn {
  constructor() {
    super(...arguments);
    this.renderer = (date) => {
      return Format.date(date);
    };
    //argh!? https://stackoverflow.com/questions/43121661/typescript-type-inference-issue-with-string-literal
    this.align = "right";
    this.width = 128;
  }
};
var datecolumn = (config) => createComponent(new DateColumn(config.id), config);
var CheckboxColumn = class extends TableColumn {
  constructor(id) {
    super(id);
    this.renderer = (val) => {
      return checkbox({
        value: val
      });
    };
    this.cls = "checkbox-select-column";
  }
};
var checkboxcolumn = (config) => createComponent(new CheckboxColumn(config.id), config);
var CheckboxSelectColumn = class extends TableColumn {
  constructor() {
    super("checkboxselect");
    this.headerRenderer = (col, headerEl, table2) => {
      return checkbox({
        listeners: {
          change: (field, newValue, oldValue) => {
            if (newValue) {
              table2.rowSelection.selectAll();
            } else {
              table2.rowSelection.clear();
            }
          }
        }
      });
    };
    this.renderer = (val, record, td, table2, rowIndex) => {
      return checkbox({
        value: val,
        listeners: {
          render: (field) => {
            field.el.addEventListener("mousedown", (ev) => {
              ev.stopPropagation();
            });
            table2.rowSelection.on("selectionchange", () => {
              field.value = table2.rowSelection.selected.indexOf(rowIndex) > -1;
            });
          },
          change: (field, newValue, oldValue) => {
            const index = table2.store.indexOf(record), selected = table2.rowSelection.selected;
            if (newValue) {
              selected.push(index);
              table2.rowSelection.selected = selected;
            } else {
              table2.rowSelection.selected = selected.filter((i) => i != index);
            }
          }
        }
      });
    };
    this.hidable = false;
    this.cls = "checkbox-select-column";
  }
};
var checkboxselectcolumn = (config) => createComponent(new CheckboxSelectColumn(), config);
var menucolumn = (...items) => column({
  width: 48,
  id: "btn",
  // headerRenderer: (col: TableColumn, headerEl: HTMLTableCellElement, table: Table) => {
  // 	headerEl.style.position = "sticky";
  // 	headerEl.style.right = "0";
  // 	return "";
  // },
  renderer: (columnValue, record, td, table2, rowIndex) => {
    items.forEach((i) => {
      i.dataSet.table = table2;
      i.dataSet.rowIndex = rowIndex;
    });
    return btn({
      icon: "more_vert",
      menu: menu({}, ...items)
    });
  }
});

// goui/script/component/form/ComboBox.ts
var ComboBox = class extends AutocompleteField {
  constructor(dataSource, displayProperty = "name", valueProperty = "id") {
    const dropDownTable = table({
      headers: false,
      fitParent: true,
      store: datasourcestore({
        dataSource,
        queryParams: {
          limit: 50
        }
      }),
      columns: [
        column({
          id: displayProperty,
          resizable: true,
          width: 312,
          sortable: true
        })
      ]
    });
    super(dropDownTable);
    this.dataSource = dataSource;
    this.displayProperty = displayProperty;
    this.valueProperty = valueProperty;
    this.picker.list.store.addScrollLoader(this.menu.el);
    this.on("autocomplete", (_field, input) => __async(this, null, function* () {
      var _a;
      this.list.store.queryParams.position = 0;
      if (!this.list.store.queryParams.filter) {
        this.list.store.queryParams.filter = {};
      }
      this.list.store.queryParams.filter[(_a = this.filterName) != null ? _a : this.displayProperty] = input;
      yield this.list.store.load();
    }));
  }
  /**
   * Pass config options for the underlying dropdown table store
   * @param storeConfig
   */
  set storeConfig(storeConfig) {
    ObjectUtil.merge(this.picker.list.store, storeConfig);
  }
  pickerRecordToValue(_field, record) {
    return record[this.valueProperty];
  }
  valueToTextField(_field, value) {
    return __async(this, null, function* () {
      if (value === "") {
        return "";
      }
      const nb = yield this.dataSource.single(value);
      return nb ? nb[this.displayProperty] : "?";
    });
  }
};
var combobox = (config) => {
  var _a, _b;
  return createComponent(new ComboBox(config.dataSource, (_a = config.displayProperty) != null ? _a : "name", (_b = config.valueProperty) != null ? _b : "id"), config);
};

// goui/script/component/form/DurationField.ts
var DurationField = class extends TextField {
  constructor() {
    super(...arguments);
    this.type = "time";
    this.min = 0;
    this.max = 1440;
    this.inMinutes = true;
  }
  validate() {
    super.validate();
    const v = Format.minutes(super.value);
    if (v && isNaN(v)) {
      this.setInvalid("Incorrect number format");
    }
    if (this.max !== void 0 && v > this.max) {
      this.setInvalid("Number is too big");
    }
    if (this.min !== void 0 && v < this.min) {
      this.setInvalid("Number is too small");
    }
  }
  set value(v) {
    if (v && isNaN(v)) {
      throw new Error("Invalid number");
    } else if (v) {
      super.value = Format.duration(v, true);
    }
  }
  get value() {
    const v = super.value;
    if (typeof v === "string") {
      return Format.minutes(v);
    }
    return v;
  }
  createControl() {
    const v = this.value, name = this.name;
    this._input = document.createElement("input");
    this._input.classList.add("text");
    this._input.type = "time";
    this._input.required = this.required;
    if (name) {
      this._input.name = name;
    }
    this._input.readOnly = this.readOnly;
    if (this.title) {
      this._input.title = this.title;
    }
    if (this.invalidMsg) {
      this.applyInvalidMsg();
    }
    if (this.min !== void 0) {
      this._input.attr("min", this.min);
    }
    if (this.max !== void 0) {
      this._input.attr("max", this.max);
    }
    if (this.interval) {
      this._input.attr("step", this.interval);
    }
    if (v !== void 0) {
      this._input.attr("value", Format.duration(v, true));
    }
    return this._input;
  }
};
var durationfield = (config) => createComponent(new DurationField(), config);

// goui/script/component/Avatar.ts
var _Avatar = class _Avatar extends Component {
  constructor() {
    super(...arguments);
    this.baseCls = "goui-avatar";
  }
  set displayName(displayName) {
    this.text = this.initials(displayName);
    this.title = displayName;
    let j = 0;
    for (let i = 0, l = this.displayName.length; i < l; i++) {
      j += displayName.charCodeAt(i);
    }
    this.el.style.backgroundColor = "#" + _Avatar.colors[j % _Avatar.colors.length];
  }
  get displayName() {
    return this.title;
  }
  set backgroundImage(imgUrl) {
    this.el.style.backgroundImage = imgUrl ? "url(" + imgUrl + ")" : "none";
    if (imgUrl) {
      this.text = "";
    }
  }
  get backgroundImage() {
    return this.el.style.backgroundImage;
  }
  /**
   * Grabs the first char of the first and last word.
   *
   * @param {string} name
   * @returns {string}
   */
  initials(name) {
    const parts = name.split(" "), l = parts.length;
    if (l > 2) {
      parts.splice(1, l - 2);
    }
    return parts.map((name2) => name2.substr(0, 1).toUpperCase()).join("");
  }
};
_Avatar.colors = [
  "C62828",
  "AD1457",
  "6A1B9A",
  "4527A0",
  "283593",
  "1565C0",
  "0277BD",
  "00838F",
  "00695C",
  "2E7D32",
  "558B2F",
  "9E9D24",
  "F9A825",
  "FF8F00",
  "EF6C00",
  "424242"
];
var Avatar = _Avatar;
var avatar = (config) => createComponent(new Avatar(), config);

// goui/script/component/CardMenu.ts
var CardMenu = class extends Component {
  constructor() {
    super("menu");
    this.baseCls = "goui-cardmenu";
    this.on("beforerender", () => {
      if (!this.cardContainer) {
        this.cardContainer = this.parent.findChildByType(CardContainer);
      }
      this.cardContainer.on("cardchange", (cardContainer, index) => {
        this.updateActiveTab();
      });
      this.createMenu();
      this.cardContainer.on("beforerender", () => {
        this.updateActiveTab();
      });
    });
  }
  focus(o) {
    var _a;
    const first = (_a = this.cardContainer) == null ? void 0 : _a.items.first();
    if (first) {
      first.focus(o);
    } else {
      super.focus(o);
    }
  }
  updateActiveTab() {
    const activeItem = this.cardContainer.items.get(this.cardContainer.activeItem);
    this.items.forEach((item, menuIndex) => {
      if (activeItem && (item.itemId == activeItem.itemId || item.itemId == activeItem.id)) {
        item.el.classList.add("active");
      } else {
        item.el.classList.remove("active");
      }
    });
  }
  createMenu() {
    this.cardContainer.items.forEach((item, index) => {
      if (!item.itemId) {
        item.itemId = "card-" + index;
      }
      if (this.findItem(item.itemId)) {
        return;
      }
      this.items.insert(
        index,
        btn({
          type: "button",
          itemId: item.itemId,
          cls: index == this.cardContainer.activeItem ? "active" : "",
          text: item.title,
          handler: () => {
            this.cardContainer.activeItem = item;
          }
        })
      );
      item.title = "";
    });
  }
};
var cardmenu = (config, ...items) => createComponent(new CardMenu(), config, items);

// goui/script/component/DescriptionList.ts
var DescriptionList = class extends Component {
  constructor() {
    super("dl");
    this.baseCls = "goui-dl";
  }
  internalRender() {
    const el = super.internalRender();
    this.renderList();
    return el;
  }
  /**
   * Set the records to display
   *
   * @example
   * ```
   * const records: DLRecord = [
   * 			['Number', record.number],
   * 			['Description', record.description],
   * 			['Created At', Format.date(record.createdAt)]
   * 		];
   * dl.setRecords(records);
   * ```
   * @param records
   */
  set records(records) {
    this._records = records;
    this.renderList();
  }
  get records() {
    return this._records || [];
  }
  renderList() {
    this.items.clear();
    this.records.forEach((record) => {
      this.items.add(comp({
        tagName: "dt",
        text: record.shift()
      }));
      record.forEach((r) => {
        const dd = comp({
          tagName: "dd"
        });
        if (typeof r == "function")
          r(dd);
        else {
          dd.text = r + "";
        }
        this.items.add(dd);
      });
    });
  }
};
var dl = (config, ...items) => createComponent(new DescriptionList(), config, items);

// goui/script/component/Paginator.ts
var Paginator = class extends Toolbar {
  constructor(store2) {
    super();
    this.store = store2;
    this.baseCls = "goui-toolbar goui-paginator";
    this.items.add(
      this.prev = btn({
        icon: "chevron_left",
        text: "Previous",
        disabled: true,
        handler: () => __async(this, null, function* () {
          yield this.store.loadPrevious();
        })
      }),
      comp({
        flex: 1
      }),
      this.next = btn({
        icon: "chevron_right",
        text: "Next",
        disabled: true,
        handler: () => __async(this, null, function* () {
          yield this.store.loadNext();
        })
      })
    );
    this.store.on("load", () => {
      this.onStoreLoad();
    });
  }
  onStoreLoad() {
    this.prev.disabled = !this.store.hasPrevious();
    this.next.disabled = !this.store.hasNext();
  }
};
var paginator = (config) => createComponent(new Paginator(config.store), config);

// goui/script/component/SearchButton.ts
var SearchButton = class extends Button {
  constructor() {
    super();
    this.buffer = 300;
    this.handler = (button, ev) => {
      this.mainTbar = button.parent;
      this.getSearchTBar().show();
      this.searchField.focus();
    };
    this.icon = "search";
    this.title = t("Search");
    this.searchField = textfield({
      label: t("Search"),
      flex: 1,
      buttons: [
        btn({
          type: "button",
          icon: "clear",
          handler: () => {
            this.reset();
          }
        })
      ]
    });
    this.searchField.on("render", () => {
      this.searchField.input.addEventListener("input", FunctionUtil.buffer(this.buffer, this.onInput.bind(this)));
      this.searchField.el.addEventListener("keydown", (e) => {
        if (e.key == "Enter") {
          e.preventDefault();
          e.stopPropagation();
        }
        if (e.key == "Escape") {
          e.preventDefault();
          e.stopPropagation();
          this.close();
        }
      });
    });
  }
  reset() {
    this.searchField.reset();
    this.close();
    this.fire("reset", this);
    this.fire("input", this, "");
    this.el.classList.remove("accent");
    this.el.classList.remove("filled");
  }
  close() {
    this.searchTBar.hide();
    this.mainTbar.show();
    this.focus();
  }
  onInput() {
    this.el.classList.toggle("accent", !!this.searchField.value);
    this.el.classList.toggle("filled", !!this.searchField.value);
    this.fire("input", this, this.searchField.value);
  }
  getSearchTBar() {
    if (!this.searchTBar) {
      this.searchTBar = tbar(
        {
          cls: "search"
        },
        btn({
          icon: "chevron_left",
          title: t("Back"),
          handler: () => {
            this.close();
          }
        }),
        this.searchField
      );
      this.mainTbar.items.add(this.searchTBar);
    }
    return this.searchTBar;
  }
};
var searchbtn = (config) => createComponent(new SearchButton(), config);

// goui/script/component/Splitter.ts
var Splitter = class extends DraggableComponent {
  /**
   *
   * @param resizeComponentPredicate 	The component to resize in height or width.
   *    Can be a component ID, itemId property, Component instance or custom function
   *
   */
  constructor(resizeComponentPredicate) {
    super("hr");
    this.resizeComponentPredicate = resizeComponentPredicate;
    /**
     * The minimum size it will set. Note that you can also put a min-width or min-height on the element with css.
     */
    this.minSize = 50;
    this.on("beforerender", () => {
      if (!(resizeComponentPredicate instanceof Component)) {
        this._resizeComponent = this.parent.findChild(this.resizeComponentPredicate);
        if (!this._resizeComponent) {
          console.warn(this.resizeComponentPredicate);
          throw "Could not find component to resize!";
        }
      } else {
        this._resizeComponent = resizeComponentPredicate;
      }
      if (this._state) {
        this.restoreState(this._state);
      }
    });
  }
  restoreState(state) {
    this._state = state;
    if (this._resizeComponent) {
      if (this._state) {
        if (this._state.width)
          this._resizeComponent.width = this._state.width;
        if (this._state.height)
          this._resizeComponent.height = this._state.height;
      }
    }
  }
  buildState() {
    return this.resizeWidth ? { width: this._resizeComponent.el.offsetWidth } : { height: this._resizeComponent.el.offsetHeight };
  }
  internalRender() {
    const el = super.internalRender();
    this.on("dragstart", (comp2, dragData2, e) => {
      if (this.resizeWidth === void 0) {
        this.resizeWidth = this.el.offsetHeight > this.el.offsetWidth;
      }
      if (this.invert === void 0) {
        const splitterIndex = this.parent.findItemIndex(this);
        const resizeCmpIndex = this.parent.findItemIndex(this._resizeComponent);
        this.invert = splitterIndex < resizeCmpIndex;
      }
      if (this.resizeWidth) {
        dragData2.data.startWidth = this._resizeComponent.el.offsetWidth;
      } else {
        dragData2.data.startHeight = this._resizeComponent.el.offsetHeight;
      }
    });
    this.on("drag", (dc, dragData2, ev) => {
      if (this.resizeWidth) {
        let offset = dragData2.x - dragData2.startX;
        if (this.invert) {
          offset *= -1;
        }
        let width = Math.max(this.minSize, dragData2.data.startWidth + offset);
        if (this.maxSize) {
          width = Math.min(this.maxSize, width);
        }
        this._resizeComponent.width = width;
      } else {
        let offset = dragData2.y - dragData2.startY;
        if (this.invert) {
          offset *= -1;
        }
        let height = Math.max(this.minSize, dragData2.data.startHeight + offset);
        if (this.maxSize) {
          height = Math.min(this.maxSize, height);
        }
        this._resizeComponent.height = height;
      }
    });
    this.on("drop", () => {
      this.saveState();
    });
    return el;
  }
};
var splitter = (config) => createComponent(new Splitter(config.resizeComponentPredicate), config);

// goui/script/component/Tree.ts
var TreeRowRenderer = (record, row, me, storeIndex) => {
  const node = E("div").cls("node"), caret = E("span").cls("caret"), icon = E("i").cls("icon"), label = E("label");
  icon.innerText = record.icon || "folder";
  label.innerText = record.text;
  if (record.children && record.children.length == 0) {
    row.cls("+no-children");
  }
  if (record.cls) {
    row.cls("+" + record.cls);
  }
  node.append(caret, icon);
  if (record.check != void 0) {
    const c = checkbox({
      value: record.check,
      listeners: {
        change: (field, newValue, oldValue) => {
          record.check = newValue;
          const top = me.findTopTree();
          top.fire("checkchange", top, void 0, record, storeIndex, newValue);
        }
      }
    });
    c.render(node);
  }
  node.append(label);
  row.append(node);
};
var Tree = class _Tree extends List {
  constructor(renderer = TreeRowRenderer) {
    super(store(), renderer);
    this.renderer = renderer;
    this.subTrees = {};
    this.expandedIds = {};
    this.parentStoreIndex = -1;
    this.baseCls = "goui goui-tree";
    this.on("rowclick", (list2, storeIndex, row, ev) => {
      if (list2 == this) {
        void this.expand(row);
      }
    });
    this.store.on("remove", (collection, item, index) => {
      delete this.subTrees[index];
    });
    if (!(this.parent instanceof _Tree)) {
      this.on("render", () => {
        this.fire("expand", this, this, void 0, -1);
      });
    }
    this.emptyStateHtml = "";
  }
  set data(records) {
    this.store.loadData(records);
  }
  /**
   * The full hierarchical store data of the tree
   */
  get data() {
    return this.store.data;
  }
  /**
   * Reload the tree if loaded via the "expand" event.
   */
  reload() {
    this.fireExpand();
    for (let id in this.subTrees) {
      this.subTrees[id].reload();
    }
  }
  fireExpand() {
    if (!(this.parent instanceof _Tree)) {
      this.fire("expand", this, this, void 0, -1);
    } else {
      const record = this.parent.store.get(this.parentStoreIndex), top = this.findTopTree();
      top.fire("expand", top, this, record, this.parentStoreIndex);
    }
  }
  /**
   * Find the first menu in the tree of submenu's
   */
  findTopTree() {
    if (this.parent instanceof _Tree) {
      return this.parent.findTopTree();
    } else {
      return this;
    }
  }
  expand(row) {
    row.cls("+expanded");
    const storeIndex = this.getRowElements().indexOf(row);
    if (this.subTrees[storeIndex]) {
      return this.subTrees[storeIndex];
    }
    const record = this.store.get(storeIndex);
    if (!record) {
      throw "Record not found";
    }
    return this.renderSubTree(row, record, storeIndex);
  }
  renderSubTree(row, record, storeIndex) {
    row.cls("+expanded");
    this.findTopTree().expandedIds[record.id] = true;
    const sub = new _Tree(this.renderer);
    sub.dropOn = this.dropOn;
    sub.dropBetween = this.dropBetween;
    sub.draggable = this.draggable;
    sub.parent = this;
    sub.parentStoreIndex = storeIndex;
    if (record.children) {
      sub.data = record.children;
    }
    sub.store.on("load", (store1, records, append) => {
      record.children = store1.data;
    });
    ["rowclick", "rowdblclick", "rowcontextmenu", "rowmousedown"].forEach((e) => {
      this.relayEvent(sub, e);
    });
    this.subTrees[storeIndex] = sub;
    sub.render(row);
    const top = this.findTopTree();
    top.fire("expand", top, sub, record, storeIndex);
    return sub;
  }
  collapse(row) {
    row.cls("-expanded");
    const storeIndex = this.getRowElements().indexOf(row);
    const record = this.store.get(storeIndex);
    if (!record) {
      throw "Record not found";
    }
    const top = this.findTopTree();
    top.expandedIds[record.id] = true;
    top.fire("collapse", top, this, record, storeIndex);
  }
  renderRow(record, storeIndex) {
    let row = super.renderRow(record, storeIndex);
    row.cls("+data");
    row.id = Component.uniqueID();
    if (this.draggable) {
      row.draggable = true;
      row.ondragstart = this.onNodeDragStart.bind(this);
    }
    row.addEventListener("contextmenu", (e) => {
      e.stopPropagation();
      e.preventDefault();
      const sub = this.expand(row);
      sub.reload();
    });
    this.bindDropEvents(row);
    row.getElementsByClassName("caret")[0].on("click", (e) => {
      row.has(".expanded") ? this.collapse(row) : this.expand(row);
      e.preventDefault();
      e.stopPropagation();
    });
    if (this.findTopTree().expandedIds[record.id]) {
      row.cls("+expanded");
      void this.renderSubTree(row, record, storeIndex);
    }
    return row;
  }
  onNodeDragEnterAllowed(e, dropRow) {
    clearTimeout(this.dragOverTimeout);
    setTimeout(() => {
      this.dragOverTimeout = setTimeout(() => {
        this.expand(dropRow);
      }, 1e3);
    });
  }
  onNodeDragLeaveAllowed(e, dropRow) {
    clearTimeout(this.dragOverTimeout);
  }
  dropAllowed(e, dropRow) {
    const topTree = this.findTopTree();
    return dragData.row && !dragData.row.contains(dropRow);
  }
  onNodeDrop(e) {
    const dropPos = this.getDropPosition(e);
    if (!dropPos) {
      return;
    }
    e.preventDefault();
    e.stopPropagation();
    const dropRow = this.findDropRow(e), dropIndex = this.getRowElements().indexOf(dropRow);
    this.clearOverClasses(dropRow);
    clearTimeout(this.dragOverTimeout);
    const dropTree = this.expand(dropRow);
    dragData.dropTree = this;
    dragData.childrenTree = dropTree;
    const topTree = this.findTopTree();
    topTree.fire("drop", topTree, e, dropRow, dropIndex, dropPos, dragData);
    return false;
  }
};
var tree = (config) => createComponent(new Tree(), config);
export {
  AbstractDataSource,
  ArrayField,
  ArrayUtil,
  AutocompleteChips,
  AutocompleteField,
  Avatar,
  BrowserStore,
  BufferedFunction,
  Button,
  CardContainer,
  CardMenu,
  CheckboxColumn,
  CheckboxField,
  CheckboxGroup,
  CheckboxSelectColumn,
  ChipsField,
  Collection,
  ColorField,
  ColorMenu,
  ColorPicker,
  ComboBox,
  CommitErrorType,
  Component,
  ContainerField,
  DataSourceForm,
  DataSourceStore,
  DateColumn,
  DateField,
  DatePicker,
  DateTime,
  DateTimeColumn,
  DescriptionList,
  DisplayField,
  DraggableComponent,
  DurationField,
  E,
  Field,
  Fieldset,
  Form,
  Format,
  FunctionUtil,
  HtmlField,
  List,
  ListPicker,
  MapField,
  Mask,
  Menu,
  MultiSelectToolbar,
  Notifier,
  NumberField,
  ObjectUtil,
  Observable,
  Paginator,
  QuoteStripper,
  RadioField,
  RangeField,
  Recurrence,
  RecurrenceField,
  RecurrencePicker,
  RestDataSource,
  Router,
  RowSelect,
  SearchButton,
  SelectField,
  Splitter,
  State,
  Store,
  Table,
  TableColumn,
  TextAreaField,
  TextField,
  Toolbar,
  Translate,
  Tree,
  TreeRowRenderer,
  Window2 as Window,
  arrayfield,
  autocomplete,
  autocompletechips,
  avatar,
  browser,
  browserStoreConnection,
  btn,
  cardmenu,
  cards,
  checkbox,
  checkboxcolumn,
  checkboxgroup,
  checkboxselectcolumn,
  chips,
  code,
  colorfield,
  colormenu,
  colorpicker,
  column,
  combobox,
  comp,
  containerfield,
  cookies,
  createComponent,
  datasourceform,
  datasourcestore,
  datecolumn,
  datefield,
  datepicker,
  datetimecolumn,
  displayfield,
  dl,
  draggable,
  durationfield,
  fieldset,
  form,
  h1,
  h2,
  h3,
  h4,
  hr,
  htmlfield,
  list,
  listpicker,
  mapfield,
  mask,
  menu,
  menucolumn,
  mstbar,
  numberfield,
  p,
  paginator,
  progress,
  radio,
  rangefield,
  recurrencefield,
  root,
  router,
  rowselect,
  searchbtn,
  section,
  select,
  small,
  splitter,
  store,
  t,
  table,
  tbar,
  tbarItems,
  textarea,
  textfield,
  translate,
  tree,
  win
};
/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Michael de Hart <mdhart@intermesh.nl>
 */
/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Joachim van de Haterd <jhaterd@intermesh.nl>
 */
//# sourceMappingURL=index.js.map
